/* -*- C++ -*- */
/*=============================================================================
//									      //
// This file is part of the TypeAccess/C-Super-Script software package.	      //
//									      //
// Copyright (C) 1995 Randall C. O'Reilly, Chadley K. Dawson, 		      //
//		      James L. McClelland, and Carnegie Mellon University     //
//     									      //
// Permission to use, copy, modify, and distribute this software and its      //
// documentation for any purpose is hereby granted without fee, provided that //
// the above copyright notice and this permission notice appear in all copies //
// of the software and related documentation.                                 //
// 									      //
// Note that the PDP++ software package, which contains this package, has a   //
// more restrictive copyright, which applies only to the PDP++-specific       //
// portions of the software, which are labeled as such.			      //
//									      //
// Note that the taString class, which is derived from the GNU String class,  //
// is Copyright (C) 1988 Free Software Foundation, written by Doug Lea, and   //
// is covered by the GNU General Public License, see ta_string.h.             //
// The iv_graphic library and some iv_misc classes were derived from the      //
// InterViews morpher example and other InterViews code, which is             //
// Copyright (C) 1987, 1988, 1989, 1990, 1991 Stanford University             //
// Copyright (C) 1991 Silicon Graphics, Inc.				      //
//									      //
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,         //
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 	      //
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  	      //
// 									      //
// IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE FOR ANY SPECIAL,    //
// INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES  //
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT     //
// ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY,      //
// ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS        //
// SOFTWARE. 								      //
==============================================================================*/

// builtin functions, constants, etc.

#include <css/machine.h>
#include <css/basic_types.h>
#include <css/c_ptr_types.h>
#include <css/css_builtin.h>
#include <css/ta_css.h>
#include <css/css_iv.h>
#include <ta/ta_base.h>
#include <ta/taiv_data.h>
#include <ta/ta_group.h>

#include <unistd.h>
#include <values.h>
#include <stdio.h>
#include <errno.h>
#include <time.h>
#include <sys/time.h>

// when using the .c version
// extern "C" {
// #include <special_math.h>
// }

// using the .cc version
#include <special_math.h>
#include <css_misc_funs.h>

#ifndef CSS_NUMBER
#include "css_parse.h"
#endif

cssElCFun*	cssBI::asgn=NULL;
cssElCFun*	cssBI::asgn_add=NULL;
cssElCFun*	cssBI::asgn_sub=NULL;
cssElCFun*	cssBI::asgn_mult=NULL;
cssElCFun*	cssBI::asgn_div=NULL;
cssElCFun*	cssBI::asgn_mod=NULL;
cssElCFun*	cssBI::asgn_lshift=NULL;
cssElCFun*	cssBI::asgn_rshift=NULL;
cssElCFun*	cssBI::asgn_and=NULL;
cssElCFun*	cssBI::asgn_xor=NULL;
cssElCFun*	cssBI::asgn_or=NULL;
cssElCFun*	cssBI::init_asgn=NULL;
cssElCFun*	cssBI::asgn_post_pp=NULL;
cssElCFun*	cssBI::asgn_pre_pp=NULL;
cssElCFun*	cssBI::asgn_post_mm=NULL;
cssElCFun*	cssBI::asgn_pre_mm=NULL; 
cssElCFun*	cssBI::new_opr=NULL;
cssElCFun*	cssBI::del_opr=NULL;
cssElCFun*	cssBI::constr=NULL;
cssElCFun*	cssBI::add=NULL;
cssElCFun*	cssBI::sub=NULL;
cssElCFun*	cssBI::mul=NULL;
cssElCFun*	cssBI::div=NULL;
cssElCFun*	cssBI::modulo=NULL;
cssElCFun*	cssBI::lshift=NULL;
cssElCFun*	cssBI::rshift=NULL;
cssElCFun*	cssBI::bit_and=NULL;
cssElCFun*	cssBI::bit_xor=NULL;
cssElCFun*	cssBI::bit_or=NULL;
cssElCFun*	cssBI::neg=NULL;
cssElCFun*	cssBI::addr_of=NULL;
cssElCFun*	cssBI::de_ptr=NULL;
cssElCFun*	cssBI::de_array=NULL;
cssElCFun*	cssBI::points_at=NULL;
cssElCFun*	cssBI::member_fun=NULL;
cssElCFun*	cssBI::scoper=NULL;
cssElCFun*	cssBI::pop=NULL;
cssElCFun*	cssBI::cast=NULL;
cssElCFun*	cssBI::cond=NULL;
cssElCFun*	cssBI::ifcd=NULL;
cssElCFun*	cssBI::switch_jump=NULL;
		
cssElCFun*	cssBI::push_root=NULL;	// pushes a root value on stack
cssElCFun*	cssBI::push_next=NULL; // pushes next program item on stack
cssElCFun*	cssBI::push_cur_this=NULL; // push current this pointer on stack
cssElCFun*	cssBI::arg_swap=NULL; // swaps args
cssElCFun*	cssBI::array_alloc=NULL;
cssElCFun*	cssBI::strstream_rewind=NULL;
		
cssInt*		cssBI::true_int=NULL;
cssInt*		cssBI::false_int=NULL;
cssElCFun*	cssBI::gt=NULL;
cssElCFun*	cssBI::lt=NULL;
cssElCFun*	cssBI::eq=NULL;
cssElCFun*	cssBI::ge=NULL;
cssElCFun*	cssBI::le=NULL;	
cssElCFun*	cssBI::ne=NULL;
cssElCFun*	cssBI::and=NULL;
cssElCFun*	cssBI::or=NULL;
cssElCFun*	cssBI::not=NULL;
cssElCFun*	cssBI::power=NULL;

cssTA_Base*	cssBI::root;		// root script element


//////////////////////////////////
// 	Internal Functions	//
//////////////////////////////////

// assign statements
static cssEl* cssElCFun_asgn_stub(int, cssEl* arg[]) {
  *(arg[1]) = *(arg[2]);
  return arg[1];
}
static cssEl* cssElCFun_init_asgn_stub(int, cssEl* arg[]) {
  (arg[2])->InitAssign(*(arg[1]));
  return arg[2];
}
static cssEl* cssElCFun_asgn_add_stub(int, cssEl* arg[]) {
  *(arg[1]) += *(arg[2]);
  return arg[1];
}
static cssEl* cssElCFun_asgn_sub_stub(int, cssEl* arg[]) {
  *(arg[1]) -= *(arg[2]);
  return arg[1];
}
static cssEl* cssElCFun_asgn_mult_stub(int, cssEl* arg[]) {
  *(arg[1]) *= *(arg[2]);
  return arg[1];
}
static cssEl* cssElCFun_asgn_div_stub(int, cssEl* arg[]) {
  *(arg[1]) /= *(arg[2]);
  return arg[1];
}
static cssEl* cssElCFun_asgn_mod_stub(int, cssEl* arg[]) {
  *(arg[1]) %= *(arg[2]);
  return arg[1];
}
static cssEl* cssElCFun_asgn_lshift_stub(int, cssEl* arg[]) {
  *(arg[1]) <<= *(arg[2]);
  return arg[1];
}
static cssEl* cssElCFun_asgn_rshift_stub(int, cssEl* arg[]) {
  *(arg[1]) >>= *(arg[2]);
  return arg[1];
}
static cssEl* cssElCFun_asgn_and_stub(int, cssEl* arg[]) {
  *(arg[1]) &= *(arg[2]);
  return arg[1];
}
static cssEl* cssElCFun_asgn_xor_stub(int, cssEl* arg[]) {
  *(arg[1]) ^= *(arg[2]);
  return arg[1];
}
static cssEl* cssElCFun_asgn_or_stub(int, cssEl* arg[]) {
  *(arg[1]) |= *(arg[2]);
  return arg[1];
}
static cssEl* cssElCFun_asgn_post_pp_stub(int, cssEl* arg[]) {
  cssEl* tmp;
  if(arg[1]->IsRef())
    tmp = arg[1]->GetActualObj()->AnonClone(); // save previous value
  else
    tmp = arg[1]->AnonClone(); // save previous value
  *(arg[1]) += *cssBI::true_int;	// an int value of 1
  return tmp;
}
static cssEl* cssElCFun_asgn_pre_pp_stub(int, cssEl* arg[]) {
  *(arg[1]) += *cssBI::true_int;	// an int value of 1
  return arg[1];
}
static cssEl* cssElCFun_asgn_post_mm_stub(int, cssEl* arg[]) {
  cssEl* tmp;
  if(arg[1]->IsRef())
    tmp = arg[1]->GetActualObj()->AnonClone(); // save previous value
  else
    tmp = arg[1]->AnonClone(); // save previous value
  *(arg[1]) -= *cssBI::true_int;	// an int value of 1
  return tmp;
}
static cssEl* cssElCFun_asgn_pre_mm_stub(int, cssEl* arg[]) {
  *(arg[1]) -= *cssBI::true_int;	// an int value of 1
  return arg[1];
}
// arg 1 is conditional, 2 is true, 3 is false
static cssEl* cssElCFun_cond_stub(int, cssEl* arg[]) {
  if((int)*arg[1])
    return arg[2];
  else
    return arg[3];
}

// if na=2 arg 1 is size, arg 2 is type, else arg 1 is type
static cssEl* cssElCFun_new_opr_stub(int na, cssEl* arg[]) {
  if (na == 2) {
    // allocate an array
    cssArray* tmp = new cssArray((int)(*arg[1]), arg[2]);
    return tmp;
  }
  else return arg[1]->NewOpr();
}
static cssEl* cssElCFun_del_opr_stub(int, cssEl* arg[]) {
  arg[1]->DelOpr();
  return &cssMisc::Void;
}

static cssEl* cssElCFun_constr_stub(int, cssEl* arg[]) {
  if(arg[1]->GetType() == cssEl::T_Class)
    ((cssClassInst*)arg[1])->ConstructToken();
  return &cssMisc::Void;
}

// cast operator
static cssEl* cssElCFun_cast_stub(int, cssEl* arg[]) {
  cssProg* cp = arg[0]->prog;
  (arg[2])->MakeTempToken(cp);		// make the thing
  cssEl* rval = cp->Stack()->Pop(); 	// result is ptr to token
  rval->CastFm(*(arg[1]));
  return rval;
}
  
// push_next

static cssEl* cssElCFun_push_next_stub(int, cssEl* arg[]) {
  cssProg* cp = arg[0]->prog;
  cssInst* nxt = cp->Next();
  if(nxt != NULL)
    return nxt->inst.El();
  else
    return &cssMisc::Void;
}

// pointer operators
static cssEl* cssElCFun_addr_of_stub(int, cssEl* arg[]) {
  return new cssPtr(arg[1]->GetAddr());
}
static cssEl* cssElCFun_de_ptr_stub(int, cssEl* arg[]) {
  return *(*arg[1]);
}
static cssEl* cssElCFun_de_array_stub(int, cssEl* arg[]) {
  return (*arg[1])[*arg[2]];
}
static cssEl* cssElCFun_points_at_stub(int, cssEl* arg[]) {
  if(arg[2]->GetType() == cssEl::T_String)
    return (arg[1])->GetMember((const char*)*arg[2]);
  else
    return (arg[1])->GetMember((int)*arg[2]);
}
static cssEl* cssElCFun_scoper_stub(int, cssEl* arg[]) {
  return (arg[1])->GetScoped(*arg[2]);
}
static cssEl* cssElCFun_member_fun_stub(int, cssEl* arg[]) {
#ifdef HAVE_VOLATILE
  volatile cssProg* cp = arg[0]->prog; // volatile because of setjmp
  volatile cssEl* tmp;
#else
  cssProg* cp = arg[0]->prog;
  cssEl* tmp;
#endif
  if(arg[2]->GetType() == cssEl::T_String)
    tmp = (arg[1])->GetMemberFun((const char*)*arg[2]);
  else 
    tmp = (arg[1])->GetMemberFun((int)*arg[2]);
  if((tmp->GetType() != cssEl::T_MbrCFun) && (tmp->GetType() != cssEl::T_MbrScriptFun)) {
    if(tmp != &cssMisc::Void)
      cssEl::Done((cssEl*)tmp);
    cssMisc::Error(cp, "Member is not a function", tmp->name);
    return &cssMisc::Void;
  }
  int old_state = cp->top->state;
  int old_step = cp->top->step_mode;
  cp->top->step_mode = 0;
  cp->Stack()->Push(&cssMisc::Void); // push the arg stop
  cssEl* ths = arg[1]->GetActualObj(); // get past references and pointers
  if(tmp->GetType() == cssEl::T_MbrScriptFun) {	// push the 'this' pointer 
    if(ths->GetType() == cssEl::T_Class)
      cp->Stack()->Push(ths);
    else {
      if(cp->CurThis() == NULL) {
	cssMisc::Error(cp, "No 'this' pointer for member function call");
	return &cssMisc::Void;
      }
      cp->Stack()->Push(cp->CurThis());
    }
  }
  // push the 'this' pointer for builtin functions
  else if(tmp->GetType() == cssEl::T_MbrCFun) {
    if(ths->GetType() == cssEl::T_Class)
      cp->Stack()->Push(ths);
  }
  cp->top->Cont();		// run rest of args 
  cp->top->state = old_state;
  cp->top->step_mode = old_step;
  
  if(!((cp->top->run==cssEl::Stopping) || (cp->top->run==cssEl::Running)))
    return &cssMisc::Void;
  
  setjmp(cssMisc::begin);		// need to set the break to here for mbr_fun do
  if(cp->top->run != cssEl::ExecError) {
    tmp->Do((cssProg*)cp);
  }
//  arg[0]->prog = cp;		// reset this, see above ^^^^
  cssEl::Done((cssEl*)tmp);	// get rid of the memberfun
  return &cssMisc::Void;	// retv pushed by the previous do
}

// arg[1]..[na-1] = sizes,   arg[na] = array
static cssEl* cssElCFun_array_alloc_stub(int na, cssEl* arg[]) {
  cssInt* rval = new cssInt();
  if(arg[na]->GetType() != cssEl::T_Array &&
     arg[na]->GetType() != cssEl::T_ArrayType) {
    cssMisc::Error(arg[0]->prog, "Array alloc did not get an array");
    rval->val = 0;
    return rval;
  }
  cssArray* tmp = (cssArray*) arg[na];
  rval->val = tmp->AllocAll(na-2, arg+1);
  return rval;
}

static cssEl* cssElCFun_add_stub(int, cssEl* arg[]) {
  return *(arg[1]) + *(arg[2]);
}
static cssEl* cssElCFun_sub_stub(int, cssEl* arg[]) {
  return *(arg[1]) - *(arg[2]);
}
static cssEl* cssElCFun_mul_stub(int, cssEl* arg[]) {
  return *(arg[1]) * *(arg[2]);
}
static cssEl* cssElCFun_div_stub(int, cssEl* arg[]) {
  return *(arg[1]) / *(arg[2]);
}
static cssEl* cssElCFun_modulo_stub(int, cssEl* arg[]) {
  return *(arg[1]) % *(arg[2]);
}
static cssEl* cssElCFun_lshift_stub(int, cssEl* arg[]) {
  if((arg[1]->GetType() == cssEl::T_TA) && (arg[2]->GetType() == cssEl::T_Bool)) {
    cssTA* ta = (cssTA*)arg[1];
    if(ta->type_def->InheritsFrom(TA_ostream)) {
      ostream* strm = (ostream*)ta->GetVoidPtrOfType(&TA_ostream);
      if(strm != NULL) {
	if(arg[2]->name == "flush") {
	  *strm << flush;	  return arg[1];
	}
	if(arg[2]->name == "endl") {
	  *strm << endl;	  return arg[1];
	}
	if(arg[2]->name == "ends") {
	  *strm << ends;	  return arg[1];
	}
	if(arg[2]->name == "dec") {
	  *strm << dec;	  return arg[1];
	}
	if(arg[2]->name == "hex") {
	  *strm << hex;	  return arg[1];
	}
	if(arg[2]->name == "oct") {
	  *strm << oct;	  return arg[1];
	}
      }
    }
  }
  return *(arg[1]) << *(arg[2]);
}
static cssEl* cssElCFun_rshift_stub(int, cssEl* arg[]) {
  if((arg[1]->GetType() == cssEl::T_TA) && (arg[2]->GetType() == cssEl::T_Bool)) {
    cssTA* ta = (cssTA*)arg[1];
    if(ta->type_def->InheritsFrom(TA_istream)) {
      istream* strm = (istream*)ta->GetVoidPtrOfType(&TA_istream);
      if(strm != NULL) {
	if(arg[2]->name == "ws") {
	  *strm >> ws;	  return arg[1];
	}
	if(arg[2]->name == "dec") {
	  *strm >> dec;	  return arg[1];
	}
	if(arg[2]->name == "hex") {
	  *strm >> hex;	  return arg[1];
	}
	if(arg[2]->name == "oct") {
	  *strm >> oct;	  return arg[1];
	}
      }
    }
  }
  return *(arg[1]) >> *(arg[2]);
}
static cssEl* cssElCFun_bit_and_stub(int, cssEl* arg[]) {
  return *(arg[1]) & *(arg[2]);
}
static cssEl* cssElCFun_bit_xor_stub(int, cssEl* arg[]) {
  return *(arg[1]) ^ *(arg[2]);
}
static cssEl* cssElCFun_bit_or_stub(int, cssEl* arg[]) {
  return *(arg[1]) | *(arg[2]);
}
static cssEl* cssElCFun_neg_stub(int, cssEl* arg[]) {
  return -(*(arg[1]));
}

static cssEl* cssElCFun_strstream_rewind_stub(int, cssEl* arg[]) {
  strstream* strm = (strstream*)*(arg[1]);
  strm->clear();
  strm->seekg(0, ios::beg);
  return arg[1];
}

static cssEl* cssElCFun_pop_stub(int, cssEl**) {
  return &cssMisc::Void;		// the args have the stack contents, will be doneArgs
}
static cssEl* cssElCFun_nop_stub(int, cssEl**) {
  return &cssMisc::Void;
}
static cssEl* cssElCFun_arg_swap_stub(int na, cssEl* arg[]) {
  if(na != 2)
    return &cssMisc::Void;
  cssProg* cp = arg[0]->prog;
  cp->Stack()->Push(arg[2]);	// push this on the stack
  return arg[1];		// and this goes after it
}

// inc after savepc & what happens there
// 0	then
// +1	else
// +2	end
// +3	condition

static cssEl* cssElCFun_ifcd_stub(int, cssEl* arg[]) {
  cssProg* cp = arg[0]->prog;
  css_progdx savepc = cp->PC();
  cssEl* cond;

  cp->SetPC(savepc + 3);
  cond = cp->top->Cont();
  if((int) *cond) {
    cp->SetPC(savepc);
    cp->EndRunPop();		// get rid of cond..
    cp->top->Cont();		// then part
    if((cp->top->run == cssEl::Stopping) || (cp->top->run == cssEl::Running))
      cp->insts[savepc + 2]->Do(); // reset pc to the end
  }
  else {
    cp->SetPC(savepc + 1);
    cp->EndRunPop();		// get rid of cond..
    cp->top->Cont();
  }
  cp->EndRunPop();
  if((cp->top->run == cssEl::Stopping) || (cp->top->run == cssEl::Running))
    ((cssElFun*)arg[0])->dostat = cssEl::Running;
  else
    ((cssElFun*)arg[0])->dostat = cp->top->run;
  return &cssMisc::Void;
}

static cssEl* cssElCFun_switch_jump_stub(int, cssEl* arg[]) {
  cssProg* cp = arg[0]->prog;

  cssElPtr ptr = cp->literals.FindName(cssSwitchJump_Name);
  cssArray* val_ary = (cssArray*)ptr.El();

  ptr = cp->FindAutoName(cssSwitchVar_Name);
  cssEl* sval = ptr.El();
  
  if((val_ary == &cssMisc::Void) || (sval == &cssMisc::Void)) {
    cssMisc::Error(cp, "case statement not in a valid switch block");
    ((cssElFun*)arg[0])->dostat = cssEl::Breaking; // bail
    return &cssMisc::Void;
  }

  ptr = val_ary->items->FindName((const char*)*sval);
  cssEl* jmp_to = ptr.El();
  if(jmp_to == &cssMisc::Void) { // not found
    ptr = val_ary->items->FindName(cssSwitchDefault_Name);
    jmp_to = ptr.El();
    if(jmp_to == &cssMisc::Void) {
      ((cssElFun*)arg[0])->dostat = cssEl::Breaking; // bail, no other case found
      return &cssMisc::Void;
    }
  }

  cp->SetPC((int)*jmp_to);	// go to that line
  return &cssMisc::Void;
}


static cssEl* cssElCFun_switch_stub(int, cssEl* arg[]) {
  cssProg* cp = arg[0]->prog;
  cp->top->Cont();		// run the switch
  cp->EndRunPop();		// get rid of junk

  if(cp->top->run == cssEl::Breaking)
    cp->top->run = cssEl::Running;	// clear the break
  return &cssMisc::Void;
}

// while loop: inc after savepc & what happens there
// 0	body
// +1	end
// +2	condition

static cssEl* cssElCFun_while_stub(int, cssEl* arg[]) {
  cssProg* cp = arg[0]->prog;
  css_progdx savepc = cp->PC();
  cssEl* cond;

  cp->SetPC(savepc + 2);
  cond = cp->top->Cont();

  while((int) *cond) {
    cp->SetPC(savepc);
    cp->EndRunPop();		// get rid of cond..
    cp->top->Cont();		// body
    if(cp->top->run == cssEl::Breaking) {
      cp->top->run = cssEl::Stopping;
      break;
    }
    else if(cp->top->run == cssEl::Continuing) {
      cp->top->run = cssEl::Stopping;
    }
    else if(!((cp->top->run==cssEl::Stopping) || (cp->top->run==cssEl::Running)))
      break;
    cp->SetPC(savepc + 2);
    cond = cp->top->Cont();
  }
  cp->EndRunPop();		// get rid of cond..
  if((cp->top->run == cssEl::Stopping) || (cp->top->run == cssEl::Running)) {
    cp->insts[savepc + 1]->Do(); // resets pc
    ((cssElFun*)arg[0])->dostat = cssEl::Running;
  }
  else 
    ((cssElFun*)arg[0])->dostat = cp->top->run;
  return &cssMisc::Void;
}

// do loop: inc after savepc & what happens there
// +0	condition
// +1	end
// +2	body

static cssEl* cssElCFun_do_stub(int, cssEl* arg[]) {
  cssProg* cp = arg[0]->prog;
  css_progdx savepc = cp->PC();
  bool cond;

  do {
    cp->SetPC(savepc+2);
    cp->top->Cont();		// body
    if(cp->top->run == cssEl::Breaking) {
      cp->top->run = cssEl::Stopping;
      break;
    }
    else if(cp->top->run == cssEl::Continuing) {
      cp->top->run = cssEl::Stopping;
    }
    else if(!((cp->top->run==cssEl::Stopping) || (cp->top->run==cssEl::Running)))
      break;
    cp->SetPC(savepc + 0);
    cssEl* cond_el = cp->top->Cont();
    cond = (bool)((int)*cond_el);
    cp->EndRunPop();		// get rid of cond..
  } while(cond);

  if((cp->top->run == cssEl::Stopping) || (cp->top->run == cssEl::Running)) {
    cp->insts[savepc + 1]->Do(); // resets pc
    ((cssElFun*)arg[0])->dostat = cssEl::Running;
  }
  else 
    ((cssElFun*)arg[0])->dostat = cp->top->run;
  return &cssMisc::Void;
}


// inc after savepc & what happens there
// 0	condition
// +1	increments
// +2	body
// +3	end
// +4	inits

static cssEl* cssElCFun_for_stub(int, cssEl* arg[]) {
  cssProg* cp = arg[0]->prog;
  css_progdx savepc = cp->PC();
  cssEl* cond;

  cp->SetPC(savepc + 4);
  cp->top->Cont();		// run initializers
  cp->EndRunPop();

  cp->SetPC(savepc);
  cond = cp->top->Cont();	// condition

  while((int) *cond) {
    cp->SetPC(savepc + 2);
    cp->EndRunPop();		// get rid of cond..
    cp->top->Cont();		// body
    if(cp->top->run == cssEl::Breaking) {
      cp->top->run = cssEl::Stopping;
      break;
    }
    else if(cp->top->run == cssEl::Continuing) {
      cp->top->run = cssEl::Stopping;
    }
    else if(!((cp->top->run==cssEl::Stopping) || (cp->top->run==cssEl::Running)))
      break;

    cp->SetPC(savepc + 1);
    cp->top->Cont();		// increments
    cp->EndRunPop();

    cp->SetPC(savepc);
    cond = cp->top->Cont();
  }
  cp->EndRunPop();		// get rid of cond..
  if((cp->top->run == cssEl::Stopping) || (cp->top->run == cssEl::Running)) {
    cp->insts[savepc + 3]->Do(); // resets pc to end
    ((cssElFun*)arg[0])->dostat = cssEl::Running;
  }
  else 
    ((cssElFun*)arg[0])->dostat = cp->top->run;
  return &cssMisc::Void;
}

static cssEl* cssElCFun_return_stub(int na, cssEl* arg[]) {
  if(na > 0) {
    cssProg* cp = arg[0]->prog;
    cssScriptFun* sf = cp->top->GetCurrentFun();
    if(sf == NULL)
      cssMisc::Error(cp, "Could not find current function scope for return value!");
    else {
      cssEl* rval = sf->argv[0].El();
      if(rval == &cssMisc::Void)
	cssMisc::Error(cp, "Attempt to return a value in a void function");
      rval->InitAssign(*(arg[1])); // use initassign!
    }
  }
  ((cssElFun*)arg[0])->dostat = cssEl::Returning;
  return &cssMisc::Void;
}
static cssEl* cssElCFun_break_stub(int, cssEl* arg[]) {
  ((cssElFun*)arg[0])->dostat = cssEl::Breaking;
  return &cssMisc::Void;
}
static cssEl* cssElCFun_continue_stub(int, cssEl* arg[]) {
  ((cssElFun*)arg[0])->dostat = cssEl::Continuing;
  return &cssMisc::Void;
}

static cssEl* cssElCFun_gt_stub(int, cssEl* arg[]) {
  return new cssInt((int)(*(arg[1]) > *(arg[2])));
}
static cssEl* cssElCFun_lt_stub(int, cssEl* arg[]) {
  return new cssInt((int)(*(arg[1]) < *(arg[2])));
}
static cssEl* cssElCFun_eq_stub(int, cssEl* arg[]) {
  return new cssInt((int)(*(arg[1]) == *(arg[2])));
}
static cssEl* cssElCFun_ge_stub(int, cssEl* arg[]) {
  return new cssInt((int)(*(arg[1]) >= *(arg[2])));
}
static cssEl* cssElCFun_le_stub(int, cssEl* arg[]) {
  return new cssInt((int)(*(arg[1]) <= *(arg[2])));
}
static cssEl* cssElCFun_ne_stub(int, cssEl* arg[]) {
  return new cssInt((int)(*(arg[1]) != *(arg[2])));
}
static cssEl* cssElCFun_and_stub(int, cssEl* arg[]) {
  return new cssInt((int)(*(arg[1]) && *(arg[2])));
}
static cssEl* cssElCFun_or_stub(int, cssEl* arg[]) {
  
  return new cssInt((int)(*(arg[1]) || *(arg[2])));
}
static cssEl* cssElCFun_not_stub(int, cssEl* arg[]) {
  return new cssInt((int)!(*(arg[1])));
}

static cssEl* cssElCFun_push_root_stub(int, cssEl**) {
  return cssBI::root;
}

static cssEl* cssElCFun_push_cur_this_stub(int, cssEl* arg[]) {
  cssProg* cp = arg[0]->prog;
  if(cp->CurThis() == NULL) {
    cssMisc::Error(cp, "No current 'this' pointer");
    return &cssMisc::Void;
  }
  return cp->CurThis();
}

// pre-processor

static cssEl* cssElCFun_include_stub(int, cssEl* arg[]) {
  cssProg* cp = arg[0]->prog;
  cp->top->SetInclude((const char*)*(arg[1]));
  return &cssMisc::Void;
}

//////////////////////////////////
//	Install_Internals	//
//////////////////////////////////

static void Install_Internals() {
  cssElInCFun_inst_ptr(cssMisc::Internal, asgn, 2, CSS_FUN);
  cssElInCFun_inst_ptr(cssMisc::Internal, init_asgn, 2, CSS_FUN);
  cssElInCFun_inst_ptr(cssMisc::Internal, asgn_add, 2, CSS_FUN);
  cssElInCFun_inst_ptr(cssMisc::Internal, asgn_sub, 2, CSS_FUN);
  cssElInCFun_inst_ptr(cssMisc::Internal, asgn_mult, 2, CSS_FUN);
  cssElInCFun_inst_ptr(cssMisc::Internal, asgn_div, 2, CSS_FUN);
  cssElInCFun_inst_ptr(cssMisc::Internal, asgn_mod, 2, CSS_FUN);
  cssElInCFun_inst_ptr(cssMisc::Internal, asgn_lshift, 2, CSS_FUN);
  cssElInCFun_inst_ptr(cssMisc::Internal, asgn_rshift, 2, CSS_FUN);
  cssElInCFun_inst_ptr(cssMisc::Internal, asgn_and, 2, CSS_FUN);
  cssElInCFun_inst_ptr(cssMisc::Internal, asgn_xor, 2, CSS_FUN);
  cssElInCFun_inst_ptr(cssMisc::Internal, asgn_or, 2, CSS_FUN);
  cssElInCFun_inst_ptr(cssMisc::Internal, asgn_post_pp, 1, CSS_FUN);
  cssElInCFun_inst_ptr(cssMisc::Internal, asgn_pre_pp, 1, CSS_FUN);
  cssElInCFun_inst_ptr(cssMisc::Internal, asgn_post_mm, 1, CSS_FUN);
  cssElInCFun_inst_ptr(cssMisc::Internal, asgn_pre_mm, 1, CSS_FUN);
  cssElInCFun_inst_ptr(cssMisc::Internal, cond, 3, CSS_FUN);
  cssElInCFun_inst_ptr(cssMisc::Internal, cast, 2, CSS_FUN);
  cssElInCFun_inst_ptr(cssMisc::Internal, add, 2, CSS_FUN);
  cssElInCFun_inst_ptr(cssMisc::Internal, sub, 2, CSS_FUN);
  cssElInCFun_inst_ptr(cssMisc::Internal, mul, 2, CSS_FUN);
  cssElInCFun_inst_ptr(cssMisc::Internal, div, 2, CSS_FUN);
  cssElInCFun_inst_ptr(cssMisc::Internal, modulo, 2, CSS_FUN);
  cssElInCFun_inst_ptr(cssMisc::Internal, lshift, 2, CSS_FUN);
  cssElInCFun_inst_ptr(cssMisc::Internal, rshift, 2, CSS_FUN);
  cssElInCFun_inst_ptr(cssMisc::Internal, bit_and, 2, CSS_FUN);
  cssElInCFun_inst_ptr(cssMisc::Internal, bit_xor, 2, CSS_FUN);
  cssElInCFun_inst_ptr(cssMisc::Internal, bit_or, 2, CSS_FUN);
  cssElInCFun_inst_ptr(cssMisc::Internal, neg, 1, CSS_FUN);
  cssElInCFun_inst_ptr(cssMisc::Internal, addr_of, 1, CSS_FUN);
  cssElInCFun_inst_ptr(cssMisc::Internal, de_ptr, 1, CSS_FUN);
  cssElInCFun_inst_ptr(cssMisc::Internal, de_array, 2, CSS_FUN);
  cssElInCFun_inst_ptr(cssMisc::Internal, points_at, 2, CSS_FUN);
  cssElInCFun_inst_ptr(cssMisc::Internal, scoper, 2, CSS_FUN);
  cssElInCFun_inst_ptr(cssMisc::Internal, member_fun, 2, CSS_FUN);
  cssElCFun_inst_ptr(cssMisc::Internal, pop, cssEl::VarArg, CSS_FUN);
  cssElCFun_inst_ptr(cssMisc::Internal, array_alloc, cssEl::VarArg, CSS_FUN);
  cssElInCFun_inst_ptr(cssMisc::Internal, push_root, cssEl::NoArg, CSS_FUN);
  cssElInCFun_inst_ptr(cssMisc::Internal, push_next, cssEl::NoArg, CSS_FUN);
  cssElInCFun_inst_ptr(cssMisc::Internal, push_cur_this, cssEl::NoArg, CSS_FUN);
  cssElInCFun_inst_ptr(cssMisc::Internal, arg_swap, 2, CSS_FUN);
  cssElInCFun_inst_ptr(cssMisc::Internal, strstream_rewind, 1, CSS_FUN);
  cssElInCFun_inst_ptr(cssMisc::Internal, switch_jump, cssEl::NoArg, CSS_FUN);

  cssElInCFun_inst_ptr(cssMisc::Internal, gt, 2, CSS_FUN);
  cssElInCFun_inst_ptr(cssMisc::Internal, lt, 2, CSS_FUN);
  cssElInCFun_inst_ptr(cssMisc::Internal, eq, 2, CSS_FUN);
  cssElInCFun_inst_ptr(cssMisc::Internal, ge, 2, CSS_FUN);
  cssElInCFun_inst_ptr(cssMisc::Internal, le, 2, CSS_FUN);
  cssElInCFun_inst_ptr(cssMisc::Internal, ne, 2, CSS_FUN);
  cssElInCFun_inst_ptr(cssMisc::Internal, and, 2, CSS_FUN);
  cssElInCFun_inst_ptr(cssMisc::Internal, or, 2, CSS_FUN);
  cssElInCFun_inst_ptr(cssMisc::Internal, not, 1, CSS_FUN);

  // internal functions
  cssElInCFun_inst	(cssMisc::Functions, for, cssEl::NoArg, CSS_FOR);
  cssElInCFun_inst	(cssMisc::Functions, while, cssEl::NoArg, CSS_WHILE);
  cssElInCFun_inst	(cssMisc::Functions, do, cssEl::NoArg, CSS_DO);
  cssElInCFun_inst_ptr_nm (cssMisc::Functions, ifcd, cssEl::NoArg, "if", CSS_IF);
  cssElInCFun_inst_nm	(cssMisc::Functions, ifcd, cssEl::NoArg, "else", CSS_ELSE);
  cssElInCFun_inst	(cssMisc::Functions, switch, cssEl::NoArg, CSS_SWITCH);
  cssElCFun_inst	(cssMisc::Functions, return, cssEl::VarArg, CSS_RETURN);
  cssElInCFun_inst	(cssMisc::Functions, break, cssEl::NoArg, CSS_BREAK);
  cssElInCFun_inst	(cssMisc::Functions, continue, cssEl::NoArg, CSS_CONTINUE);
  cssElCFun_inst_nm     (cssMisc::Functions, new_opr, cssEl::VarArg, "new", CSS_NEW);
  cssElCFun_inst_nm     (cssMisc::Functions, del_opr, 1, "delete", CSS_DELETE);
  cssElInCFun_inst_ptr	(cssMisc::Internal, constr, 1, CSS_FUN);

  // stuff just for parsing
  cssElCFun_inst_nm	(cssMisc::Parse, nop, 	cssEl::NoArg, "class", CSS_CLASS);
  cssElCFun_inst_nm	(cssMisc::Parse, nop, 	cssEl::NoArg, "struct", CSS_CLASS);
  cssElCFun_inst_nm	(cssMisc::Parse, nop, 	cssEl::NoArg, "union", CSS_CLASS);
  cssElCFun_inst_nm	(cssMisc::Parse, nop, 	cssEl::NoArg, "enum", CSS_ENUM);
  cssElCFun_inst_nm	(cssMisc::Parse, nop, 	cssEl::NoArg, "extern", CSS_EXTERN);
  cssElCFun_inst_nm	(cssMisc::Parse, nop, 	cssEl::NoArg, "static", CSS_STATIC);
  cssElCFun_inst_nm	(cssMisc::Parse, nop, 	cssEl::NoArg, "private", CSS_PRIVATE);
  cssElCFun_inst_nm	(cssMisc::Parse, nop, 	cssEl::NoArg, "public", CSS_PUBLIC);
  cssElCFun_inst_nm	(cssMisc::Parse, nop, 	cssEl::NoArg, "protected", CSS_PROTECTED);
  cssElCFun_inst_nm	(cssMisc::Parse, nop, 	cssEl::NoArg, "virtual", CSS_VIRTUAL);
  cssElCFun_inst_nm	(cssMisc::Parse, nop, 	cssEl::NoArg, "const", CSS_CONST);
  cssElCFun_inst_nm	(cssMisc::Parse, nop, 	cssEl::NoArg, "inline", CSS_INLINE);
  cssElCFun_inst_nm	(cssMisc::Parse, nop, 	cssEl::NoArg, "case", CSS_CASE);
  cssElCFun_inst_nm	(cssMisc::Parse, nop, 	cssEl::NoArg, "default", CSS_DEFAULT);

  // pre-processor
  cssElInCFun_inst(cssMisc::PreProcessor, include,  	1, CSS_PP_INCLUDE);
  cssElCFun_inst_nm(cssMisc::PreProcessor, nop,  	cssEl::NoArg, "define", CSS_PP_DEFINE);
  cssElCFun_inst_nm(cssMisc::PreProcessor, nop,  	cssEl::NoArg, "undef", CSS_PP_UNDEF);
  cssElCFun_inst_nm(cssMisc::PreProcessor, nop,  	cssEl::NoArg, "ifdef", CSS_PP_IFDEF);
  cssElCFun_inst_nm(cssMisc::PreProcessor, nop,  	cssEl::NoArg, "ifndef", CSS_PP_IFNDEF);
  cssElCFun_inst_nm(cssMisc::PreProcessor, nop,  	cssEl::NoArg, "else", CSS_PP_ELSE);
  cssElCFun_inst_nm(cssMisc::PreProcessor, nop,  	cssEl::NoArg, "endif", CSS_PP_ENDIF);

  cssMisc::Defines.Push(new cssDef(0, "__CSS__"));
}


//////////////////////////
// 	Commands	//
//////////////////////////

static cssEl* cssElCFun_alias_stub(int, cssEl* arg[]) {
  cssElCFun* tmp = (cssElCFun *)((cssRef*)arg[2])->ptr.El();
  cssMisc::Commands.Push(new cssElCFun(tmp->argc, tmp->funp, (const char*)*(arg[1]),
				 tmp->parse));  
  return &cssMisc::Void;
}
static cssEl* cssElCFun_chsh_stub(int na, cssEl* arg[]) {
  cssProg* cp = arg[0]->prog;
  if(na > 0) {
    if(arg[1]->GetType() == cssEl::T_SubShell) {
      cp->top->SetShell(&(((cssSubShell*)arg[1])->prog_space));
      return &cssMisc::Void;
    }
    if(arg[1]->GetType() != cssEl::T_TA) {
      cssMisc::Error(cp, "chsh argument is not a pointer to a cssProgSpace");
      return &cssMisc::Void;
    }
    cssTA* ta = (cssTA*)arg[1];
    if(ta->type_def->InheritsFrom("cssProgSpace") && (ta->ptr != NULL)) {
      cssProgSpace* ps = (cssProgSpace*)ta->GetVoidPtr();
      cp->top->SetShell(ps);
    }
    else {
      cssMisc::Error(cp, "chsh argument is not a valid pointer to a cssProgSpace");
    }
  }
  else
    cp->top->SetShell(cp->top);	// shell itself
  return &cssMisc::Void;
}
static cssEl* cssElCFun_clearall_stub(int, cssEl* arg[]) {
  cssProg* cp = arg[0]->prog;
  cp->top->SetClearAll();
  return &cssMisc::Void;
}
static cssEl* cssElCFun_cont_stub(int na, cssEl* arg[]) {
  cssProg* cp = arg[0]->prog;
  if(na > 0)
    cp->top->SetCont((int)*(arg[1]));
  else 
    cp->top->SetCont(-1);
  return &cssMisc::Void;
}
static cssEl* cssElCFun_goto_stub(int, cssEl* arg[]) {
  cssProg* cp = arg[0]->prog;
  cp->top->SetCont((int)*(arg[1]));
  return &cssMisc::Void;
}
static cssEl* cssElCFun_commands_stub(int, cssEl* arg[]) {
  cssProg* cp = arg[0]->prog;
  cssMisc::Commands.NameList(*(cp->top->fout));
  return &cssMisc::Void;
}
static cssEl* cssElCFun_constants_stub(int, cssEl* arg[]) {
  cssProg* cp = arg[0]->prog;
  cssMisc::Constants.List(*(cp->top->fout));
  return &cssMisc::Void;
}
static cssEl* cssElCFun_debug_stub(int, cssEl* arg[]) {
  cssProg* cp = arg[0]->prog;
  cp->top->SetDebug((int)*(arg[1])); 
  return &cssMisc::Void;
}
static cssEl* cssElCFun_define_stub(int, cssEl* arg[]) {
  cssProg* cp = arg[0]->prog;
  cp->top->SetDefn();
  return &cssMisc::Void;
}
static cssEl* cssElCFun_defines_stub(int, cssEl* arg[]) {
  cssProg* cp = arg[0]->prog;
  cssMisc::Defines.List(*(cp->top->fout));
  return &cssMisc::Void;
}
static cssEl* cssElCFun_edit_stub(int na, cssEl* arg[]) {
  cssInt* rval = new cssInt();
  if(na > 1)
    rval->val = arg[1]->Edit((int)*(arg[2]));
  else if(na == 1)
    rval->val = arg[1]->Edit();
  else
    cssMisc::Error(arg[0]->prog, "Edit requires an argument of the object to be edited");
  return rval;
}
static cssEl* cssElCFun_enums_stub(int, cssEl* arg[]) {
  cssProg* cp = arg[0]->prog;
  cssMisc::Enums.NameList(*(cp->top->fout));
  return &cssMisc::Void;
}
static cssEl* cssElCFun_frame_stub(int na, cssEl* arg[]) {
  cssProg* cp = arg[0]->prog;
  if(na > 0) {
    int back = (int)*arg[1];
    if(back >= cp->top->size) {
      cssMisc::Error(cp, "Not that many frames:",String(back),"in program, has only:",
		      String((int)(cp->top->size-1)));
      return &cssMisc::Void;
    }
    if(back == cp->top->size-1)
      cp->top->statics.List(*(cp->top->fout));
    else
      cp->top->Prog(cp->top->size-1-back)->ListSpace();
  }
  else {
    cp->ListSpace();
    cp->top->statics.List(*(cp->top->fout));
  }
  return &cssMisc::Void;
}
static cssEl* cssElCFun_functions_stub(int, cssEl* arg[]) {
  cssProg* cp = arg[0]->prog;
  cssMisc::Functions.NameList(*(cp->top->fout));
  cp->top->hard_funs.NameList(*(cp->top->fout));
  cssMisc::HardFuns.NameList(*(cp->top->fout));
  return &cssMisc::Void;
}
static cssEl* cssElCFun_globals_stub(int, cssEl* arg[]) {
  cssProg* cp = arg[0]->prog;
  cp->top->hard_vars.List(*(cp->top->fout));
  cssMisc::HardVars.List(*(cp->top->fout));
  cssMisc::Externs.List(*(cp->top->fout));
  return &cssMisc::Void;
}
static cssEl* cssElCFun_help_stub(int, cssEl* arg[]) {
  cssProg* cp = arg[0]->prog;
  cp->top->Help();
  return &cssMisc::Void;
}
static cssEl* cssElCFun_inherit_stub(int, cssEl* arg[]) {
  cssProg* cp = arg[0]->prog;
  arg[1]->InheritInfo(*(cp->top->fout));
  *(cp->top->fout) << "\n";
  cp->top->fout->flush();
  return &cssMisc::Void;
}
static cssEl* cssElCFun_list_stub(int na, cssEl* arg[]) {
  cssProg* cp = arg[0]->prog;
  if(na > 0) {
    if((arg[1]->GetType() == cssEl::T_ScriptFun) ||
       (arg[1]->GetType() == cssEl::T_MbrScriptFun))
    {
      cssScriptFun* fe = (cssScriptFun*)arg[1];
      if(na > 1)
	fe->fun->List(cout, fe->fun->CurSrcLn(0), *arg[2]);
      else
	fe->fun->List(cout, fe->fun->CurSrcLn(0), 100);	// give it the whole thing..
    }
    else {
      if(na > 1)
	cp->top->List(*arg[1], *arg[2]);
      else
	cp->top->List(*arg[1]);
    }
  }
  else
    cp->top->List();
  return &cssMisc::Void;
}
static cssEl* cssElCFun_load_stub(int, cssEl* arg[]) {
  cssProg* cp = arg[0]->prog;
  cp->top->SetCompile((const char*)*(arg[1]));
  return &cssMisc::Void;
}

static cssEl* cssElCFun_mallinfo_stub(int, cssEl* arg[]) {
  cssProg* cp = arg[0]->prog;
  taMisc::MallocInfo(*(cp->top->fout));
  return &cssMisc::Void;
}
static cssEl* cssElCFun_print_stub(int, cssEl* arg[]) {
  cssProg* cp = arg[0]->prog;
  arg[1]->Print(*(cp->top->fout));
  if(cp->top->debug > 1)
    *(cp->top->fout) << "\trefn: " << arg[1]->refn;
  *(cp->top->fout) << "\n";
  cp->top->fout->flush();
  return &cssMisc::Void;
}
static cssEl* cssElCFun_printr_stub(int, cssEl* arg[]) {
  cssProg* cp = arg[0]->prog;
  arg[1]->PrintR(*(cp->top->fout));
  *(cp->top->fout) << "\n";
  cp->top->fout->flush();
  return &cssMisc::Void;
}

static cssEl* cssElCFun_remove_stub(int, cssEl* arg[]) {
  cssProg* cp = arg[0]->prog;
  if(!cssMisc::TypesSpace.Replace(arg[1], &cssMisc::Void)) {
    if(!cp->top->DelVar(arg[1]))
      cssMisc::Error(cp, "Could not delete type/variable:", (char*)arg[1]->GetName());
  }
  return &cssMisc::Void;
}
static cssEl* cssElCFun_reload_stub(int, cssEl* arg[]) {
  cssProg* cp = arg[0]->prog;
  cp->top->SetReCompile(); 
  return &cssMisc::Void;
}
static cssEl* cssElCFun_reset_stub(int, cssEl* arg[]) {
  cssProg* cp = arg[0]->prog;
  cp->top->SetReset();
  return &cssMisc::Void;
}
static cssEl* cssElCFun_restart_stub(int, cssEl* arg[]) {
  cssProg* cp = arg[0]->prog;
  cp->top->SetRestart();
  return &cssMisc::Void;
}
static cssEl* cssElCFun_run_stub(int, cssEl* arg[]) {
  cssProg* cp = arg[0]->prog;
  cp->top->SetRun();
  return &cssMisc::Void;
}
static cssEl* cssElCFun_setbp_stub(int, cssEl* arg[]) {
  cssProg* cp = arg[0]->prog;
  cp->top->SetBreak((int)*(arg[1])); 
  return &cssMisc::Void;
}
static cssEl* cssElCFun_setout_stub(int, cssEl* arg[]) {
  cssProg* cp = arg[0]->prog;
  ostream* strm = (ostream*)*(arg[1]);
  cp->top->fout = strm; 
  return &cssMisc::Void;
}
static cssEl* cssElCFun_settings_stub(int, cssEl* arg[]) {
  cssProg* cp = arg[0]->prog;
  int i;
  *(cp->top->fout) << "Include Paths:\n";
  taMisc::include_paths.List(*(cp->top->fout));
  *(cp->top->fout) << "\n";
  for(i=1; i<cssMisc::Settings.size; i++) {
    cssMisc::Settings.FastEl(i)->Print(*(cp->top->fout));
    *(cp->top->fout) << "\n";
  }
  cp->top->fout->flush();
  return &cssMisc::Void;
}
static cssEl* cssElCFun_showbp_stub(int, cssEl* arg[]) {
  cssProg* cp = arg[0]->prog;
  cp->top->ShowBreaks();
  return &cssMisc::Void;
}
static cssEl* cssElCFun_source_stub(int, cssEl* arg[]) {
  cssProg* cp = arg[0]->prog;
  cp->top->SetSource((const char*)*(arg[1]));
  return &cssMisc::Void;
}
static cssEl* cssElCFun_stack_stub(int, cssEl* arg[]) {
  cssProg* cp = arg[0]->prog;
  cp->Stack()->List();
  cp->saved_stack.List();
  return &cssMisc::Void;
}
static cssEl* cssElCFun_status_stub(int, cssEl* arg[]) {
  cssProg* cp = arg[0]->prog;
  cp->top->Status();
  return &cssMisc::Void;
}
static cssEl* cssElCFun_step_stub(int, cssEl* arg[]) {
  cssProg* cp = arg[0]->prog;
  cp->top->step_mode = (int)*(arg[1]);
  return &cssMisc::Void;
}

static cssEl* cssElCFun_system_stub(int, cssEl* arg[]) {
  system((const char*)*arg[1]);
  return &cssMisc::Void;
}

static cssEl* cssElCFun_tokens_stub(int, cssEl* arg[]) {
  cssProg* cp = arg[0]->prog;
  arg[1]->TokenInfo(*(cp->top->fout));
  *(cp->top->fout) << "\n";
  cp->top->fout->flush();
  return &cssMisc::Void;
}
static cssEl* cssElCFun_trace_stub(int na, cssEl* arg[]) {
  cssProg* cp = arg[0]->prog;
  if(na > 0) {
    int tr_lev = (int)*arg[1];
    if(tr_lev > 2)
      cp->top->ListSpace();
    else
      cp->top->Trace(tr_lev);
  }
  else
    cp->top->Trace();
  return &cssMisc::Void;
}
static cssEl* cssElCFun_type_stub(int na, cssEl* arg[]) {
  cssProg* cp = arg[0]->prog;
  if(na > 0) {
    arg[1]->TypeInfo(*(cp->top->fout));
    *(cp->top->fout) << "\n";
    cp->top->fout->flush();
  }
  else
    cssMisc::TypesSpace.NameList(*(cp->top->fout));
  return &cssMisc::Void;
}
static cssEl* cssElCFun_unsetbp_stub(int, cssEl* arg[]) {
  cssProg* cp = arg[0]->prog;
  cp->top->unSetBreak((int)*(arg[1]));
  return &cssMisc::Void;
}
static cssEl* cssElCFun_undo_stub(int na, cssEl* arg[]) {
  cssProg* cp = arg[0]->prog;
  if(na > 0)
    cp->top->SetUndo((Int)*(arg[1]));
  else
    cp->top->SetUndo(-1);
  return &cssMisc::Void;
}



//////////////////////////////////
//	Install_Commands	//
//////////////////////////////////

static void Install_Commands() {
  cssElCFun_inst(cssMisc::Commands, alias, 		2, CSS_ALIAS);
  cssElCFun_inst(cssMisc::Commands, chsh, 		cssEl::VarArg, CSS_COMMAND);
  cssElCFun_inst(cssMisc::Commands, clearall, 		0, CSS_COMMAND);
  cssElCFun_inst(cssMisc::Commands, commands,		0, CSS_COMMAND);
  cssElCFun_inst(cssMisc::Commands, constants,		0, CSS_COMMAND);
  cssElCFun_inst(cssMisc::Commands, cont, 		0, CSS_CONT);
  cssElCFun_inst(cssMisc::Commands, goto, 		1, CSS_COMMAND);
  cssElCFun_inst(cssMisc::Commands, define, 		0, CSS_COMMAND);
  cssElCFun_inst(cssMisc::Commands, defines, 		0, CSS_COMMAND);
  cssElCFun_inst(cssMisc::Commands, debug,   		1, CSS_COMMAND);
  cssElCFun_inst(cssMisc::Commands, edit,   		cssEl::VarArg, CSS_COMMAND);
  cssElCFun_inst(cssMisc::Commands, enums,		0, CSS_COMMAND);
  cssElCFun_inst(cssMisc::Commands, frame,		cssEl::VarArg, CSS_COMMAND);
  cssElCFun_inst(cssMisc::Commands, functions,		0, CSS_COMMAND);
  cssElCFun_inst(cssMisc::Commands, globals,		0, CSS_COMMAND);
  cssElCFun_inst(cssMisc::Commands, help, 		0, CSS_COMMAND);
  cssElCFun_inst(cssMisc::Commands, inherit, 		1, CSS_TYPECMD);
  cssElCFun_inst(cssMisc::Commands, list, 		cssEl::VarArg, CSS_LIST);
  cssElCFun_inst(cssMisc::Commands, load, 		1, CSS_COMMAND);
  cssElCFun_inst(cssMisc::Commands, mallinfo, 		0, CSS_COMMAND);
  cssElCFun_inst(cssMisc::Commands, print, 		1, CSS_COMMAND);
  cssElCFun_inst(cssMisc::Commands, printr,		1, CSS_COMMAND);
  cssElCFun_inst(cssMisc::Commands, reload, 		0, CSS_COMMAND);
  cssElCFun_inst(cssMisc::Commands, remove,		1, CSS_REMOVE);
  cssElCFun_inst(cssMisc::Commands, reset, 		0, CSS_COMMAND);
  cssElCFun_inst(cssMisc::Commands, restart,   		0, CSS_COMMAND);
  cssElCFun_inst(cssMisc::Commands, run,  		0, CSS_COMMAND);
  cssElCFun_inst(cssMisc::Commands, setbp, 		1, CSS_COMMAND);
  cssElCFun_inst(cssMisc::Commands, setout, 		1, CSS_COMMAND);
  cssElCFun_inst(cssMisc::Commands, settings,		0, CSS_COMMAND);
  cssElCFun_inst_nm(cssMisc::Commands, system, 		1, "shell", CSS_COMMAND);
  cssElCFun_inst(cssMisc::Commands, showbp,  		0, CSS_COMMAND);
  cssElCFun_inst(cssMisc::Commands, stack, 		0, CSS_COMMAND);
  cssElCFun_inst(cssMisc::Commands, status, 		0, CSS_STATUS);
  cssElCFun_inst(cssMisc::Commands, step, 		1, CSS_COMMAND);
  cssElCFun_inst(cssMisc::Commands, source, 		1, CSS_COMMAND);
  cssElCFun_inst(cssMisc::Commands, tokens, 		1, CSS_TYPECMD);
  cssElCFun_inst(cssMisc::Commands, trace, 		cssEl::VarArg, CSS_COMMAND);
  cssElCFun_inst(cssMisc::Commands, type, 		cssEl::VarArg, CSS_TYPECMD);
  cssElCFun_inst(cssMisc::Commands, unsetbp, 		1, CSS_COMMAND);
  cssElCFun_inst(cssMisc::Commands, undo, 		cssEl::VarArg, CSS_COMMAND);
  cssElCFun_inst_nm(cssMisc::Commands, nop, 		0, "exit", CSS_EXIT);
}


//////////////////////////
//    Math Functions	//
//////////////////////////

static cssEl* cssElCFun_max_stub(int, cssEl* arg[]) {
  return (*(arg[1]) > *(arg[2])) ? arg[1] : arg[2];
}
static cssEl* cssElCFun_min_stub(int, cssEl* arg[]) {
  return (*(arg[1]) < *(arg[2])) ? arg[1] : arg[2];
}

static cssRealFun_stub1(acos);
static cssRealFun_stub1(asin);
static cssRealFun_stub1(atan);
static cssRealFun_stub2(atan2);
static cssRealFun_stub1(ceil);
static cssRealFun_stub1(cos);
static cssRealFun_stub1(cosh);
static cssRealFun_stub0(drand48);
static cssEl* cssElCFun_lrand48_stub(int, cssEl**) {
  return new cssInt((int)lrand48());
}
static cssRealFun_stub1(exp);
static cssRealFun_stub1(fabs);
static cssRealFun_stub1(floor);
static cssRealFun_stub2(fmod);
static cssRealFun_stub1(log);
static cssRealFun_stub1(log10);
static cssEl* cssRealFun_pow_stub(int, cssEl* arg[]) {
  return new cssReal((Real)pow((Real)*(arg[1]), (Real)*(arg[2])));
}
static cssEl* cssElCFun_random_stub(int, cssEl**) {
  return new cssInt((int)lrand48());
}
static cssEl* cssElCFun_srand48_stub(int, cssEl* arg[]) {
  srand48((int)*(arg[1]));
  return &cssMisc::Void;
}

static cssRealFun_stub1(sin);
static cssRealFun_stub1(sinh);
static cssRealFun_stub1(sqrt);
static cssRealFun_stub1(tan);
static cssRealFun_stub1(tanh);

static cssRealFun_stub1(acosh);
static cssRealFun_stub1(asinh);
static cssRealFun_stub1(atanh);


// stuff from special math:

static cssRealFun_stub1(fact_ln);
static cssRealFun_stub2(bico_ln);

static cssRealFun_stub4(hyperg);

static cssRealFun_stub1(gamma_ln);
static cssRealFun_stub2(gamma_p);
static cssRealFun_stub2(gamma_q);
static cssRealFun_stub2(beta);
static cssRealFun_stub3(beta_i);

static cssRealFun_stub3(binom_den);
static cssRealFun_stub3(binom_cum);
static cssRealFun_stub2(binom_dev); 

static cssRealFun_stub2(poisson_den);
static cssRealFun_stub2(poisson_cum);
static cssRealFun_stub1(poisson_dev);

static cssRealFun_stub3(gamma_den);
static cssRealFun_stub3(gamma_cum);
static cssRealFun_stub1(gamma_dev);

static cssRealFun_stub1(gauss_den);
static cssRealFun_stub1(gauss_cum);
static cssRealFun_stub0(gauss_dev);
static cssRealFun_stub1(gauss_inv);
static cssRealFun_stub1(erf);
static cssRealFun_stub1(erf_c);

static cssRealFun_stub2(chisq_p);
static cssRealFun_stub2(chisq_q);
static cssRealFun_stub2(students_cum);
static cssRealFun_stub2(students_den);
static cssRealFun_stub3(Ftest_q);


//////////////////////////////////
//	Install_Math		//
//////////////////////////////////

static void Install_Math() {
  // global real functions
  cssMisc::Functions.Push(cssBI::power = new cssElCFun(2, cssRealFun_pow_stub, "pow", CSS_FUN));

  // parsed internal real functions
  cssRealFun_inst(cssMisc::Functions, acos, 1);
  cssRealFun_inst(cssMisc::Functions, asin, 1);
  cssRealFun_inst(cssMisc::Functions, atan, 1);
  cssRealFun_inst(cssMisc::Functions, atan2, 2);
  cssRealFun_inst(cssMisc::Functions, ceil, 1);
  cssRealFun_inst(cssMisc::Functions, cos, 1);
  cssRealFun_inst(cssMisc::Functions, cosh, 1);
  cssRealFun_inst(cssMisc::Functions, drand48, 0);
  cssElCFun_inst(cssMisc::Functions, lrand48, 0, CSS_FUN);
  cssRealFun_inst(cssMisc::Functions, exp, 1);
  cssRealFun_inst(cssMisc::Functions, fabs, 1);
  cssRealFun_inst(cssMisc::Functions, floor, 1);
  cssRealFun_inst(cssMisc::Functions, fmod, 2);
  cssRealFun_inst(cssMisc::Functions, log, 1);
  cssRealFun_inst(cssMisc::Functions, log10, 1);
  cssElCFun_inst(cssMisc::Functions, max, 2, CSS_FUN);
  cssElCFun_inst(cssMisc::Functions, min, 2, CSS_FUN);
  cssElCFun_inst_nm(cssMisc::Functions, max, 2, "MAX", CSS_FUN);
  cssElCFun_inst_nm(cssMisc::Functions, min, 2, "MIN", CSS_FUN);
  cssElCFun_inst(cssMisc::Functions, random, 0, CSS_FUN);
  cssElCFun_inst(cssMisc::Functions, srand48, 1, CSS_FUN);
  cssRealFun_inst(cssMisc::Functions, sin, 1);
  cssRealFun_inst(cssMisc::Functions, sinh, 1);
  cssRealFun_inst(cssMisc::Functions, sqrt, 1);
  cssRealFun_inst(cssMisc::Functions, tan, 1);
  cssRealFun_inst(cssMisc::Functions, tanh, 1);

  cssRealFun_inst(cssMisc::Functions, fact_ln   , 1);
  cssRealFun_inst(cssMisc::Functions, bico_ln   , 2);

  cssRealFun_inst(cssMisc::Functions, hyperg    , 4);

  cssRealFun_inst(cssMisc::Functions, gamma_ln  , 1);
  cssRealFun_inst(cssMisc::Functions, gamma_p   , 2);
  cssRealFun_inst(cssMisc::Functions, gamma_q   , 2);
  cssRealFun_inst(cssMisc::Functions, beta      , 2);
  cssRealFun_inst(cssMisc::Functions, beta_i    , 3);

  cssRealFun_inst(cssMisc::Functions, binom_den , 3);
  cssRealFun_inst(cssMisc::Functions, binom_cum , 3);
  cssRealFun_inst(cssMisc::Functions, binom_dev , 2);

  cssRealFun_inst(cssMisc::Functions, poisson_den, 2);
  cssRealFun_inst(cssMisc::Functions, poisson_cum,2);
  cssRealFun_inst(cssMisc::Functions, poisson_dev,1);

  cssRealFun_inst(cssMisc::Functions, gamma_den , 3);
  cssRealFun_inst(cssMisc::Functions, gamma_cum , 3);
  cssRealFun_inst(cssMisc::Functions, gamma_dev , 1);

  cssRealFun_inst(cssMisc::Functions, gauss_den , 1);
  cssRealFun_inst(cssMisc::Functions, gauss_cum , 1);
  cssRealFun_inst(cssMisc::Functions, gauss_inv , 1);
  cssRealFun_inst(cssMisc::Functions, gauss_dev , 0);
  cssRealFun_inst(cssMisc::Functions, erf       , 1);
  cssRealFun_inst(cssMisc::Functions, erf_c     , 1);

  cssRealFun_inst(cssMisc::Functions, chisq_p   , 2);
  cssRealFun_inst(cssMisc::Functions, chisq_q   , 2);
  cssRealFun_inst(cssMisc::Functions, students_cum, 2);
  cssRealFun_inst(cssMisc::Functions, students_den, 2);
  cssRealFun_inst(cssMisc::Functions, Ftest_q   , 3);

// hp does not include these unless you have _INCLUDE_HPUX_SOURCE...
#ifndef HP800
  cssRealFun_inst(cssMisc::Functions, acosh, 1);
  cssRealFun_inst(cssMisc::Functions, asinh, 1);
  cssRealFun_inst(cssMisc::Functions, atanh, 1);
#endif // HP800
  
  cssReal_inst_nm(cssMisc::Constants, 3.14159265358979323846,  "PI"   );
  cssReal_inst(cssMisc::Constants, 2.71828182845904523536,  E     );
  cssReal_inst(cssMisc::Constants, 0.57721566490153286060,  GAMMA ); // Euler
  cssReal_inst(cssMisc::Constants, 57.29577951308232087680, DEG   ); // deg/radian
  cssReal_inst(cssMisc::Constants, 1.61803398874989484820,  PHI   ); // golden ratio
  cssReal_inst(cssMisc::Constants, 96484.56,                FARADAY); // coulombs/mole
  cssReal_inst(cssMisc::Constants, 8.31441,                 R     );  // molar gas constant, joules/mole/deg-K

  cssReal_inst_nm(cssMisc::Constants, MAXFLOAT,             "MAXFLOAT");
}


//////////////////////////////////
// 	Misc Functions  	//
//////////////////////////////////

static cssEl* cssElCFun_CancelEditObj_stub(int, cssEl* arg[]) {
  cssEl* aobj = arg[1]->GetActualObj();
  if(aobj->GetType() != cssEl::T_Class)
    cssMisc::Error(arg[0]->prog, "CancelEditObj: arg not a class instance");
  else {
    cssClassInst* obj = (cssClassInst*)aobj;
    cssivSession::CancelObjEdits(obj);
  }
  return &cssMisc::Void;
}

static cssEl* cssElCFun_Extern_stub(int, cssEl* arg[]) {
  cssElPtr ptr = cssMisc::Externs.FindName((const char*)*arg[1]);
  return ptr.El();
}

static cssEl* cssElCFun_Token_stub(int, cssEl* arg[]) {
  return arg[1]->GetToken((int)*(arg[2]));
}

static cssEl* cssElCFun_Type_stub(int, cssEl* arg[]) {
  TypeDef* td = taMisc::types.FindName((const char*)*arg[1]);
  if(td == NULL) {
    cssMisc::Error(arg[0]->prog, "Could not find type:", (const char*)*arg[1]);
    return &cssMisc::Void;
  }
  return new cssTA(td, 1, &TA_TypeDef);
}

static cssEl* cssElCFun_Dir_stub(int na, cssEl* arg[]) {
  String dir_nm = ".";
  if(na > 0)
    dir_nm = (const char*)*(arg[1]);
  String_Array& rval = Dir((const char*)dir_nm);
  static cssTA* rv_ta = NULL;
  if(rv_ta == NULL) {
    rv_ta = new cssTA(NULL, 1, &TA_String_Array);
    cssEl::Ref(rv_ta);
  }
  rv_ta->ptr = (void*)&rval;
  return new cssRef(rv_ta);
}

// this reads a line from a file and returns a string array with
// one entry per column from the file
static cssEl* cssElCFun_ReadLine_stub(int, cssEl* arg[]) {
  istream* fh = (istream*)*(arg[1]);
  if(fh == NULL)
    return &cssMisc::Void;
  String_Array& rval = ReadLine(*fh);
  static cssTA* rv_ta = NULL;
  if(rv_ta == NULL) {
    rv_ta = new cssTA(NULL, 1, &TA_String_Array);
    cssEl::Ref(rv_ta);
  }
  rv_ta->ptr = (void*)&rval;
  return new cssRef(rv_ta);
}

// old-fashioned C file io

static cssEl* cssElCFun_fopen_stub(int, cssEl* arg[]) {
  fstream* strm = new fstream;	// this guy never gets free'd..
  String optype = arg[2]->GetStr();

  if(optype == "r")
    strm->open(*arg[1], ios::in);
  else if(optype == "w")
    strm->open(*arg[1], ios::out);
  else if(optype == "a")
    strm->open(*arg[1], ios::app);
  else {
    delete strm;
    cssMisc::Error(arg[0]->prog, "fopen: open type not recognized: ", (char*)optype);
    return &cssMisc::Void;
  }
  return new cssTA((void*)strm, 1, &TA_fstream);
}
static cssEl* cssElCFun_fclose_stub(int, cssEl* arg[]) {
  fstream* fh = (fstream*)*(arg[1]);
  if((fh != NULL) && fh->rdbuf()->is_open()) {
    fh->close();
  }
  return &cssMisc::Void;
}
static cssEl* cssElCFun_fprintf_stub(int na, cssEl* arg[]) {
  ostream* fh = (ostream*)*(arg[1]);
  if((fh == NULL) || !fh->good())
    return &cssMisc::Void;
  int i;
  for(i=2; i <= na; i++) {
    (arg[i])->PrintF(*fh);
  }
  fh->flush();
  return &cssMisc::Void;
}
static cssEl* cssElCFun_printf_stub(int na, cssEl* arg[]) {
  cssProg* cp = arg[0]->prog;
  int i;
  for(i=1; i <= na; i++) {
    (arg[i])->PrintF(*(cp->top->fout));
  }
  cp->top->fout->flush();
  return &cssMisc::Void;
}

//////////////////////////////////
// 	POSIX Functions  	//
//////////////////////////////////

static cssEl* cssElCFun_access_stub(int, cssEl* arg[]) {
  cssInt* rval = new cssInt();
  rval->val = access((const char*)*arg[1], (int)*arg[2]);
  return rval;
}
static cssEl* cssElCFun_alarm_stub(int, cssEl* arg[]) {
  cssInt* rval = new cssInt();
  rval->val = alarm((Int)*arg[1]);
  return rval;
}
static cssEl* cssElCFun_chdir_stub(int, cssEl* arg[]) {
  cssInt* rval = new cssInt();
  rval->val = chdir((const char*)*arg[1]);
  return rval;
}
static cssEl* cssElCFun_chown_stub(int, cssEl* arg[]) {
  cssInt* rval = new cssInt();
  rval->val = chown((const char*)*arg[1], (int)*arg[2], (int)*arg[3]);
  return rval;
}
static cssEl* cssElCFun_ctermid_stub(int, cssEl**) {
  cssString* rval = new cssString();
  rval->val = ctermid(NULL);
  return rval;
}
static cssEl* cssElCFun_cuserid_stub(int, cssEl**) {
  cssString* rval = new cssString();
  rval->val = cuserid(NULL);
  return rval;
}
static cssEl* cssElCFun_getcwd_stub(int, cssEl**) {
  cssString* rval = new cssString();
  char* cwd = getcwd(NULL, 1024);
  rval->val = cwd;
  free(cwd);
  return rval;
}
static cssEl* cssElCFun_getenv_stub(int, cssEl* arg[]) {
  cssString* rval = new cssString();
  char* env_val = getenv((const char*)*arg[1]);
  if(env_val != NULL)
    rval->val = env_val;
  return rval;
}
static cssEl* cssElCFun_getegid_stub(int, cssEl**) {
  cssInt* rval = new cssInt();
  rval->val = (int)getegid();
  return rval;
}
static cssEl* cssElCFun_geteuid_stub(int, cssEl**) {
  cssInt* rval = new cssInt();
  rval->val = (int)geteuid();
  return rval;
}
static cssEl* cssElCFun_getgid_stub(int, cssEl**) {
  cssInt* rval = new cssInt();
  rval->val = (int)getegid();
  return rval;
}
static cssEl* cssElCFun_getuid_stub(int, cssEl**) {
  cssInt* rval = new cssInt();
  rval->val = (int)getuid();
  return rval;
}
static cssEl* cssElCFun_getlogin_stub(int, cssEl**) {
  cssString* rval = new cssString();
  rval->val = getlogin();
  return rval;
}
static cssEl* cssElCFun_getpgrp_stub(int, cssEl**) {
  cssInt* rval = new cssInt();
#if defined(SUN4) && !defined(SOLARIS)
  rval->val = (int)getpgrp(0);
#else
  rval->val = (int)getpgrp();
#endif
  return rval;
}
static cssEl* cssElCFun_getpid_stub(int, cssEl**) {
  cssInt* rval = new cssInt();
  rval->val = (int)getpid();
  return rval;
}
static cssEl* cssElCFun_getppid_stub(int, cssEl**) {
  cssInt* rval = new cssInt();
  rval->val = (int)getppid();
  return rval;
}
static cssEl* cssElCFun_isatty_stub(int, cssEl* arg[]) {
  cssInt* rval = new cssInt();
  rval->val = (int)isatty((int)*arg[1]);
  return rval;
}
static cssEl* cssElCFun_link_stub(int, cssEl* arg[]) {
  cssInt* rval = new cssInt();
  rval->val = (int)link((const char*)*arg[1], (const char*)*arg[2]);
  return rval;
}

#if defined(SUN4)|| defined(SUN4solaris)
extern "C" int symlink(const char*, const char*);
extern "C" int rename(const char*, const char*);
#endif

static cssEl* cssElCFun_symlink_stub(int, cssEl* arg[]) {
  cssInt* rval = new cssInt();
  rval->val = (int)symlink((const char*)*arg[1], (const char*)*arg[2]);
  return rval;
}
static cssEl* cssElCFun_unlink_stub(int, cssEl* arg[]) {
  cssInt* rval = new cssInt();
  rval->val = (int)unlink((const char*)*arg[1]);
  return rval;
}
static cssEl* cssElCFun_pause_stub(int, cssEl**) {
  cssInt* rval = new cssInt();
  rval->val = (int)pause();
  return rval;
}
static cssEl* cssElCFun_rename_stub(int, cssEl* arg[]) {
  cssInt* rval = new cssInt();
  rval->val = (int)rename((const char*)*arg[1], (const char*)*arg[2]);
  return rval;
}
static cssEl* cssElCFun_rmdir_stub(int, cssEl* arg[]) {
  cssInt* rval = new cssInt();
  rval->val = (int)rmdir((const char*)*arg[1]);
  return rval;
}
static cssEl* cssElCFun_setgid_stub(int, cssEl* arg[]) {
  cssInt* rval = new cssInt();
  rval->val = (int)setgid((int)*arg[1]);
  return rval;
}
static cssEl* cssElCFun_setpgid_stub(int, cssEl* arg[]) {
  cssInt* rval = new cssInt();
  rval->val = (int)setpgid((int)*arg[1], (int)*arg[2]);
  return rval;
}
static cssEl* cssElCFun_setuid_stub(int, cssEl* arg[]) {
  cssInt* rval = new cssInt();
  rval->val = (int)setuid((int)*arg[1]);
  return rval;
}
static cssEl* cssElCFun_sleep_stub(int, cssEl* arg[]) {
  cssInt* rval = new cssInt();
  rval->val = (int)sleep((int)*arg[1]);
  return rval;
}
static cssEl* cssElCFun_tcgetpgrp_stub(int, cssEl* arg[]) {
  cssInt* rval = new cssInt();
  rval->val = (int)tcgetpgrp((int)*arg[1]);
  return rval;
}
static cssEl* cssElCFun_tcsetpgrp_stub(int, cssEl* arg[]) {
  cssInt* rval = new cssInt();
  rval->val = (int)tcsetpgrp((int)*arg[1], (int)*arg[2]);
  return rval;
}
static cssEl* cssElCFun_ttyname_stub(int, cssEl* arg[]) {
  cssString* rval = new cssString();
  rval->val = ttyname((int)*arg[1]);
  return rval;
}
static cssEl* cssElCFun_perror_stub(int, cssEl* arg[]) {
  perror((const char*)*arg[1]);
  return &cssMisc::Void;
}

static cssEl* cssElCFun_putenv_stub(int, cssEl* arg[]) {
  static String_Array env_vals;	// need to keep strings around since putenv uses it..
  String env_val = (const char*)*arg[1];
  String env_cue = env_val.through('=');	// must have an equals..

  int i;
  for(i=0; i<env_vals.size; i++) {
    if(env_vals.FastEl(i).at(0,env_cue.length()) == env_cue) { // redefining existing
      env_vals.FastEl(i) = env_val;
      if(putenv((char*)env_vals.FastEl(i)))
	return cssBI::true_int; // no sense in creating more baggage..
      else
	return cssBI::false_int;
    }
  }
  env_vals.Add(env_val);
  if(putenv((char*)env_vals.Peek()))
    return cssBI::true_int; // no sense in creating more baggage..
  else
    return cssBI::false_int;
}

// possibly non-posix but nonetheless useful functions

#if defined(SUN4) && !defined(SOLARIS)
extern "C" {
  int gettimeofday(struct timeval*, struct timezone*);
  clock_t clock();
}
#define CLOCKS_PER_SEC 1000000
#endif

static cssEl* cssElCFun_gettimesec_stub(int, cssEl**) {
  struct timezone tzp;
  struct timeval tp;
  gettimeofday(&tp,&tzp);
  return new cssInt((int)tp.tv_sec);
}

static cssEl* cssElCFun_gettimemsec_stub(int, cssEl**) {
  struct timezone tzp;
  struct timeval tp;
  gettimeofday(&tp,&tzp);
  return new cssInt((int)tp.tv_usec);
}

static cssEl* cssElCFun_clock_stub(int, cssEl**) {
  clock_t clk = clock();
  double sec = (double)clk / (double)CLOCKS_PER_SEC;
  return new cssReal(sec);
}


// apparently this is not available on all platforms..
//
// #include <sys/resource.h>
// 
// extern "C" getrusage(int, struct rusage*);
// 
// static cssEl* cssElCFun_getrusage_stub(int, cssEl**) {
//   static int_Array vals;
//   static cssTA* rv_ta = NULL;
//   struct rusage rus;
//   getrusage(RUSAGE_SELF,&rus);
//   vals.EnforceSize(18);
//   vals[0] = (int)rus.ru_utime.tv_sec;  // user time used, secs
//   vals[1] = (int)rus.ru_utime.tv_usec; // user time used, microsecs
//   vals[2] = (int)rus.ru_stime.tv_sec;  // system time used, secs
//   vals[3] = (int)rus.ru_stime.tv_usec; // system time used, microsecs
//   vals[4] = (int)rus.ru_maxrss;	  // max resident size
//   vals[5] = (int)rus.ru_ixrss;	  // integral shared text memory size
//   vals[6] = (int)rus.ru_idrss;	  // integral unshared data size
//   vals[7] = (int)rus.ru_isrss;	  // integral unshared stack size
//   vals[8] = (int)rus.ru_minflt;	  // page reclaims
//   vals[9] = (int)rus.ru_majflt;	  // page faults
//   vals[10]= (int)rus.ru_nswap;	  // swaps
//   vals[11]= (int)rus.ru_inblock;	  // block input operations
//   vals[12]= (int)rus.ru_oublock;	  // block output operations
//   vals[13]= (int)rus.ru_msgsnd;	  // messages sent
//   vals[14]= (int)rus.ru_msgrcv;	  // messages received
//   vals[15]= (int)rus.ru_nsignals;	  // signals received
//   vals[16]= (int)rus.ru_nvcsw;	  // voluntary context switches
//   vals[17]= (int)rus.ru_nivcsw;	  // involuntary context switches
//   if(rv_ta == NULL) {
//     rv_ta = new cssTA(NULL, 1, &TA_int_Array);
//     cssEl::Ref(rv_ta);
//   }
//   rv_ta->ptr = (void*)&vals;
//   return new cssRef(rv_ta);
// }
// 

//////////////////////////////////
//	Install_MiscFun		//
//////////////////////////////////

static void Install_MiscFun() {

  // misc functions 
  cssElCFun_inst_nm(cssMisc::Functions, printr, 1, "PrintR", CSS_FUN);
  cssElCFun_inst_nm(cssMisc::Functions, edit, cssEl::VarArg, "EditObj", CSS_FUN);
  cssElCFun_inst(cssMisc::Functions, CancelEditObj, 1, CSS_FUN);
  cssElCFun_inst(cssMisc::Functions, Extern, 1, CSS_FUN);
  cssElCFun_inst(cssMisc::Functions, Token, 2, CSS_FUN);
  cssElCFun_inst(cssMisc::Functions, Type, 1, new cssTA(NULL, 1, &TA_TypeDef));
  cssElCFun_inst(cssMisc::Functions, Dir, cssEl::VarArg, new cssTA(NULL, 1, &TA_String_Array));
  cssElCFun_inst(cssMisc::Functions, ReadLine, 1, new cssTA(NULL, 1, &TA_String_Array));

  // old file-io
  cssElCFun_inst(cssMisc::Functions, fprintf, cssEl::VarArg, CSS_FUN);
  cssElCFun_inst(cssMisc::Functions, fopen, 2, CSS_FUN);
  cssElCFun_inst(cssMisc::Functions, fclose, 1, CSS_FUN);
  cssElCFun_inst(cssMisc::Functions, printf, cssEl::VarArg, CSS_FUN);

  // POSIX functions
  cssElCFun_inst(cssMisc::Functions, access, 2, CSS_FUN);
  cssEnum_inst_nm(cssMisc::Enums, R_OK, "R_OK");
  cssEnum_inst_nm(cssMisc::Enums, W_OK, "W_OK");
  cssEnum_inst_nm(cssMisc::Enums, X_OK, "X_OK");
  cssEnum_inst_nm(cssMisc::Enums, F_OK, "F_OK");
  cssElCFun_inst(cssMisc::Functions, alarm, 1, CSS_FUN);
  cssElCFun_inst(cssMisc::Functions, chdir, 1, CSS_FUN);
  cssElCFun_inst(cssMisc::Functions, chown, 3, CSS_FUN);
  cssElCFun_inst(cssMisc::Functions, ctermid, 0, CSS_FUN);
  cssElCFun_inst(cssMisc::Functions, cuserid, 0, CSS_FUN);
  cssElCFun_inst(cssMisc::Functions, getcwd, 0, CSS_FUN);
  cssElCFun_inst(cssMisc::Functions, getenv, 1, CSS_FUN);
  cssElCFun_inst(cssMisc::Functions, getegid, 0, CSS_FUN);
  cssElCFun_inst(cssMisc::Functions, geteuid, 0, CSS_FUN);
  cssElCFun_inst(cssMisc::Functions, getgid, 0, CSS_FUN);
  cssElCFun_inst(cssMisc::Functions, getuid, 0, CSS_FUN);
  cssElCFun_inst(cssMisc::Functions, getlogin, 0, CSS_FUN);
  cssElCFun_inst(cssMisc::Functions, getpgrp, 0, CSS_FUN);
  cssElCFun_inst(cssMisc::Functions, getpid, 0, CSS_FUN);
  cssElCFun_inst(cssMisc::Functions, getppid, 0, CSS_FUN);
  cssElCFun_inst(cssMisc::Functions, isatty, 1, CSS_FUN);
  cssElCFun_inst(cssMisc::Functions, link, 2, CSS_FUN);
  cssElCFun_inst(cssMisc::Functions, symlink, 2, CSS_FUN);
  cssElCFun_inst(cssMisc::Functions, unlink, 1, CSS_FUN);
  cssElCFun_inst(cssMisc::Functions, pause, 0, CSS_FUN);
  cssElCFun_inst(cssMisc::Functions, rename, 2, CSS_FUN);
  cssElCFun_inst(cssMisc::Functions, rmdir, 1, CSS_FUN);
  cssElCFun_inst(cssMisc::Functions, setgid, 1, CSS_FUN);
  cssElCFun_inst(cssMisc::Functions, setpgid, 2, CSS_FUN);
  cssElCFun_inst(cssMisc::Functions, setuid, 1, CSS_FUN);
  cssElCFun_inst(cssMisc::Functions, sleep, 1, CSS_FUN);
  cssElCFun_inst(cssMisc::Functions, tcgetpgrp, 1, CSS_FUN);
  cssElCFun_inst(cssMisc::Functions, tcsetpgrp, 2, CSS_FUN);
  cssElCFun_inst(cssMisc::Functions, ttyname, 1, CSS_FUN);
  cssElCFun_inst(cssMisc::Functions, perror, 1, CSS_FUN);
  cssCPtr_int_inst(cssMisc::HardVars, errno);
  cssElCFun_inst(cssMisc::Functions, putenv, 1, CSS_FUN);
  cssElCFun_inst(cssMisc::Functions, system, 1, CSS_FUN);

  cssElCFun_inst(cssMisc::Functions, gettimesec, 0, CSS_FUN);
  cssElCFun_inst(cssMisc::Functions, gettimemsec, 0, CSS_FUN);
  cssElCFun_inst(cssMisc::Functions, clock, 0, CSS_FUN);
//  cssElCFun_inst(cssMisc::Functions, getrusage, 0, CSS_FUN);

  // parsed internal ptrs
  cssMisc::HardVars.Push(new cssTA((void*)&cout, 1, &TA_ostream, "cout"));
  cssMisc::HardVars.Push(new cssTA((void*)&cin, 1, &TA_istream, "cin"));
  cssMisc::HardVars.Push(new cssTA((void*)&cerr, 1, &TA_ostream, "cerr"));

  cssMisc::HardVars.Push(new cssTA((void*)&cout, 1, &TA_ostream, "stdout"));
  cssMisc::HardVars.Push(new cssTA((void*)&cin, 1, &TA_istream, "stdin"));
  cssMisc::HardVars.Push(new cssTA((void*)&cerr, 1, &TA_ostream, "stderr"));

  // magic stream codes!
  cssMisc::Constants.Push(new cssBool(false, "ws"));
  cssMisc::Constants.Push(new cssBool(false, "flush"));
  cssMisc::Constants.Push(new cssBool(false, "endl"));
  cssMisc::Constants.Push(new cssBool(false, "ends"));
  cssMisc::Constants.Push(new cssBool(false, "dec"));
  cssMisc::Constants.Push(new cssBool(false, "hex"));
  cssMisc::Constants.Push(new cssBool(false, "oct"));
  cssMisc::Constants.Push(new cssBool(false, "lock"));
  cssMisc::Constants.Push(new cssBool(false, "unlock"));

  // constants  
  cssMisc::Enums.Push(new cssBool(true, "true"));
  cssMisc::Enums.Push(new cssBool(false, "false"));
  cssBI::true_int = new cssInt(1, "true");
  cssBI::false_int = new cssInt(0, "false");
  cssMisc::Internal.Push(cssBI::true_int); // keep them on this list (to refn)
  cssMisc::Internal.Push(cssBI::false_int);
  cssInt_inst	(cssMisc::Constants, NULL, NULL);
}


//////////////////////////////////
//	Install_Types		//
//////////////////////////////////

static void Install_Types() {
  // basic CSS types
  cssMisc::TypesSpace.Push(&cssMisc::Void);
  cssMisc::TypesSpace.Push(new cssEl("void"));
  cssInt_inst_nm(cssMisc::TypesSpace, 0, "Int");
  cssInt_inst_nm(cssMisc::TypesSpace, 0, "int");
  cssMisc::TypesSpace.Push(new cssBool(false, "bool"));
  cssMisc::TypesSpace.Push(new cssBool(false, "Bool"));
  cssReal_inst_nm(cssMisc::TypesSpace, 0.0, "Real");
  cssReal_inst_nm(cssMisc::TypesSpace, 0.0, "real");
  cssReal_inst_nm(cssMisc::TypesSpace, 0.0, "float");
  cssReal_inst_nm(cssMisc::TypesSpace, 0.0, "double");
  cssString_inst_nm(cssMisc::TypesSpace, "", "String");
  cssString_inst_nm(cssMisc::TypesSpace, "", "string");
  cssMisc::TypesSpace.Push(new cssChar(' ', "char"));
  cssCPtr_inst_nm(cssMisc::TypesSpace, 0, 1, "c_ptr");
  cssMisc::TypesSpace.Push(new cssSubShell("SubShell"));

  // CPtr types
  cssMisc::TypesSpace.Push(new cssCPtr_int(NULL, 1, "c_int"));
  cssMisc::TypesSpace.Push(new cssCPtr_bool(NULL, 1, "c_bool"));
  cssMisc::TypesSpace.Push(new cssCPtr_short(NULL, 1, "c_short"));
  cssMisc::TypesSpace.Push(new cssCPtr_long(NULL, 1, "c_long"));
  cssMisc::TypesSpace.Push(new cssCPtr_char(NULL, 1, "c_char"));
  cssMisc::TypesSpace.Push(new cssCPtr_enum(NULL, 1, "c_enum"));
  cssMisc::TypesSpace.Push(new cssCPtr_double(NULL, 1, "c_double"));
  cssMisc::TypesSpace.Push(new cssCPtr_float(NULL, 1, "c_float"));
  cssMisc::TypesSpace.Push(new cssCPtr_String(NULL, 1, "c_String"));

  // ta types
  cssTA_inst_nm (cssMisc::TypesSpace, NULL, 1, &TA_void, "TA");
  cssTA_Base_inst_nm (cssMisc::TypesSpace, NULL, 1, &TA_taBase, "TAPtr");
  cssTA_Base_inst_nm (cssMisc::TypesSpace, NULL, 1, &TA_taBase, "taBase");
  cssMisc::TypesSpace.Push(new cssFStream("fstream"));
  cssMisc::TypesSpace.Push(new cssStrStream("strstream"));
  cssTA_inst_nm (cssMisc::TypesSpace, NULL, 1, &TA_fstream, "FILE");
  cssMisc::TypesSpace.Push(new cssLeafItr("taLeafItr"));

  // add all user-defined types..
  int i, j;
  for(i=0; i<taMisc::types.size; i++) {
    TypeDef* tmp = taMisc::types.FastEl(i);
    if(tmp->InheritsFormal(TA_class)) {
      // don't add any template instances that have a single further subclass
      // (use the subclass instead)
      if(tmp->InheritsFormal(TA_templ_inst)) {
	if((tmp->children.size != 1) || (tmp->children.FastEl(0)->parents.size > 1)) {
	  if(tmp->InheritsFrom(TA_taBase))
	    cssTA_Base_inst_nm(cssMisc::TypesSpace, tmp->GetInstance(), 1, tmp, tmp->name);
	  else
	    cssTA_inst_nm(cssMisc::TypesSpace, tmp->GetInstance(), 1, tmp, tmp->name);
	}
      }
      else if(!((tmp->members.size==0) && (tmp->methods.size==0)) &&
	      (tmp->name != "fstream") && (tmp->name != "class") && 
	      (tmp->name != "taString") && (tmp->name != "taLeafItr") &&
	      (tmp->name != "strstream"))
      {
	if(tmp->InheritsFrom(TA_taBase))
	  cssTA_Base_inst_nm(cssMisc::TypesSpace, tmp->GetInstance(), 1, tmp, tmp->name);
	else
	  cssTA_inst_nm(cssMisc::TypesSpace, tmp->GetInstance(), 1, tmp, tmp->name);
      }
    }
    else if(tmp->InheritsFormal(TA_enum)) {
      cssEnum_inst_nm(cssMisc::Enums, 0, tmp->name);
      for(j=0; j < tmp->enum_vals.size; j++)
	cssInt_inst_nm(cssMisc::Enums, tmp->enum_vals.FastEl(j)->enum_no,
			tmp->enum_vals.FastEl(j)->name);
    }
    else if(tmp->InheritsFrom(TA_taRegFun)) {
      for(j=0; j < tmp->methods.size; j++) {
	MethodDef* md = tmp->methods.FastEl(j);
	if(md->stubp != NULL)
	  cssMisc::HardFuns.Push(new cssMbrCFun(md->fun_argc, NULL, md->stubp,
					     (const char*)md->name));
      }
    }
  }
}

//////////////////////////////////
//   Completion in Readline	//
//////////////////////////////////

extern "C" {
  typedef int rl_function(void);
  typedef char* rl_generator_fun(char*, int);
  extern int (*rl_attempted_completion_function)(void);
  extern char** completion_matches (char* text, rl_generator_fun* gen);
  extern char** css_attempted_completion(char* text, int start, int end);
  extern char* css_path_generator(char* text, int state);
  extern char* css_scoped_generator(char* text, int state);
  extern char* css_keyword_generator(char* text, int state);
  extern char *rl_line_buffer;
}

char** css_attempted_completion(char* text, int start, int) {
  char** matches = NULL;
  String txt_str = text;

  // if starting with a quote, assume its a file name
  if((start > 0) && (rl_line_buffer[start-1] == '\"')) // "
    return matches;
  // if it starts with a dot, then its a path
  if(text[0] == '.')
    matches = completion_matches(text, css_path_generator);
  // check for a scoped type
  else if(txt_str.contains("::"))
    matches = completion_matches(text, css_scoped_generator);
  // otherwise, it must be a keyword of some sort..
  else
    matches = completion_matches(text, css_keyword_generator);

  return matches;
}

static char* css_malloc_string(const String& str) {
  char* rval = (char*)malloc((str.length()+1) * sizeof(char*));
  strcpy(rval, (const char*)str);
  return rval;
}

static char* css_cmd_gen_get_nm(cssSpace* spc, int& idx, char* text, int len) {
  while(idx < spc->size) {
    cssEl* el = spc->FastEl(idx++);
    if(el->name(0,len) == text)
      return css_malloc_string(el->name);
  }
  return NULL;
}

char* css_keyword_generator(char* text, int state) {
  static int spc_idx;		// index in spaces searching through
  static int item_idx;		// index in items searching through
  static int len;
  
  if(state == 0) {
    spc_idx = 0;
    item_idx = 0;
    len = strlen(text);
  }

  cssSpace* spc = NULL;
  int extra_spcs = 1;
  if(cssMisc::cur_class != NULL)
    extra_spcs++;

  if(spc_idx == 0)
    spc = &cssMisc::TypesSpace;
  else if((cssMisc::cur_class != NULL) && (spc_idx == 1))
    spc = cssMisc::cur_class->types;
  else
    spc = cssMisc::cur_top->GetParseSpace(spc_idx-extra_spcs);

  char* rval = NULL;
  while((spc != NULL) && ((rval = css_cmd_gen_get_nm(spc, item_idx, text, len)) == NULL)) {
    item_idx = 0;
    // use postfix because subtracting one due to TypesSpace being first
    spc = cssMisc::cur_top->GetParseSpace(spc_idx++);
  }
  return rval;
}

static String* css_path_search(TypeDef* td, int lst_idx, int& idx, String& mb_name) {
  if(lst_idx == 0) {
    while(idx < td->members.size) {
      MemberDef* md = td->members.FastEl(idx++);
      if(md->name(0,mb_name.length()) == mb_name)
	return &(md->name);
    }
    return NULL;
  }
  else {
    while(idx < td->methods.size) {
      MethodDef* md = td->methods.FastEl(idx++);
      if(md->name(0,mb_name.length()) == mb_name)
	return &(md->name);
    }
    return NULL;
  }
}

char* css_path_generator(char* text, int state) {
  static int lst_idx;		// list index
  static int item_idx;		// index for searching within list
  static String mb_name;
  static String par_path;
  static TAPtr parent;
  static TypeDef* par_td;

  if(cssBI::root == NULL)
    return NULL;

  if(state == 0) {
    par_path = text;
    par_path = par_path.before('.',-1);
    if(par_path == "") {
      parent = (TAPtr)cssBI::root->ptr;
      if(parent != NULL)
	par_td = parent->GetTypeDef();
    }
    else {
      MemberDef* md = NULL;
      TAPtr root = (TAPtr)cssBI::root->ptr;
      if(root == NULL)
	parent = NULL;
      else {
	parent = root->FindFromPath(par_path, md);
	if(parent != NULL) {
	  par_td = parent->GetTypeDef();
// no need to do this..
//	  par_path = parent->GetPath();
	}
      }
    }
    mb_name = text;
    mb_name = mb_name.after('.', -1);
    item_idx = 0;
    lst_idx = 0;
  }

  if(parent == NULL)
    return NULL;

  String* rval = NULL;
  while((lst_idx < 2) &&
	((rval = css_path_search(par_td, lst_idx, item_idx, mb_name)) == NULL))
  {
    item_idx = 0;
    lst_idx++;
  }

  if(rval == NULL)
    return NULL;

  String full_val = par_path;
  full_val += ".";
  full_val += *rval;
  return css_malloc_string(full_val);
}

static String* css_scope_search(TypeDef* td, int lst_idx, int& idx, int& sub_itm, 
				String& mb_name)
{
  if(lst_idx == 0) {
    while(idx < td->members.size) {
      MemberDef* md = td->members.FastEl(idx++);
      if(md->name(0,mb_name.length()) == mb_name)
	return &(md->name);
    }
    return NULL;
  }
  else if(lst_idx == 1) {
    while(idx < td->methods.size) {
      MethodDef* md = td->methods.FastEl(idx++);
      if(md->name(0,mb_name.length()) == mb_name)
	return &(md->name);
    }
    return NULL;
  }
  else {
    while(idx < td->sub_types.size) {
      TypeDef* st = td->sub_types.FastEl(idx);
      if(st->InheritsFormal(TA_enum)) {
	while(sub_itm < st->enum_vals.size) {
	  EnumDef* en = st->enum_vals.FastEl(sub_itm++);
	  if(en->name(0,mb_name.length()) == mb_name)
	    return &(en->name);
	}
      }
      idx++;
      sub_itm = 0;
      if(st->name(0,mb_name.length()) == mb_name)
	return &(st->name);
    }
    return NULL;
  }
}

char* css_scoped_generator(char* text, int state) {
  static int lst_idx;		// list index
  static int item_idx;		// index for searching within list
  static int sub_item_idx;	// index of sub-item within item (if applic)
  static String mb_name;
  static String par_path;
  static TypeDef* par_td;

  if(state == 0) {
    par_path = text;
    par_path = par_path.before("::",-1);
    par_td = taMisc::types.FindName(par_path);
    mb_name = text;
    mb_name = mb_name.after("::", -1);
    item_idx = 0;
    lst_idx = 0;
    sub_item_idx = 0;
  }

  if(par_td == NULL)
    return NULL;

  String* rval = NULL;
  while((lst_idx < 3) &&
	((rval = css_scope_search(par_td, lst_idx, item_idx, sub_item_idx, mb_name))
	 == NULL)) {
    item_idx = 0;
    sub_item_idx = 0;
    lst_idx++;
  }

  if(rval == NULL)
    return NULL;

  String full_val = par_path;
  full_val += "::";
  full_val += *rval;
  return css_malloc_string(full_val);
}


//////////////////////////////////
//	Initialize		//
//////////////////////////////////

void cssMisc::Initialize(int argc, const char** argv) {
  // use our completion function
  rl_attempted_completion_function = (rl_function*)css_attempted_completion;

  // functions
  cssEl::Ref(&cssMisc::Void);		// reference this to keep it around

  cssMisc::Top = new cssProgSpace("C^c Top Level");	// top level
  cssMisc::ConstExprTop = new cssProgSpace("Constant Expression Top Level");
  cssMisc::ConstExpr = cssMisc::ConstExprTop->Prog(); // this guy is child of top
  cssMisc::CDtorProg = new cssProg("Constructor/Destructor Prog");
  cssMisc::CallFunProg = new cssProg("CallFun Prog");
  cssProg::Ref(cssMisc::CDtorProg);
  cssProg::Ref(cssMisc::CallFunProg);
  cssMisc::cur_top = cssMisc::Top;

  Install_Internals();
  Install_Commands();
  Install_Types();
  Install_Math();
  Install_MiscFun();

  cssMisc::s_argv = new cssArray(1, cssMisc::VoidString.Clone(), "argv");
  cssMisc::s_argc = new cssInt(0, "argc");
  cssMisc::HardVars.Push(cssMisc::s_argv);
  cssMisc::HardVars.Push(cssMisc::s_argc);

  // setting variables
  cssMisc::Settings.Push(new cssTA(&taMisc::include_paths, 1, &TA_String_PArray,
				    "include_paths"));

  cssMisc::Settings.Push(new cssTA(TA_taMisc.GetInstance(), 1, &TA_taMisc,
				    "cssSettings"));

  cssMisc::TypesSpace.Sort();		// must sort before anything happens
  cssMisc::Commands.Sort();
  cssMisc::Functions.Sort();		// must sort before anything happens
  cssMisc::Constants.Sort();		// must sort before anything happens

  String home = getenv("HOME");
  taMisc::include_paths.AddUnique(home); // this is first on the list..

#ifndef CYGWIN
  String css_dir = "/usr/local/pdp++"; // default css home directory
#else
  String css_dir = "C:/PDP++"; // default pdp home directory
#endif
  char* css_dir_env = getenv("CSSDIR");
  if(css_dir_env != NULL)
    css_dir = css_dir_env;

  taMisc::include_paths.AddUnique(css_dir);
  taMisc::include_paths.AddUnique(css_dir+"/css/include");

  if(argc > 0) {
    // args can get munged together when #!/usr/local/css, so need to pre-process
    String_Array act_args;
    int i;
    for(i=0;i<argc;i++) {
      String tmp = argv[i];
      if(tmp.contains('-') && tmp.contains(' ')) { // args munged from auto start script
	while(tmp.contains('-') && tmp.contains(' ')) {
	  String t2 = tmp.before(' ');
	  tmp = tmp.after(' ');
	  act_args.Add(t2);
	}
	if(tmp.contains('-'))
	  act_args.Add(tmp);
      }
      else
	act_args.Add(tmp);
    }
    cssMisc::s_argv->Alloc(act_args.size); // allocate proper number of args
    *cssMisc::s_argc = 0;
    i = 0;
    while(i < act_args.size) {
      String tmp = act_args[i];
      if(((tmp == "-f") || (tmp == "-file")) && (i+1 < act_args.size)) { // compile the file
	cssMisc::startup_file = act_args[++i];
      }
      else if(((tmp == "-e") || (tmp == "-exec")) && (i+1 < act_args.size)) { // compile the file
	cssMisc::startup_code = act_args[++i];
//	cssMisc::startup_code.gsub(";\\n", ";\n"); // translate cr.s
	cssMisc::startup_code += "\n";		  // add a final cr for good measure
      }
      else if((tmp == "-i") || (tmp == "-interactive")) {
	cssMisc::init_interactive = true;
      }
      else if(tmp(0,2) == "-v") {
	String vls = tmp.after(1);
	int vl = (int)vls;
	if(vl > 0)
	  cssMisc::init_debug = vl;
	else
	  cssMisc::init_debug = 1;
      }
      else if(((tmp == "-b") || (tmp == "-bp")) && (i+1 < act_args.size)) {
	int vl = (int)act_args[++i];
	cssMisc::init_bpoint = vl;
      }
      else {	   // pass the arg to the script
	int ac = (int)(*cssMisc::s_argc);
	cssString* av = (cssString*)(*cssMisc::s_argv)[ac];
	*av = tmp;
	*cssMisc::s_argc = (int)(*cssMisc::s_argc) + 1;
	if(i == 0) {
	  if(tmp.contains('/'))
	    tmp = tmp.after('/',-1);
	  cssMisc::prompt = tmp; // default prompt is this program
	}
      }
      i++;
    }
  }
}
