/* -*- C++ -*- */
/*=============================================================================
//									      //
// This file is part of the PDP++ software package.			      //
//									      //
// Copyright (C) 1995 Randall C. O'Reilly, Chadley K. Dawson, 		      //
//		      James L. McClelland, and Carnegie Mellon University     //
//     									      //
// Permission to use, copy, and modify this software and its documentation    //
// for any purpose other than distribution-for-profit is hereby granted	      //
// without fee, provided that the above copyright notice and this permission  //
// notice appear in all copies of the software and related documentation.     //
//									      //
// Permission to distribute the software or modified or extended versions     //
// thereof on a not-for-profit basis is explicitly granted, under the above   //
// conditions. 	HOWEVER, THE RIGHT TO DISTRIBUTE THE SOFTWARE OR MODIFIED OR  //
// EXTENDED VERSIONS THEREOF FOR PROFIT IS *NOT* GRANTED EXCEPT BY PRIOR      //
// ARRANGEMENT AND WRITTEN CONSENT OF THE COPYRIGHT HOLDERS.                  //
// 									      //
// Note that the taString class, which is derived from the GNU String class,  //
// is Copyright (C) 1988 Free Software Foundation, written by Doug Lea, and   //
// is covered by the GNU General Public License, see ta_string.h.             //
// The iv_graphic library and some iv_misc classes were derived from the      //
// InterViews morpher example and other InterViews code, which is             //
// Copyright (C) 1987, 1988, 1989, 1990, 1991 Stanford University             //
// Copyright (C) 1991 Silicon Graphics, Inc.				      //
//									      //
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,         //
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 	      //
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  	      //
// 									      //
// IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE FOR ANY SPECIAL,    //
// INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES  //
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT     //
// ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY,      //
// ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS        //
// SOFTWARE. 								      //
==============================================================================*/

#ifndef pdplog_h
#define pdplog_h

#include <pdp/process.h>
#include <ta_misc/datatable.h>
#include <ta_misc/ta_file.h>
#include <ta_misc/fontspec.h>

class LogView;
class GridLogView;
class GraphLogView;
class TextLogView;
class NetLogView;

class PDPLog : public WinMgr {
  // ##EXT_log Records data from processes, displaying and saving it
public:
  taFile	log_file;	// optional file for saving
  int		log_lines;	// #NO_SAVE number of lines in the log
  LogData	log_data;	// #NO_SAVE #HIDDEN our own log data for reading fm file

  DataTable	data;		// data for the log
  int		data_bufsz;	// how big a data buffer size to keep
  float		data_shift;	// percentage to shift buffer upon overflow
  MinMaxInt	data_range;	// #NO_SAVE range of lines in the data buffer (in log lines)
  bool		record_proc_name; // whether to record process name in log file or not

  Process_Group log_proc;	// #LINK_GROUP processes which use this log
  SchedProcess*	cur_proc;	// #READ_ONLY #NO_SAVE current process sending to log
  

  virtual void 	GetHeaders();
  // #MENU #MENU_ON_Actions #CONFIRM Get Header information from proceses \n and clear out all current data
  virtual void	SetSaveFile(const char* nm = NULL, bool no_dlg = false);
  // #MENU #MENU_ON_LogFile #ARGC_0 open file to save log data to (overwrite existing file)
  virtual void	SetAppendFile(const char* nm = NULL, bool no_dlg = false);
  // #MENU #ARGC_0 open file to append log data to
  virtual void	LoadFile(const char* nm = NULL, bool no_dlg = false);
  // #MENU #ARGC_0 read in existing log file data
  virtual void	CloseFile();
  // #MENU #MENU_SEP_AFTER close any open files being logged to
  virtual void	HeadToFile();
  // #MENU send current header to log file
  virtual void	BufferToFile();
  // #MENU #CONFIRM send current buffer of data to log file

  virtual void 	Buffer_F();	// forward by data_shift
  virtual void 	Buffer_R();	// rewind by data_shift
  virtual void 	Buffer_FF();	// forward to end
  virtual void 	Buffer_FR();	// rewind to begining

  virtual void	Clear();	// clears out the data 
  virtual void	UpdateViewHeaders();           // update headers for all views

  virtual void	AddUpdater(SchedProcess* updt_proc);
  /* #NEW_FUN #MENU #MENU_ON_Object #MENU_SEP_BEFORE
     Tell updt_proc to update this log with new log data as the process runs */
  virtual void	RemoveUpdater(SchedProcess* updt_proc);
  // #MENU #MENU_ON_Object #FROM_GROUP_log_proc remove given updating process
  virtual void	RemoveAllUpdaters(); 		// remove this from all process logs
  virtual void	SyncLogViewUpdaters(); 		// #IGNORE 

  // for processes communicating stuff to the log (and log acting on this)
  virtual void  NewData(LogData& ld, SchedProcess* sproc);
  // This is the primary call to make when sending data to the log
  virtual void	NewHead(LogData& ld, SchedProcess* sproc);
  // #IGNORE use this to specifically update the field headers without updating data
  virtual void	HeadToLogFile(LogData& ld);	// #IGNORE
  virtual void	DataToLogFile(LogData& ld);	// #IGNORE
  virtual void	HeadToBuffer(LogData& ld);	// #IGNORE
  virtual void	DataToBuffer(LogData& ld);	// #IGNORE

  // for reading log files from disk to buffer
  virtual int 	LogLineType(char* lnc);		// #IGNORE determine type of given line
  virtual int  	FileScanTo(istream& strm, int log_ln); // #IGNORE scan to log_ln no
  virtual int	GetFileLength();		// #IGNORE gets length of current file
  virtual void	HeadFromLogFile(String& hdln);	// #IGNORE put log header into buffer
  virtual void	DataFromLogFile(String& dtln);	// #IGNORE put log data into buffer
  virtual void	ReadNewLogFile();		// #IGNORE a new log file was opened, read it
  virtual void	LogFileToBuffer();		// #IGNORE read current data_range into buffer
  virtual void 	InitFile();			// #IGNORE initialize log_file object

  // helper routines to set header/data for buffer field
  virtual void	SetFieldHead(LogData& ld, int fld, int top_cnt, int subgp_n,
			     int subgp_max, int subgp_cnt); // #IGNORE
  virtual void	SetFieldData(LogData& ld, int fld, int top_cnt, int subgp_n,
			     int subgp_max, int subgp_cnt); // #IGNORE
  static ostream& LogColumn(ostream& strm, String& str, int tabs);
  // #IGNORE output one column of data (string str), tabs is width (1 or 2)

  // for maintaining an internal version of the buffer format in log-data terms
  virtual DataItem* DataItemFromDataArray(DataArray_impl* da); 	   // #IGNORE
  virtual void 	LogDataFromDataTable(DataTable* dt, int st_idx=0); // #IGNORE
  virtual void	LogDataFromBuffer();	// #IGNORE get log data records from current buffer

  TypeDef*	GetDefaultView()	{ return &TA_LogView; }

  int	Dump_Save_Value(ostream& strm, TAPtr par, int indent); // check for open log_file..
  void	UpdateAfterEdit();

  void	Initialize();
  void 	Destroy();
  void	InitLinks();
  void	CutLinks();
  void	Copy_(const PDPLog& cp); 
  COPY_FUNS(PDPLog, WinMgr);
  TA_BASEFUNS(PDPLog);
};

BaseGroup_of(PDPLog);

class PDPLog_MGroup : public PDPMGroup {
public:
  void	Initialize() 		{ SetBaseType(&TA_PDPLog); }
  void 	Destroy()		{ };
  TA_BASEFUNS(PDPLog_MGroup);
};

class TextLog : public PDPLog {
public:
  TypeDef*	GetDefaultView()	{ return &TA_TextLogView; }
  void	Initialize() 		{ };
  void 	Destroy()		{ };
  TA_BASEFUNS(TextLog);
};

class NetLog: public TextLog {
public:
  virtual void	SetNetwork(Network* net);
  // #MENU #MENU_ON_Object select given network as the one to update views on

  TypeDef*	GetDefaultView()	{ return &TA_NetLogView; }
  void	Initialize() 		{ };
  void 	Destroy()		{ };
  TA_BASEFUNS(NetLog);
};
  
class GraphLog : public TextLog {
public:
  TypeDef*	GetDefaultView()	{ return &TA_GraphLogView; }
  void	Initialize() 		{ };
  void 	Destroy()		{ };
  TA_BASEFUNS(GraphLog);
};

class GridLog : public TextLog {
public:
  TypeDef*	GetDefaultView()	{ return &TA_GridLogView; }
  void	Initialize() 		{ };
  void 	Destroy()		{ };
  TA_BASEFUNS(GridLog);
};

class ivLayoutKit;		// #IGNORE
class ivWidgetKit;		// #IGNORE
class ivButton;			// #IGNORE

class LogView : public PDPView {
  // parent class of all log views
public:
  int		view_bufsz;	// Maximum number of lines in visible buffer
  float		view_shift;	// precentage of buffer to shift
  MinMaxInt	view_range;	// #NO_SAVE range of visible lines (in log lines)
  DT_ViewSpec*	viewspec;	// #NO_SAVE global view spec for all data 

  DT_ViewSpec 	viewspecs;	// #HIDDEN specification for display of datatable
  PDPLog* 	log_mgr;	// #READ_ONLY #NO_SAVE Parent Log

  // graphical implementation stuff
  ivButton*	disp_tog_but;	// #IGNORE display toggle button
  ivButton*	frev_but;	// #IGNORE full rev
  ivButton*	fsrev_but;	// #IGNORE fast rev
  ivButton*	rev_but;	// #IGNORE rev
  ivButton*	fwd_but;	// #IGNORE fwd
  ivButton*	fsfwd_but;	// #IGNORE fast fwd
  ivButton*	ffwd_but;	// #IGNORE full fwd
  ivButton*	clear_but;	// #IGNORE 
  ivButton*	init_but;	// #IGNORE 
  ivButton*	update_but;	// #IGNORE 

  // log interface
  virtual void  NewHead(); // new header information is in log data
  virtual void  NewData(){ }; // new data information is in log's datatable

  // relation to the buffer
  virtual void	UpdateFromBuffer() { }; // update view from buffer

  // view control
  virtual void 	View_F();	// forward by view_shift
  virtual void 	View_R();	// rewind by view_shift
  virtual void 	View_FSF();	// fast forward 
  virtual void 	View_FSR();	// fast rewind 
  virtual void 	View_FF();	// forward to end
  virtual void 	View_FR();	// rewind to begining

  virtual void 	Clear();	// Clear the display and the data
  virtual void	Clear_impl();

  void	InitDisplay();
  void  UpdateDisplay(TAPtr updtr=NULL);
  void	UpdateCallback()	{ UpdateDisplay(NULL); } // #IGNORE

  virtual void 	ToggleDisplay();
  virtual void 	SetToggle(bool value);

  virtual void	Reset() { };	// just clear the display, not the data..

  void		EditViewSpec(taBase* column);
  // #MENU #MENU_ON_Actions #ARGC_1 #FROM_GROUP_viewspec edit view specs for given column
  void		ToggleVisibility(taBase* column);
  // #MENU #ARGC_1 #FROM_GROUP_viewspec Toggle the visibility of given column

  virtual void		CreateViewSpec(); // #IGNORE create view spec of appropriate type
  virtual TypeDef*	DT_ViewSpecType(); // #IGNORE get type of DataTable for this log
  virtual TypeDef*	DA_ViewSpecType(); // #IGNORE get type of DataArray for this log

  // gui implementation stuff
  virtual void		CreateLogButtons(); // #IGNORE just make them
  virtual ivGlyph*	GetLogButtons(); // #IGNORE create and package

  // winview stuff
  void		AddNotify(TAPtr ud); // #IGNORE
  void	 	RemoveNotify(TAPtr ud);	// #IGNORE
  void  	SetWinName();
  void		CloseWindow();
  ivGlyph*	GetPrintData();	// #IGNORE

  void 	UpdateAfterEdit();
  void 	Initialize();
  void 	Destroy()	{ CutLinks(); }
  void 	InitLinks();
  void	CutLinks();
  void	Copy_(const LogView& cp);
  COPY_FUNS(LogView, PDPView);
  TA_BASEFUNS(LogView);
};

class ivPolyGlyph; // #IGNORE
class ivGlyph; // #IGNORE
class ivPatch; // #IGNORE

class TextLogView : public LogView {
  // a textual view of the log data
public:
  ivPolyGlyph*  sb;		// #IGNORE scrollbox for screenlog
  ivGlyph*	h_line;		// #IGNORE headerline for screenlog
  ivPatch*	patch;		// #IGNORE update patch;
  ivPatch*	scrb;		// #IGNORE update scrollable;
  ivPolyGlyph*	vbox;		// #IGNORE vbox with headerline and scrollbox
  ivPolyGlyph*	ovbox;		// #IGNORE outer vbox with buttons and vbox
  float		font_width;	// #IGNORE width of font spacing
  ivPolyGlyph*  cur_line;	// #IGNORE current Line we are creating

  void  		NewHead();
  void  		NewData();

  float		GetFieldWidth(DA_TextViewSpec* vs);// determines width of a field
  void		MakeLine(int from_end = 0); // creates a data line 
  void		UpdateFromBuffer(); 
  
  void		Reset();
  void		Clear_impl();

  int		ShowMore(int newsize); // #IGNORE
  int		ShowLess(int newsize); // #IGNORE

  TypeDef*	DT_ViewSpecType(); // #IGNORE get type of DataTable for this log
  TypeDef*	DA_ViewSpecType(); // #IGNORE get type of DataArray for this log

  // winview stuff
  void 		GetBodyRep();	// #IGNORE
  ivGlyph*	GetPrintData();	// #IGNORE

  void		UpdateDisplay(TAPtr updtr=NULL);
  void 		InitDisplay();
  void 		CloseWindow();

  // gui implementation stuff
  virtual void	AddToLine(const char* name, float width); // add name of width to line
  virtual void	AddLine();			  // add line to scrollbox
  virtual void  RemoveLine(int index);		  // remove line from scrollbox, -1 = all

  virtual void  StartHeaderLine();
  virtual void  AddToHeaderLine(const char* name, float width);
  virtual void  EndHeaderLine();
  virtual void	FixSize();

  void 	UpdateAfterEdit();
  void	Initialize();
  void	Destroy()	{ };
  TA_BASEFUNS(TextLogView);
};

class NetLogView : public LogView {
  // displays log information in the network view window
public:
  Network*      network;		// Network to whose views  data is logged

  void		SetNetwork(Network* net);
  // #MENU #MENU_ON_Object select given network as the one to update views on
  void 		NewHead();
  void 		NewData();
  void		UpdateDisplay(TAPtr updtr=NULL);
  void 		InitDisplay();
  void		UpdateFromBuffer(); 
  void 		UpdateAfterEdit();
  TypeDef*	DA_ViewSpecType(); // #IGNORE get type of DataArray for this log

  void	Initialize();
  void	Destroy()	{ CutLinks(); }
  void	InitLinks();
  void	CutLinks();
  TA_BASEFUNS(NetLogView);
};

class DTEditor;		// #IGNORE

class GridLogView : public LogView {
  // displays log information as colored grids
public:
  DT_GridViewSpec::BlockFill	fill_type;
  // how the grid blocks are filled in to display their value
  ColorScaleSpec* colorspec; 	// The color spectrum for this display
  int		block_size;	// *maximum* block size -- blocks will be smaller if needed to fit
  int		block_border_size; // size of the border around the blocks
  bool		header_on;	// is the log header visible?
  bool		auto_scale;	// whether to auto-scale on values or not
  FontSpec	view_font;	// the font to use for the labels in the display
  MinMax        scale_range;	// #HIDDEN range of scalebar
  MinMaxInt	actual_range;	// #HIDDEN #NO_SAVE range in actual lines of data
  
  // gui implementation
  DTEditor* 	editor;		// #IGNORE
  ivButton*	auto_sc_but;	// #IGNORE toggle auto-scale button
  ivButton*	head_tog_but;	// #IGNORE header toggle button
  ivAction*	adjuster;	// #IGNORE callback for the auto-scale button action

  void  	NewHead();
  void  	NewData();
  void		UpdateFromBuffer();
  void 		Clear_impl();
  
  TypeDef*	DT_ViewSpecType(); // #IGNORE get type of DataTable for this log
  TypeDef*	DA_ViewSpecType(); // #IGNORE get type of DataArray for this log
  
  // winview stuff
  void 		GetBodyRep();		// #IGNORE
  void		UpdateDisplay(TAPtr updtr=NULL);
  void 		InitDisplay();
  void 		CloseWindow();
  ivGlyph*	GetPrintData();	// #IGNORE
  
  virtual void	SetBlockFill(DT_GridViewSpec::BlockFill fill_typ=DT_GridViewSpec::COLOR);
  // #MENU #MENU_ON_Actions #MENU_SEP_BEFORE #ARGC_1 Set the fill style of the grid blocks
  virtual void	UpdateGridLayout(DT_GridViewSpec::MatrixLayout grid_layout=DT_GridViewSpec::LFT_RGT_BOT_TOP);
  // #MENU #ARGC_1 arrange columns to fit without gaps, according to geometry

  virtual void	AllBlockTextOn();
  // #MENU #MENU_SEP_BEFORE turn text on for all block displayed items
  virtual void	AllBlockTextOff();
  // #MENU turn text off for all block displayed items

  virtual void	ToggleHeader();  // toggle header on or off
  virtual void	ToggleAutoScale();  // toggle header on or off

  // gui implementation
  void		CreateLogButtons(); // #IGNORE
  ivGlyph*	GetLogButtons(); // #IGNORE

  void 	UpdateAfterEdit();
  void  InitLinks();
  void  CutLinks();
  void	Initialize();
  void 	Destroy();
  void	Copy_(const GridLogView& cp);
  COPY_FUNS(GridLogView, LogView);
  TA_BASEFUNS(GridLogView);
};

class GraphEditor;		// #IGNORE

class GraphLogViewLabel : public ViewLabel {
  // view labels for graph logs
public:
  void		GetMasterViewer();

  void	Initialize()	{ };
  void	Destroy()	{ };
  TA_BASEFUNS(GraphLogViewLabel);
};

class GraphLogView : public LogView {
  // View log data as a graph of lines
public:
  int			x_axis_index;	// index of index in array
  GraphEditor*		editor;		// #IGNORE handles all the display stuff
  ViewLabel_List	labels;	// misc labels in the graph view

  void		UpdateLineFeatures(); 
  // #MENU #MENU_ON_Actions #MENU_SEP_BEFORE update color, line type, point type, etc of lines
  void		SetLineFeatures
  (DT_GraphViewSpec::ColorType color_type,
   DT_GraphViewSpec::SequenceType sequence1=DT_GraphViewSpec::COLORS,
   DT_GraphViewSpec::SequenceType sequence2=DT_GraphViewSpec::LINES,
   DT_GraphViewSpec::SequenceType sequence3=DT_GraphViewSpec::POINTS,
   bool update_only_visible=true);
  // #MENU set color, line type, point type of lines by cycling through values

  void 		SetBackground(RGBA background);
  // #MENU set the display's background to the given color
  void  	NewHead();
  void  	NewData();

  void		UpdateFromBuffer();
  void 		Clear_impl();

  virtual int	SetXAxis(char* nm); // set the x axis to be this field
  virtual void	UpdateViewRange();  // update view range info from log

  TypeDef*	DT_ViewSpecType(); // #IGNORE get type of DataTable for this log
  TypeDef*	DA_ViewSpecType(); // #IGNORE get type of DataArray for this log

  void		InitColors();
  void 		GetBodyRep();	// #IGNORE
  void		CloseWindow();
  void 		UpdateDisplay(TAPtr updtr=NULL);
  void 		InitDisplay();

  ivGlyph*	GetPrintData();	// #IGNORE

  void 	UpdateAfterEdit();
  void	Initialize();
  void  Destroy();
  void	CutLinks();
  void 	InitLinks();
  void	Copy_(const GraphLogView& cp);
  COPY_FUNS(GraphLogView, LogView);
  TA_BASEFUNS(GraphLogView);
};

#endif // pdplog.h

