/* -*- C++ -*- */
/*=============================================================================
//									      //
// This file is part of the PDP++ software package.			      //
//									      //
// Copyright (C) 1995 Randall C. O'Reilly, Chadley K. Dawson, 		      //
//		      James L. McClelland, and Carnegie Mellon University     //
//     									      //
// Permission to use, copy, and modify this software and its documentation    //
// for any purpose other than distribution-for-profit is hereby granted	      //
// without fee, provided that the above copyright notice and this permission  //
// notice appear in all copies of the software and related documentation.     //
//									      //
// Permission to distribute the software or modified or extended versions     //
// thereof on a not-for-profit basis is explicitly granted, under the above   //
// conditions. 	HOWEVER, THE RIGHT TO DISTRIBUTE THE SOFTWARE OR MODIFIED OR  //
// EXTENDED VERSIONS THEREOF FOR PROFIT IS *NOT* GRANTED EXCEPT BY PRIOR      //
// ARRANGEMENT AND WRITTEN CONSENT OF THE COPYRIGHT HOLDERS.                  //
// 									      //
// Note that the taString class, which is derived from the GNU String class,  //
// is Copyright (C) 1988 Free Software Foundation, written by Doug Lea, and   //
// is covered by the GNU General Public License, see ta_string.h.             //
// The iv_graphic library and some iv_misc classes were derived from the      //
// InterViews morpher example and other InterViews code, which is             //
// Copyright (C) 1987, 1988, 1989, 1990, 1991 Stanford University             //
// Copyright (C) 1991 Silicon Graphics, Inc.				      //
//									      //
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,         //
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 	      //
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  	      //
// 									      //
// IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE FOR ANY SPECIAL,    //
// INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES  //
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT     //
// ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY,      //
// ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS        //
// SOFTWARE. 								      //
==============================================================================*/

#ifndef log_iv_h
#define log_iv_h
#include <pdplog.h>

// these contains the graphics part of their pdplog counterparts

class ivGlyph;
class DragPage;
class AdjustWrapper;

class LogView_IV_Impl {
public:
  LogView*		log_view;
  ivLayoutKit*		layout;
  ivWidgetKit*		wkit;

  ivButton*		ckbox;

  virtual void  InitDisplay()      {};
  virtual void	UpdateDisplay()    {}; // reallocate and redraw
  virtual void	UpdateFromBuffer() {}; // more extensive UpdateDisplay

  virtual void		SetToggle(bool onoff);
  virtual void		ToggleDisplay();
  virtual ivGlyph*	GetLogButtons();
  virtual ivGlyph*	GetBodyRep() {return NULL;}
  virtual ivGlyph*	GetPrintData(){return log_view->win_box;}
  LogView_IV_Impl(LogView* view);

  virtual ~LogView_IV_Impl();
};


class TextLogView_IV_Impl : public LogView_IV_Impl {
public:
  ivPolyGlyph*  sb;		// scrollbox for screenlog
  ivGlyph*	h_line;		// headerline for screenlog
  ivPatch*	patch;		// update patch;
  ivPatch*	scrb;		// update scrollable;
  ivPolyGlyph*	vbox;		// vbox with headerline and scrollbox
  ivPolyGlyph*	ovbox;		// outer vbox with buttons and vbox
  float		font_width;	//  width of font spacing
  ivPolyGlyph*  cur_line;	// current Line we are creating

  ivGlyph*	GetBodyRep();
  ivGlyph*	GetPrintData();

  void		UpdateDisplay(); // reallocate and redraw
  void		UpdateFromBuffer(); // more extensive UpdateDisplay

  virtual void	AddToLine(const char* name, ivCoord width); // add name of width to line
  virtual void	AddLine();			  // add line to scrollbox
  virtual void  RemoveLine(int index);		  // remove line from scrollbox, -1 = all

  virtual void  StartHeaderLine();
  virtual void  AddToHeaderLine(const char* name, ivCoord width);
  virtual void  EndHeaderLine();

  TextLogView_IV_Impl(LogView * view);
  ~TextLogView_IV_Impl();
};

class PageLogView_IV_Impl : public LogView_IV_Impl {
public:
  ivPolyGlyph*  sb;		// scrollbox data
  ivPolyGlyph*  tmp;		/// temp storage area
  DragPage*	h_page;		// header page for headers
  AdjustWrapper* h_page_wrap;	// header page wrapepr
  ivPatch*	patch;		// update patch;
  ivPatch*	scrb;		// update scrollable;
  ivPolyGlyph*	vbox;		// vbox with headerline and scrollbox
  ivPolyGlyph*	ovbox;		// outer vbox with buttons and vbox
  float		font_width;	//  width of font spacing
  ivPolyGlyph*  cur_line;	// current Line we are creating

  ivGlyph*	GetBodyRep();
  ivGlyph*	GetPrintData();

  void		UpdateDisplay(); // reallocate and redraw
  void		UpdateFromBuffer(); // more extensive UpdateDisplay

  virtual void	AddToLine(const char* name,ivCoord x, ivCoord y);       // add name to line
  virtual void	AddLine();			  // add line to scrollbox
  virtual void  RemoveLine(int index);		  // remove line from scrollbox, -1 = all

  virtual void  StartHeaderPage();
  virtual void  AddToHeaderPage(const char* name,ivCoord x, ivCoord y);
  virtual void  EndHeaderPage();

  virtual void  ToggleEdit();
  virtual void  AutoLayout();
  virtual void  GetPosition(int i, ivCoord& x, ivCoord& y);

  PageLogView_IV_Impl(LogView * view);
  ~PageLogView_IV_Impl();
};

class GridLogView_IV_Impl : public LogView_IV_Impl {
public:
  ivAction*	adjuster;	// 
  ivGlyph*	data_rep;	// main data representation
  ivPatch*	data_patch;	// patch to hold data (gets updated)

  ivGlyph*	GetBodyRep();
  ivGlyph*	GetPrintData();

  void		InitDisplay();
  void		UpdateDisplay();

  virtual void	ScaleOnAllBlocks(); // Toggle all block header labels on
  virtual void	CreateAdjuster();

  GridLogView_IV_Impl(LogView * view);
  ~GridLogView_IV_Impl();
  
};

class GraphLogView_IV_Impl : public LogView_IV_Impl {
public:
  ivGlyph* 	GetBodyRep();
  ivGlyph*	GetPrintData();

  GraphLogView_IV_Impl(LogView * view);
  ~GraphLogView_IV_Impl() {}
};

#endif // log_iv_h
