/* -*- C++ -*- */
/*=============================================================================
//									      //
// This file is part of the PDP++ software package.			      //
//									      //
// Copyright (C) 1995 Randall C. O'Reilly, Chadley K. Dawson, 		      //
//		      James L. McClelland, and Carnegie Mellon University     //
//     									      //
// Permission to use, copy, and modify this software and its documentation    //
// for any purpose other than distribution-for-profit is hereby granted	      //
// without fee, provided that the above copyright notice and this permission  //
// notice appear in all copies of the software and related documentation.     //
//									      //
// Permission to distribute the software or modified or extended versions     //
// thereof on a not-for-profit basis is explicitly granted, under the above   //
// conditions. 	HOWEVER, THE RIGHT TO DISTRIBUTE THE SOFTWARE OR MODIFIED OR  //
// EXTENDED VERSIONS THEREOF FOR PROFIT IS *NOT* GRANTED EXCEPT BY PRIOR      //
// ARRANGEMENT AND WRITTEN CONSENT OF THE COPYRIGHT HOLDERS.                  //
// 									      //
// Note that the taString class, which is derived from the GNU String class,  //
// is Copyright (C) 1988 Free Software Foundation, written by Doug Lea, and   //
// is covered by the GNU General Public License, see ta_string.h.             //
// The iv_graphic library and some iv_misc classes were derived from the      //
// InterViews morpher example and other InterViews code, which is             //
// Copyright (C) 1987, 1988, 1989, 1990, 1991 Stanford University             //
// Copyright (C) 1991 Silicon Graphics, Inc.				      //
//									      //
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,         //
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 	      //
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  	      //
// 									      //
// IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE FOR ANY SPECIAL,    //
// INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES  //
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT     //
// ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY,      //
// ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS        //
// SOFTWARE. 								      //
==============================================================================*/

// PDPShell based on C^c C super script language

#include <pdp/pdpshell.h>
#include <pdp/sched_proc.h>
#include <pdp/net_iv.h>
#include <pdp/pdp_iv.h>
// some functions are instantited here for these extra procs
#include <pdp/procs_extra.h>
#include <ta/taiv_type.h>
#include <ta/ta_dump.h>
#include <css/ta_css.h>
#include <css/css_iv.h>
#include <css/css_builtin.h>
#include <signal.h>
#include <memory.h>
#include <malloc.h>
#include <unistd.h>
#include <strstream.h>

#include <ta/enter_iv.h>
#include <IV-look/kit.h>
#include <IV-look/dialogs.h>
#include <InterViews/window.h>
#include <InterViews/layout.h>
#include <InterViews/style.h>
#include <InterViews/action.h>
#include <InterViews/event.h>
#include <InterViews/patch.h>
#include <InterViews/stencil.h>
#include <InterViews/background.h>
#ifndef CYGWIN
#include <IV-X11/xwindow.h>		// this is for the window title hack
#endif
#include <InterViews/bitmap.h>
#include <ta/leave_iv.h>


static ivOptionDesc PDP_options[] = {
    { NULL }
};

static ivPropertyData PDP_defs[] = {
  {"PDP++*gui", "sgimotif"},
  {"PDP++*PopupWindow*overlay", "true"},
  {"PDP++*PopupWindow*saveUnder", "on"},
  {"PDP++*TransientWindow*saveUnder", "on"},
  {"PDP++*double_buffered",	"on"},
  {"PDP++*flat",		"#c0c4d3"},
  {"PDP++*background",  	"#70c0d8"},
  {"PDP++*name*flat",		"#70c0d8"},
  {"PDP++*apply_button*flat",	"#c090b0"},
  {"PDP++*FieldEditor*background", "white"},
  {"PDP++*FileChooser*filter", 	"on"},
  {"PDP++*FileChooser.rows", 	"20"},
  {"PDP++*FileChooser.width", 	"100"},
  {"PDP++*PaletteButton*minimumWidth", "72.0"},
  {"PDP++*PushButton*minimumWidth", "72.0"},
  {"PDP++*TaIVButton*SmallWidth", "46.0"},
  {"PDP++*TaIVButton*MediumWidth", "72.0"},
  {"PDP++*TaIVButton*BigWidth", "115.0"},
  {"PDP++*toggleScale",		"1.5"},
#ifndef CYGWIN
  {"PDP++*font",		"*-helvetica-medium-r-*-*-10*"},
  {"PDP++*name*font",		"*-helvetica-medium-r-*-*-10*"},
  {"PDP++*title*font",		"*-helvetica-bold-r-*-*-10*"},
  {"PDP++*small_menu*font",	"*-helvetica-medium-r-*-*-10*"},
  {"PDP++*small_submenu*font",	"*-helvetica-medium-r-*-*-10*"},
  {"PDP++*big_menu*font",	"*-helvetica-medium-r-*-*-12*"},
  {"PDP++*big_submenu*font",	"*-helvetica-medium-r-*-*-12*"},
  {"PDP++*big_menubar*font",	"*-helvetica-bold-r-*-*-14*"},
  {"PDP++*big_italic_menubar*font","*-helvetica-bold-o-*-*-14*"},
#else
  {"PDP++*font",		"*Arial*medium*--12*"},
  {"PDP++*name*font",		"*Arial*medium*--12*"},
  {"PDP++*title*font",		"*Arial*bold*--12*"},
  {"PDP++*small_menu*font",	"*Arial*medium*--12*"},
  {"PDP++*small_submenu*font",	"*Arial*medium*--12*"},
  {"PDP++*big_menu*font",	"*Arial*medium*--12*"},
  {"PDP++*big_submenu*font",	"*Arial*medium*--12*"},
  {"PDP++*big_menubar*font",	"*Arial*bold*--14*"},
  {"PDP++*big_italic_menubar*font","*Arial*italic*--14*"},
  // following are def'd in smf_kit.cpp
  {"PDP++*MenuBar*font", 	"*Arial*bold*--12*"},
  {"PDP++*MenuItem*font", 	"*Arial*bold*--12*"},
  {"PDP++*MenuBar*font", 	"*Arial*bold*--12*"},
  {"PDP++*MenuItem*font", 	"*Arial*bold*--12*"},
#endif
  { NULL } 
};


//////////////////////////////////////////////////////////////////////////////////
//     	Functions from other files which reference Project or root in passing	//
//////////////////////////////////////////////////////////////////////////////////

//////////////////////////
//	spec.h		//
//////////////////////////

BaseSpec_MGroup* SpecPtr_impl::GetSpecGroup() {
  Project* prj = GET_OWNER(owner,Project);
  if(prj == NULL)
    return NULL;
  return &(prj->specs);
}

//////////////////////////
//	netstru.h	//
//////////////////////////


void ConSpec::InitWeights() {
  Project* prj = (Project *) GET_MY_OWNER(Project);
  if (prj == NULL) return;
  Network* net;
  taLeafItr ni;
  FOR_ITR_EL(Network, net, prj->networks., ni){
    net->InitWtState();
    net->UpdateAllViews();
  }
}

void UnitSpec::UpdateAfterEdit() {
  BaseSpec::UpdateAfterEdit();
  if((bias_con_type != NULL) && (bias_spec.spec != NULL) && !bias_con_type->InheritsFrom(bias_spec.spec->min_con_type)) {
    taMisc::Error("Bias con type of:", bias_con_type->name,
		  "is not of the correct type for the bias con spec,"
		  "which needs at least a:", bias_spec.spec->min_con_type->name);
    return;
  }
  if(taMisc::is_loading || !taMisc::iv_active)	return;
  Project* prj = (Project *) GET_MY_OWNER(Project);
  if (prj == NULL) return;
  Network* net;
  taLeafItr ni;
  FOR_ITR_EL(Network, net, prj->networks., ni) {
    NetView* vw;
    taLeafItr vi;
    FOR_ITR_EL(NetView, vw, net->views., vi) {
      vw->UpdateButtons();	// update buttons to reflect need to build or not..
    }
  }
}

void UnitSpec::BuildBiasCons() {
  Project* prj = (Project *) GET_MY_OWNER(Project);
  if (prj == NULL) return;
  Network* net;
  taLeafItr ni;
  FOR_ITR_EL(Network, net, prj->networks., ni){
    net->Build();
  }
}

void ConSpec::ReplacePointersHook(TAPtr old) {
  Project* prj = (Project *) GET_MY_OWNER(Project);
  if (prj == NULL) return;
  winbMisc::DelayedMenuUpdate(prj);
  Network* net;
  taLeafItr ni;
  FOR_ITR_EL(Network, net, prj->networks., ni){
    net->ReplaceConSpec((ConSpec*)old, this);
  }

  // now go through specs!
  BaseSpec* sp;
  taLeafItr si;
  FOR_ITR_EL(BaseSpec, sp, prj->specs., si) {
    if(!sp->InheritsFrom(TA_UnitSpec)) continue;
    UnitSpec* us = (UnitSpec*)sp;
    if(us->bias_spec.spec == old) {
      us->bias_spec.SetSpec(this); // update to new
    }
  }
}

void UnitSpec::ReplacePointersHook(TAPtr old) {
  Project* prj = (Project *) GET_MY_OWNER(Project);
  if (prj == NULL) return;
  winbMisc::DelayedMenuUpdate(prj);
  Network* net;
  taLeafItr ni;
  FOR_ITR_EL(Network, net, prj->networks., ni){
    net->ReplaceUnitSpec((UnitSpec*)old, this);
  }
}

void LayerSpec::ReplacePointersHook(TAPtr old) {
  Project* prj = (Project *) GET_MY_OWNER(Project);
  if (prj == NULL) return;
  winbMisc::DelayedMenuUpdate(prj);
  Network* net;
  taLeafItr ni;
  FOR_ITR_EL(Network, net, prj->networks., ni){
    net->ReplaceLayerSpec((LayerSpec*)old, this);
  }
}

void ProjectionSpec::ReplacePointersHook(TAPtr old) {
  Project* prj = (Project *) GET_MY_OWNER(Project);
  if (prj == NULL) return;
  winbMisc::DelayedMenuUpdate(prj);
  Network* net;
  taLeafItr ni;
  FOR_ITR_EL(Network, net, prj->networks., ni){
    net->ReplacePrjnSpec((ProjectionSpec*)old, this);
  }
}

void Projection::GridViewWeights(GridLog* grid_log, int un_x, int un_y, int wt_x, int wt_y) {
  if(grid_log == NULL) {
    Project* proj = GET_MY_OWNER(Project);
    if(proj == NULL)    return;
    grid_log = (GridLog*) proj->logs.NewEl(1, &TA_GridLog);
    winbMisc::DelayedMenuUpdate(proj);
  }
  else {
    grid_log->Clear();
  }
  grid_log->name = (String)"GridViewWeights: " + layer->name + ", " + name;

  DataTable* md = &(grid_log->data);
  md->RemoveAll();
  
  DataArray_impl* da = md->NewEl(1, &TA_float_Data);
  da->name = "$row";

  int rx = (un_x < 0) ? layer->act_geom.x : un_x;
  int ry = (un_y < 0) ? layer->act_geom.y : un_y;

  int wtx = (wt_x < 0) ? from->act_geom.x : wt_x;
  int wty = (wt_y < 0) ? from->act_geom.y : wt_y;

  int totn = wtx * wty;

  // these are the columns
  int i;
  for(i=0; i<rx; i++) {
    DataTable* dt = (DataTable*)md->NewGp(1);
    dt->name = (String)"c" + (String)i;
    dt->EnforceSize(totn);
    dt->FastEl(0)->name = (String)"<" + (String)totn + ">c0";
    int j;
    for(j=1; j<totn; j++)
      dt->FastEl(j)->name = (String)"c" + (String)j;
  }

  md->ResetData();

  for(i=0;i<ry;i++) {
    md->AddBlankRow();		// add the rows
    ((float_Data*)md->FastEl(0))->ar->FastEl(i) = ry - i -1;	// label in reverse order
  }

  int uni = 0;
  int ux, uy;
  for(uy=0;uy<ry; uy++) {
    if(uni >= layer->units.leaves)
      break;
    for(ux=0; ux<layer->act_geom.x; ux++, uni++) {
      if(uni >= layer->units.leaves)
	break;
      if(ux >= rx) {
	uni += layer->act_geom.x - rx; break;
      }
      DataTable* dt = (DataTable*)md->gp.FastEl(ux);
      Unit* un = (Unit*)layer->units.Leaf(uni);
      Con_Group* cg = un->recv.FindFrom(from);
      if(cg == NULL)
	break;
      int wi;
      for(wi=0;wi<cg->size;wi++) {
	Unit* su = cg->Un(wi);
	Unit_Group* ownr = ((Unit_Group*)su->owner);
	int sx, sy;
	sx = su->pos.x + ownr->pos.x; 
	sy = su->pos.y + ownr->pos.y;
	if((sx >= wtx) || (sy >= wty)) continue;
	int sidx = (sy * wtx) + sx;
	((float_Data*)dt->FastEl(sidx))->ar->FastEl(ry-uy-1) = cg->Cn(wi)->wt;
      }
    }
  }
  
  DT_GridViewSpec* vspc = (DT_GridViewSpec*)((GridLogView*)grid_log->views[0])->viewspec;
  vspc->RemoveAll();
  vspc->BuildFromDataTable();

  for(i=0; i<rx; i++) {
    DT_GridViewSpec* da = (DT_GridViewSpec*)vspc->gp[i];
    da->geom.x = wtx; 
    da->geom.y = wty;
    da->name = (String)"c" + (String)i;
    da->display_name = (String)"c" + (String)i;
  }
  ((DA_GridViewSpec*)vspc->FastEl(0))->display_style = DA_GridViewSpec::TEXT;
  vspc->UpdateLayout();
  grid_log->InitAllViews();
  grid_log->data_range.max = ry-1;
  grid_log->InitAllViews();
  ((GridLogView*)grid_log->views[0])->View_FR();
}

void Network::RemoveMonitors() {
  if(proj == NULL) return;
  taLeafItr pi;
  SchedProcess* sp;
  FOR_ITR_EL(SchedProcess, sp, proj->processes., pi) {
    taLeafItr si;
    Stat* st;
    FOR_ITR_EL(Stat, st, sp->loop_stats., si) {
      if(st->InheritsFrom(TA_MonitorStat))
	((MonitorStat*)st)->ptrs.RemoveAll();
    }
    FOR_ITR_EL(Stat, st, sp->final_stats., si) {
      if(st->InheritsFrom(TA_MonitorStat))
	((MonitorStat*)st)->ptrs.RemoveAll();
    }
  }
}

void Network::UpdateMonitors() {
  if(proj == NULL) return;
  taLeafItr pi;
  SchedProcess* sp;
  FOR_ITR_EL(SchedProcess, sp, proj->processes., pi) {
    taLeafItr si;
    Stat* st;
    FOR_ITR_EL(Stat, st, sp->loop_stats., si) {
      if(st->InheritsFrom(TA_MonitorStat))
	((MonitorStat*)st)->UpdateAfterEdit();
    }
    FOR_ITR_EL(Stat, st, sp->final_stats., si) {
      if(st->InheritsFrom(TA_MonitorStat))
	((MonitorStat*)st)->UpdateAfterEdit();
    }
  }
}

// cfront requires this to be outside class function
enum NetSection {NS_NONE, NS_DEFINITIONS, NS_CONSTRAINTS,
		   NS_NETWORK, NS_BIASES};

ConSpec* GetNSConSpec(char name,BaseSpec_MGroup * bsmg,
		      BaseSpec_MGroup* master, TypeDef* conspec_type,
		      bool skip_dots=true){
  if((skip_dots == true) && (name == '.')) return NULL;
  int i;
  ConSpec* result;
  for(i=0;i<bsmg->leaves;i++){
    result = (ConSpec *) bsmg->Leaf(i);
    if(!result->name.empty() && (result->name[0] == name)) return result;
  }
  if((name == 'r') || (name == 'p') ||
     (name == 'n') || (name == '.')) { // pre-defined specs
    result  = (ConSpec *) master->New(1,conspec_type);
    bsmg->Link(result);
    result->name = name;
    result->rnd.type = Random::UNIFORM;
    if(name == 'r') {result->rnd.mean = 0.0; result->rnd.var = .5;}
    else if(name == 'p') {result->rnd.mean = 0.5; result->rnd.var = .5;}
    else if(name == 'n') {result->rnd.mean = -0.5; result->rnd.var = .5;}
    else if(name == '.') {result->rnd.mean = 0; result->rnd.var = 0;}
    return result;
  }
  // search for uppercase version
  for(i=0;i<bsmg->leaves;i++){
    result = (ConSpec *) bsmg->Leaf(i);
    String upnm = result->name;
    upnm.upcase();
    if(!upnm.empty() && (upnm[0] == name)){
      master->DuplicateEl(result);
      result = (ConSpec *) master->Leaf(master->leaves-1);
      result->name = name;
    }
    // todo set lrate or something to 0 to make upcase
    // version unmodifiable
  }
  taMisc::Error("Connection Specification letter \"", String(name) , "\" not found");
  return NULL;
}


void Network::ReadOldPDPNet(istream& strm, bool skip_dots){
  NetSection ns = NS_NONE;
  int nunits= 0;
  int ninputs = 0;
  int nhiddens = 0;
  int noutputs= 0;
  Unit_Group flat_units;
  BaseSpec_MGroup conspec_group;
  int send_start = 0;
  int send_end = 0;
  int recv_start = 0;
  int recv_end = 0;
  int cur_send = 0;
  int cur_recv = 0;
  char con_char = '\0';
  Project* prj = (Project *) GET_MY_OWNER(Project);
  Layer* lay = (Layer *) layers.New(1);
  Projection* prjn = (Projection *) lay->projections.New(1);
  TypeDef* conspec_type = prjn->con_spec.type;
  lay->projections.Remove(prjn);
  layers.Remove(lay);

  strm.seekg(0);
  while(strm.good() && !strm.eof()){
    int c = taMisc::read_alnum(strm);
    if((c == EOF) || taMisc::LexBuf.empty())
      continue;
    int fc = taMisc::LexBuf.firstchar();
    // skip comments
    if((fc == '#') || (taMisc::LexBuf.before(2) == "//")) {
      taMisc::read_till_eol(strm);
      continue;
    }
    if(ns == NS_NONE){
      if(taMisc::LexBuf == "") continue; // not a section line
      taMisc::LexBuf.upcase();
      if(taMisc::LexBuf == "DEFINITIONS:") { ns = NS_DEFINITIONS; continue;}
      if(taMisc::LexBuf == "CONSTRAINTS:") { ns = NS_CONSTRAINTS; continue;}
      if(taMisc::LexBuf == "NETWORK:") {
	ns = NS_NETWORK;
	// default fully-connected weight configuration
	send_end = recv_end = nunits;
	continue;
      }
      if(taMisc::LexBuf == "BIASES:") { ns = NS_BIASES; continue;}
    }
    else if( ns == NS_DEFINITIONS) {
      if(upcase(taMisc::LexBuf) == "NUNITS") {
	c = taMisc::read_alnum(strm);
	if((c == EOF) || taMisc::LexBuf.empty()){
	  taMisc::Error("Unspecified number of units in OldPDPNet file");
	  return;
	}

	// this is how many units we should have

	nunits = (int) taMisc::LexBuf;
	continue;
      }
      if(upcase(taMisc::LexBuf) == "NINPUTS") {
	c = taMisc::read_alnum(strm);
	if((c == EOF) || taMisc::LexBuf.empty()){
	  taMisc::Error("Unspecified number of inputs in OldPDPNet file");
	  return;
	}

	// this is how many input units we should have

	ninputs = (int) taMisc::LexBuf;
	continue;
      }
      if(upcase(taMisc::LexBuf) == "NOUTPUTS") {
	c = taMisc::read_alnum(strm);
	if((c == EOF) || taMisc::LexBuf.empty()){
	  taMisc::Error("Unspecified number of outputs in OldPDPNet file");
	  return;
	}

	// this is how many output units we should have

	noutputs = (int) taMisc::LexBuf;
	continue;
      }

      if(taMisc::LexBuf == "end") {
	ns = NS_NONE;
	if((ninputs + noutputs) > nunits) {
	  taMisc::Error("ninputs + noutputs > noutputs in OldPDPNet file");
	  return;
	}
	Layer* tlay;
	taLeafItr ti;
	FOR_ITR_EL(Layer, tlay, layers., ti){
	  flat_units.Borrow(tlay->units);
	}
	if(flat_units.leaves > 0) {
	  if(flat_units.leaves != nunits){
	    String nun = (int) nunits;
	      taMisc::Error("Existing network structure does not have ",
			    nun, " units");
	    flat_units.Reset();
	    return;
	  }
	  else { // use existing network;
	    nhiddens = ninputs = noutputs = 0;
	  }
	}
	else {
	  if(ninputs + noutputs != 0)
	    nhiddens = nunits - ninputs - noutputs;
	}
	Layer* input_layer;
	Layer* output_layer;
	Layer* hidden_layer;
	if((ninputs + noutputs + nhiddens == 0) &&
	   (flat_units.leaves == 0) && (nunits > 0)){
	  // we need some units, all in the same layer
	  input_layer = (Layer *) layers.New(1);
	  input_layer->n_units = nunits;
	  input_layer->Build();
	  flat_units.Borrow(input_layer->units);
	  continue;
	}
	if(ninputs > 0) { // user wants an inputs layer, make it the first
	  if(layers.leaves > 0)
	    input_layer = (Layer *) layers.Leaf(0);
	  else {
	    input_layer = (Layer *) layers.New(1);
	    input_layer->name = "Input Layer";
	  }
	  input_layer->n_units = ninputs;
	  input_layer->Build();
	}
	if(nhiddens > 0) { // user wants an hidden layer, make it the second
			  // or first if no input layer
	  if(layers.leaves > (1 - (ninputs == 0)))
	    hidden_layer = (Layer *) layers.Leaf(1 - (ninputs ==0));
	  else {
	    hidden_layer = (Layer *) layers.New(1);
	    hidden_layer->name = "Hidden Layer";
	  }
	  hidden_layer->n_units = nhiddens;
	  hidden_layer->Build();
	}
	if(noutputs > 0) { // user wants an inputs layer, make it the last
			  // but not the hidden layer!
	  if(layers.leaves > ((ninputs != 0) + (nhiddens != 0)))
	    output_layer = (Layer *) layers.Leaf(layers.leaves-1);
	  else {
	    output_layer = (Layer *) layers.New(1);
	    output_layer->name = "Output Layer";
	  }
	  output_layer->n_units = noutputs;
	  output_layer->Build();
	}

	Layer* lay; flat_units.Reset();
	taLeafItr i;
	FOR_ITR_EL(Layer, lay, layers., i){
	  flat_units.Borrow(lay->units);
	}
      }
    }
    else if (ns == NS_NETWORK) {
      if(taMisc::LexBuf == "end") {
	ns = NS_NONE;
	continue;
      }
      if (fc == '%') { // special case
	if(taMisc::LexBuf.length() == 2) // same contype for whole range
	  con_char = taMisc::LexBuf[1];
	c = taMisc::read_alnum_noeol(strm);
        if ((c == EOF) || taMisc::LexBuf.empty()) {
	  taMisc::Error("Improper connection range specification in network file");
	  return;
	}
	recv_start = (int)taMisc::LexBuf;
	c = taMisc::read_alnum_noeol(strm);
        if ((c == EOF) || taMisc::LexBuf.empty()) {
	  taMisc::Error("Improper connection range specification in network file");
	  return;
	}
	recv_end  = recv_start + ( (int) taMisc::LexBuf) -1;
	c = taMisc::read_alnum_noeol(strm);
        if ((c == EOF) ||  taMisc::LexBuf.empty()) {
	  taMisc::Error("Improper connection range specification in network file");
	  return;
	}
	send_start = (int)taMisc::LexBuf;
	c = taMisc::read_alnum_noeol(strm);
        if ((c == EOF) || taMisc::LexBuf.empty()) {
	  taMisc::Error("Improper connection range specification in network file");
	  return;
	}
	send_end  = send_start + ( (int) taMisc::LexBuf) -1;

	if(con_char != '\0') { // same type on con for all of range
	  ConSpec* conspec = GetNSConSpec(con_char,&conspec_group,
					  &prj->specs,conspec_type,skip_dots);
	  if (conspec != NULL) 
	    for(int j=recv_start;j<=recv_end;j++){
	      for(int i =send_start;i<=send_end;i++){
		ConnectUnits(flat_units.Leaf(j),flat_units.Leaf(i),
			     false,conspec);
	      }
	    }
	  con_char = '\0';
	  continue;
	}
        // otherwise prepare to begin a block
	cur_send = send_start;
	cur_recv = recv_start;
	continue;
      }
      // otherwise it must be a line of connections
      // check to make sure that line is send_end - send_start chars in length
      // step cur_send through line for each char

      if((int)taMisc::LexBuf.length() != (send_end - send_start)) {
	String curlen = (int) taMisc::LexBuf.length();
	String expt = send_end - send_start;
	taMisc::Error("Connection Line:", taMisc::LexBuf, " has ",
		      curlen, "sending weights. (expected ", expt, ")");
	return;
      }
      if(cur_recv > recv_end) {
	String cur = cur_recv;
	String expt = recv_end;
	taMisc::Error("Too many connection lines:", cur, 
		      "than expected number:", expt);
	return;
      }
      Unit* recv_unit = flat_units.Leaf(cur_recv);
      Layer* recv_layer = GET_OWNER(recv_unit,Layer);
      for(int i=0;i<(send_end-send_start);i++){
	con_char = taMisc::LexBuf[i];
	ConSpec* conspec = GetNSConSpec(con_char, &conspec_group,
					&prj->specs,conspec_type);
	if(conspec == NULL) continue;
	Unit* send_unit = (Unit *) flat_units.Leaf(i);
	Layer* send_layer = GET_OWNER(send_unit,Layer);
	Projection* pjn = NULL;
	taLeafItr p;
	// check to see if a pjrn already exists
	FOR_ITR_EL(Projection, pjn, recv_layer->projections., p) {
	  if((pjn->from == send_layer) && 
	     (pjn->spec->InheritsFrom(&TA_CustomPrjnSpec)) &&
	     (pjn->con_spec == conspec))
	    break; // ok found one
	}
	if(pjn==NULL) { // no projection
	  pjn = (Projection *) recv_layer->projections.New(1);
	  if(recv_layer == send_layer) // self con
	    pjn->from_type = Projection::SELF;
	  else
	    pjn->from_type = Projection::CUSTOM;

	  pjn->spec.type = &TA_CustomPrjnSpec;
	  pjn->spec.UpdateAfterEdit();
	  pjn->con_spec.SetSpec(conspec);
	  taBase::SetPointer((TAPtr*)&(pjn->from), send_layer);
	  pjn->projected = true;
	  pjn->UpdateAfterEdit();
	}
	pjn->send_idx = pjn->recv_idx = -1; // clear idicies
	// find the connection group on the units
	Con_Group* rgrp = NULL;
	Con_Group* sgrp = NULL;
	int rg,sg;
	FOR_ITR_GP(Con_Group,rgrp,recv_unit->recv.,rg){
	  if(rgrp->prjn == pjn) { pjn->recv_idx = rg; break;}
	}
	FOR_ITR_GP(Con_Group,sgrp,send_unit->send.,sg){
	  if(sgrp->prjn == pjn) { pjn->send_idx = sg; break;}
	}
	recv_unit->ConnectFrom(send_unit,pjn);
	cur_send++;
      }
      con_char = '\0';
      cur_send = send_start;
      cur_recv++;
      continue;
    }
    else if(ns == NS_BIASES) {
      if(taMisc::LexBuf == "end") {ns = NS_NONE;continue;}
    }
    else if (ns == NS_CONSTRAINTS) {
      if(taMisc::LexBuf == "end") {ns = NS_NONE;continue;}
      if(taMisc::LexBuf.length() > 1) { // constriant variable is not a char
	taMisc::Error("Constraint Character:", taMisc::LexBuf,
		      "was not a character");
	return;
      }
      ConSpec* nsc  = (ConSpec *) prj->specs.New(1,conspec_type);
      conspec_group.Link(nsc);
      nsc->name = taMisc::LexBuf;
      nsc->rnd.type = Random::UNIFORM;
      while((c!='\n') && (c != EOF) && !taMisc::LexBuf.empty()){
	c = taMisc::read_alnum_noeol(strm);
	taMisc::LexBuf.upcase();
	if(taMisc::LexBuf == "POSITIVE") {
	  nsc->wt_limits.type = WeightLimits::GT_MIN;
	  nsc->wt_limits.min = 0.0f;
	  continue;
	}
	if(taMisc::LexBuf == "NEGATIVE") {
	  nsc->wt_limits.type = WeightLimits::LT_MAX;
	  nsc->wt_limits.max = 0.0f;
	  continue;
	}
	if(taMisc::LexBuf == "LINKED") {
	  // todo linked
	}
	if(taMisc::LexBuf == "RANDOM") {
	  nsc->rnd.mean = 0;
	  nsc->rnd.var =  0.5;
	  continue;
	}
	// otherwise it must be a number
	nsc->rnd.var = 0.0;
	nsc->rnd.mean = (float) taMisc::LexBuf;
      }
    }
  }
  flat_units.Reset();
  UpdateAfterEdit();
}

//////////////////////////
//	process.h	//
//////////////////////////

void Process::SetDefaultPNEPtrs() {
  if(project == NULL)
    project = GET_MY_OWNER(Project);
  if(project == NULL)	return;

  if(network == NULL)
    taBase::SetPointer((TAPtr*)&network, project->networks.DefaultEl());
  if(environment == NULL)
    taBase::SetPointer((TAPtr*)&environment, project->environments.DefaultEl());
}

//////////////////////////
//	sched_proc.h	//
//////////////////////////

void SchedProcess::InitAllLogs() {
  if(project == NULL)
    return;
  taLeafItr i;
  PDPLog* lg;
  FOR_ITR_EL(PDPLog, lg, project->logs., i)
    lg->Clear();
}

void ForkProcess::SetDefaultPNEPtrs() {
  SchedProcess::SetDefaultPNEPtrs();
  if(project == NULL) return;
  if(second_network == NULL)
    taBase::SetPointer((TAPtr*)&second_network, project->networks.DefaultEl());
  if(second_environment == NULL)
    taBase::SetPointer((TAPtr*)&second_environment, project->environments.DefaultEl());
}

void SyncEpochProc::SetDefaultPNEPtrs() {
  EpochProcess::SetDefaultPNEPtrs();
  if(project == NULL) return;
  if(second_network == NULL)
    taBase::SetPointer((TAPtr*)&second_network, project->networks.DefaultEl());
}

//////////////////////////
//	enviro.h	//
//////////////////////////

Network* PatternSpec::GetDefaultNetwork() {
  Project* proj = GET_MY_OWNER(Project);
  if(proj == NULL)    return NULL;
  Network* net = (Network*)proj->networks.DefaultEl();
  return net;
}

Network* EnviroView::GetDefaultNetwork() {
  Project* proj = GET_MY_OWNER(Project);
  if(proj == NULL)    return NULL;
  Network* net = (Network*)proj->networks.DefaultEl();
  return net;
}

void Environment::ReplacePointersHook(TAPtr old) {
  Project* prj = (Project *) GET_MY_OWNER(Project);
  if (prj == NULL) return;
  winbMisc::DelayedMenuUpdate(prj);
  prj->processes.ReplaceEnvPtrs((Environment*)old, this);
}

void Environment::DistMatrixGrid(GridLog* disp_log, int pat_no, float_RArray::DistMetric metric,
			     bool norm, float tol)
{
  if(events.leaves == 0)
    return;

  if(disp_log == NULL) {
    Project* proj = GET_MY_OWNER(Project);
    if(proj == NULL)    return;
    disp_log = (GridLog*) proj->logs.NewEl(1, &TA_GridLog);
    winbMisc::DelayedMenuUpdate(proj);
  }

  disp_log->name = "DistMatrixGrid: " + name + " Pattern: " + String(pat_no);
  DataTable* dt = &(disp_log->data);

  dt->Reset();
  dt->NewEl(1, &TA_String_Data);
  dt->El(0)->name = "$Event";
  DataTable* dtgp = (DataTable*)dt->NewGp(1);

  bool first = true;
  taLeafItr i;
  Event* ev;
  FOR_ITR_EL(Event, ev, events., i) {
    String nm = ev->name;
    if(first) {
      nm = String("<") + String(events.leaves) + ">" + nm;
      first = false;
    }
    float_Data* da = (float_Data*)dtgp->NewEl(1, &TA_float_Data);
    da->name = nm;
  }
  dt->ResetData();

  int cnt = 0;			// index into data
  FOR_ITR_EL(Event, ev, events., i) {
    Pattern* trg_pat = (Pattern*)ev->patterns.Leaf(pat_no);
    if(trg_pat == NULL) {
      taMisc::Error("*** Pattern number:", String(pat_no), "not found");
      return;
    }
    ((String_Array*)dt->El(0)->AR())->Add(ev->name);
    taLeafItr j;
    Event* oev;
    FOR_ITR_EL(Event, oev, events., j) {
      Pattern* pat = (Pattern*)oev->patterns.Leaf(pat_no);
      if(pat == NULL) {
	taMisc::Error("*** Pattern number:", String(pat_no), "not found");
	return;
      }
      float dst = trg_pat->value.Dist(pat->value, metric, norm, tol);
      ((float_RArray*)dtgp->El(cnt)->AR())->Add(dst);
    }
    cnt++;
  }
  GridLogView* lv = ((GridLogView*)disp_log->views.El(0));
  lv->viewspec->Reset();
  lv->viewspec->BuildFromDataTable();
  DT_GridViewSpec* vsp = (DT_GridViewSpec*)lv->viewspec->gp[0];
  vsp->use_gp_name = false; // display each element separately
  vsp->geom.x = vsp->size;
  vsp->geom.y = 1;
  ((DT_GridViewSpec*)lv->viewspec)->UpdateLayout();

  disp_log->InitAllViews();
}

void Environment::ClusterPlot(GraphLog* disp_log, int pat_no, bool,
			      float_RArray::DistMetric metric,
			      bool norm, float tol)
{
  // initialize with leaves
  ClustNode root;
  taLeafItr ei;
  Event* ev;
  FOR_ITR_EL(Event, ev, events., ei) {
    Pattern* trg_pat = (Pattern*)ev->patterns.Leaf(pat_no);
    if(trg_pat == NULL) {
      taMisc::Error("*** Pattern number:", String(pat_no), "not found");
      return;
    }
    ClustNode* nd = new ClustNode;
    nd->name = ev->name;
    nd->pat = &(trg_pat->value);
    root.AddChild(nd);
  }

  root.Cluster(metric, norm, tol);
  
  if(disp_log == NULL) {
    Project* proj = GET_MY_OWNER(Project);
    if(proj == NULL)    return;
    disp_log = (GraphLog*) proj->logs.NewEl(1, &TA_GraphLog);
    winbMisc::DelayedMenuUpdate(proj);
  }

  disp_log->name = "ClusterPlot: " + name + " Pattern: " + String(pat_no);
  root.GraphData(&(disp_log->data));

  GraphLogView* lv = ((GraphLogView*)disp_log->views.El(0));
  lv->viewspec->Reset();
  lv->viewspec->BuildFromDataTable();
  // axis spec is prior data
  DA_GraphViewSpec* xax = (DA_GraphViewSpec*)lv->viewspec->El(0);
  DA_GraphViewSpec* yax = (DA_GraphViewSpec*)lv->viewspec->El(1);
  DA_GraphViewSpec* labs = (DA_GraphViewSpec*)lv->viewspec->El(2);
  taBase::SetPointer((TAPtr*)&(labs->axis_spec), yax);
  xax->max_mode = DA_GraphViewSpec::FIXED;
  xax->range.max = ((float_RArray*)xax->data_array->AR())->range.max * 1.15f; // add some room for the labels
  yax->max_mode = DA_GraphViewSpec::FIXED;
  yax->min_mode = DA_GraphViewSpec::FIXED;
  yax->range.max = ((float_RArray*)yax->data_array->AR())->range.max + .3f; // add some room for the labels
  yax->range.min = 0.2f;
  disp_log->InitAllViews();
}  

void Environment::EnvToGrid(GridLog* grid_log, int pat_no, int ev_x, int ev_y, int pt_x, int pt_y) {
  if(event_specs.size == 0) return;
  PatternSpec* ps = (PatternSpec*)((EventSpec*)event_specs[0])->patterns[pat_no];
  if(ps == NULL) return;

  if(grid_log == NULL) {
    Project* proj = GET_MY_OWNER(Project);
    if(proj == NULL)    return;
    grid_log = (GridLog*) proj->logs.NewEl(1, &TA_GridLog);
    winbMisc::DelayedMenuUpdate(proj);
  }
  else {
    grid_log->Clear();
  }
  grid_log->name = (String)"EnvToGrid: " + name  + " Pattern: " + String(pat_no);

  DataTable* md = &(grid_log->data);
  md->RemoveAll();
  
  DataArray_impl* da = md->NewEl(1, &TA_float_Data);
  da->name = "$row";

  int evx = ev_x;
  int evy = ev_y;
  if(evx < 0) {
    evx = (int)sqrt(events.leaves);
  }
  if((evy < 0) || (evx * evy < events.leaves)) {
    evy = events.leaves / evx;
    while((evx * evy) < events.leaves)
      evy++;
  }

  int ptx = (pt_x < 0) ? ps->geom.x : pt_x;
  int pty = (pt_y < 0) ? ps->geom.y : pt_y;
  int totn = ptx * pty;

  // these are the columns
  int i;
  for(i=0; i<evx; i++) {
    DataTable* dt = (DataTable*)md->NewGp(1);
    dt->name = (String)"c" + (String)i;
    dt->EnforceSize(totn);
    dt->FastEl(0)->name = (String)"<" + (String)totn + ">c0";
    int j;
    for(j=1; j<totn; j++)
      dt->FastEl(j)->name = (String)"c" + (String)j;
  }

  md->ResetData();

  for(i=0;i<evy;i++) {
    md->AddBlankRow();		// add the rows
    ((float_Data*)md->FastEl(0))->ar->FastEl(i) = evy - i -1;	// label in reverse order
  }

  int uni = 0;
  int ux, uy;
  for(uy=0;uy<evy; uy++) {
    if(uni >= events.leaves)
      break;
    for(ux=0; ux<evx; ux++, uni++) {
      if(uni >= events.leaves)
	break;
      Pattern* pat = (Pattern*)((Event*)events.Leaf(uni))->patterns[pat_no];
      DataTable* dt = (DataTable*)md->gp.FastEl(ux);
      dt->PutArrayToRow(pat->value, evy-uy-1);
    }
  }
  
  DT_GridViewSpec* vspc = (DT_GridViewSpec*)((GridLogView*)grid_log->views[0])->viewspec;
  vspc->RemoveAll();
  vspc->BuildFromDataTable();

  for(i=0; i<evx; i++) {
    DT_GridViewSpec* da = (DT_GridViewSpec*)vspc->gp[i];
    da->geom.x = ptx; 
    da->geom.y = pty;
    da->name = (String)"c" + (String)i;
    da->display_name = (String)"c" + (String)i;
  }
  ((DA_GridViewSpec*)vspc->FastEl(0))->display_style = DA_GridViewSpec::TEXT;
  vspc->UpdateLayout();
  grid_log->InitAllViews();
  grid_log->data_range.max = evy-1;
  grid_log->InitAllViews();
  ((GridLogView*)grid_log->views[0])->View_FR();
}

//////////////////////////
//	net_iv.h	//
//////////////////////////

Process_MGroup* NetEditor::GetProcessGroup() {
  Project* proj = GET_OWNER(owner, Project);  
  if(proj == NULL)    return NULL;
  return &(proj->processes);
}


//////////////////////////////////////////
//	PDPRoot: structural root	//
//////////////////////////////////////////

#include <ta/enter_iv.h>
declareActionCallback(PDPRoot)
implementActionCallback(PDPRoot)
#include <ta/leave_iv.h>

String PDPRoot::version_no = "2.0";

void PDPRoot::Initialize() {
  name = "root";
  projects.SetBaseType(&TA_Project);
  colorspecs.SetBaseType(&TA_ColorScaleSpec);
}

void PDPRoot::InitLinks() {
  taBase::Own(projects, this);
  taBase::Own(colorspecs, this);
  WinBase::InitLinks();
}

void PDPRoot::Settings() {
  if(taMisc::iv_active)
    TA_taMisc.ive->Edit(TA_taMisc.GetInstance(),window);
}

TAPtr PDPRoot::Browse(const char* init_path) {
  if(!taMisc::iv_active) return NULL;
  
  TAPtr iob = this;
  if(init_path != NULL) {
    String ip = init_path;
    iob = FindFromPath(ip);
    if(iob == NULL) iob = this;
  }

  taivObjChooser* chs = new taivObjChooser(iob, taivM->wkit, taivM->wkit->style(),
					   "Browse for Object", false);
  chs->Choose();
  TAPtr retv = chs->sel_obj;
  delete chs;
  return retv;
}

void PDPRoot::Info() {
  String info = "PDP Info\n";
  info += "This is the PDP++ software package, version: ";
  info += version_no + "\n\n";
  info += "Mailing List:       pdp++request@cnbc.cmu.edu            \n";
  info += "Bug Reports:        pdp++bugreport@cnbc.cmu.edu           \n";
  info += "WWW Page:           http://www.cnbc.cmu.edu/PDP++/PDP++.html\n";
  info += "Anonymous FTP Site: cnbc.cmu.edu/pub/pdp++              \n";
  info += "\n\n";
    
  info += "Copyright (C) 1995-2000 Randall C. O'Reilly, Chadley K. Dawson,\n\
                    James L. McClelland, and Carnegie Mellon University\n\
\n\
Permission to use, copy, and modify this software and its documentation\n\
for any purpose other than distribution-for-profit is hereby granted\n\
without fee, provided that the above copyright notice and this permission\n\
notice appear in all copies of the software and related documentation\n\
\n\
Permission to distribute the software or modified or extended versions\n\
thereof on a not-for-profit basis is explicitly granted, under the above\n\
conditions. 	HOWEVER, THE RIGHT TO DISTRIBUTE THE SOFTWARE OR MODIFIED OR\n\
EXTENDED VERSIONS THEREOF FOR PROFIT IS *NOT* GRANTED EXCEPT BY PRIOR\n\
ARRANGEMENT AND WRITTEN CONSENT OF THE COPYRIGHT HOLDERS\n\
\n\
Note that the taString class, which is derived from the GNU String class\n\
is Copyright (C) 1988 Free Software Foundation, written by Doug Lea, and\n\
is covered by the GNU General Public License, see ta_string.h\n\
The iv_graphic library and some iv_misc classes were derived from the\n\
InterViews morpher example and other InterViews code, which is\n\
Copyright (C) 1987, 1988, 1989, 1990, 1991 Stanford University\n\
Copyright (C) 1991 Silicon Graphics, Inc\n\
\n\
THE SOFTWARE IS PROVIDED 'AS-IS' AND WITHOUT WARRANTY OF ANY KIND\n\
EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY\n\
WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE\n\
\n\
IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE FOR ANY SPECIAL\n\
INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES\n\
WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT\n\
ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY\n\
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS\n\
SOFTWARE";

  taMisc::Choice(info, "Ok");
}
  
void PDPRoot::Quit() {
  projects.RemoveAll();
  cssMisc::Top->ExitShell();	// just exit from top-level shell
}

void PDPRoot::Destroy() { 
  projects.RemoveAll();
  colorspecs.RemoveAll();
  if(taMisc::iv_active) {
    if(ta_file != NULL)
      taRefN::unRefDone(ta_file);
    ta_file = NULL;
    int i;
    for(i=0; i<10; i++) {
      cssivSession::RunPending();
      if(cssivSession::WaitProc != NULL) 	// then its safe to do some work
	(*cssivSession::WaitProc)();
    }
    CloseWindow();
    cssivSession::Quit();	    // then quit..
  }
  taMisc::types.RemoveAll();	// get rid of all the types before global dtor!
}


void PDPRoot::SetWinName() {
  if(!taMisc::iv_active) return;
  if(window == NULL)    return;
  String nw_name = String(ivSession::instance()->classname()) + ": Root";
  if(nw_name == win_name) return;
  win_name = nw_name;
  if(window->style() == NULL)
    window->style(new ivStyle(wkit->style()));
  window->style()->attribute("title", win_name);
  window->style()->attribute("name", win_name);
  window->style()->attribute("iconName", win_name);
#ifndef CYGWIN
  if(window->is_mapped()) {
    ivManagedWindowRep* mwRep = window->rep();
    mwRep->do_set(window, &ivManagedWindowRep::set_name);
    mwRep->do_set(window, &ivManagedWindowRep::set_icon_name);
  }
#endif
}


void PDPRoot::GetWindow() {
  IconGlyph* ig = 
    new IconGlyph(new ivBackground(win_box, wkit->background()),NULL,this);
  ig->SetIconify = winbMisc::SetIconify;
  ig->ScriptIconify= winbMisc::ScriptIconify;
  window = new ivApplicationWindow(ig);
  ig->SetWindow(window);
}

bool PDPRoot::ThisMenuFilter(MethodDef* md) {
  if((md->name == "Load") || (md->name == "Save") || (md->name == "SaveAs") ||
     (md->name == "Close") || (md->name == "Print") || (md->name == "Print_Data") ||
     (md->name == "GetAllWinPos") || (md->name == "ScriptAllWinPos") ||
     (md->name == "SetWinPos") || (md->name == "CopyFrom") || (md->name == "CopyTo") ||
     (md->name == "DuplicateMe") || (md->name == "ChangeMyType"))
    return false;
  return true;
}

void PDPRoot::SetWinPos(float left, float bottom, float width, float height) {
  WinGeometry* wn_pos = &win_pos;
  if(projects.leaves > 0)
    wn_pos = &(((Project*)projects.Leaf(0))->root_win_pos);
    
  bool was_set = false;
  if((left != 0) && (bottom != 0)) {
    wn_pos->lft = left;
    wn_pos->bot = bottom;
    was_set = true;
  }
  if((width != 0) && (height != 0)) {
    wn_pos->wd = width;
    wn_pos->ht = height;
    was_set = true;
  }
  if(was_set) {
    wn_pos->UpdateAfterEdit();	// puts it on the update_winpos list
  }
  else {
    wn_pos->SetWinPos();	// does it now
  }
}

void PDPRoot::GetWinPos() {
  WinGeometry* wn_pos = &win_pos;
  if(projects.leaves > 0)
    wn_pos = &(((Project*)projects.Leaf(0))->root_win_pos);
  wn_pos->GetWinPos();
}

//////////////////////////////////////////
//	Project: pdp projects		//
//////////////////////////////////////////

int TypeDefault_MGroup::Dump_Load_Value(istream& strm, TAPtr par) {
  Reset();			// get rid of any existing defaults before loading
  return PDPMGroup::Dump_Load_Value(strm, par);
}

void ProjViewState::Initialize() {
  in_use = false;
  iconify = false;
  draw_links = false;
  selected = false;
}

void ProjViewState_List::UnSetInUse() {
  int i;
  for(i=0;i<size;i++)
    FastEl(i)->in_use = false;
}

void ProjViewState_List::RemoveNotInUse() {
  int i;
  for(i=size-1;i>=0;i--) 
    if(!FastEl(i)->in_use) Remove(i);
}

void Project::Initialize() {
  defaults.SetBaseType(&TA_TypeDefault);
  specs.SetBaseType(&TA_BaseSpec);
  networks.SetBaseType(&TA_Network);
  environments.SetBaseType(&TA_Environment);
  processes.SetBaseType(&TA_SchedProcess);
  logs.SetBaseType(&TA_TextLog);
  scripts.SetBaseType(&TA_Script);

  editor = NULL;
  obj_width = 11;
  view_colors.SetBaseType(&TA_RGBA);
  the_colors.SetBaseType(&TA_TAColor);
  view_states.SetBaseType(&TA_ProjViewState);
  mnu_updating = false;
}

void Project::InitLinks() {
  taBase::Own(defaults, this);
  taBase::Own(specs, this);
  taBase::Own(networks, this);
  taBase::Own(environments, this);
  taBase::Own(processes, this);
  taBase::Own(logs, this);
  taBase::Own(scripts, this);
  taBase::Own(root_win_pos, pdpMisc::root); // owned by root..

  taBase::Own(view_font, this);
  taBase::Own(view_colors, this);
  taBase::Own(the_colors, this);
  taBase::Own(view_states, this);

  // need a small font!
#ifndef CYGWIN
  view_font.SetFont("*-helvetica-medium-r-*-*-8*");
#else
  view_font.SetFont("*Arial*medium*--9*");
#endif

  view_colors.EnforceSize(COLOR_COUNT);
  view_colors.El(TEXT)->name = "black";
  view_colors.El(BACKGROUND)->r = .752941;
  view_colors.El(BACKGROUND)->g = .768627;
  view_colors.El(BACKGROUND)->b = .827451;
  view_colors.El(NETWORK)->name = "hot pink";
  view_colors.El(ENVIRONMENT)->name = "green";
  view_colors.El(SCHED_PROC)->name = "yellow";
  view_colors.El(STAT_GROUP)->name = "LightSteelBlue2";
  view_colors.El(SUBPROC_GROUP)->name = "wheat";
  view_colors.El(STAT_PROC)->name = "SteelBlue1";
  view_colors.El(OTHER_PROC)->name = "gold";
  view_colors.El(PDPLOG)->name = "violet";
  view_colors.El(STAT_AGG)->name = "aquamarine";
  view_colors.El(GEN_GROUP)->name = "grey64";

  the_colors.EnforceSize(COLOR_COUNT);
  if(taMisc::iv_active) {
    int i;
    for(i=0;i<COLOR_COUNT;i++) {
      view_colors.El(i)->UpdateAfterEdit();
      the_colors.El(i)->SetColor(view_colors.El(i));
    }
  }

  LoadDefaults();

  WinBase::InitLinks();
}

void Project::LoadDefaults() {
  bool loading = taMisc::is_loading; // cache is loading flag
  fstream def;
  if(!pdpMisc::root->default_file.empty() && cssProgSpace::GetFile(def, pdpMisc::root->default_file)) {
    defaults.Load(def);
    def.close();
    //    cerr << "Successfully loaded default file: " << pdpMisc::root->default_file << "\n";
  }
  else if(pdpMisc::defaults_str != NULL) {
    istrstream sdef(pdpMisc::defaults_str, strlen(pdpMisc::defaults_str));
    sdef.seekg(0);
    defaults.Load(sdef);
    cerr << "Using standard pre-compiled defaults file\n";
  }
  else {
    taMisc::Error("Warning: no default file was available - created objects will not automatically be of correct type");
  }
  taMisc::is_loading = loading;
}

void Project::CutLinks() {
  if(editor != NULL) { delete editor; editor = NULL; }
  WinBase::CutLinks();	// close windows, etc
  scripts.CutLinks();
  logs.CutLinks();
  processes.CutLinks();
  environments.CutLinks();
  networks.CutLinks();
  specs.CutLinks();
  defaults.CutLinks();

  view_font.CutLinks();
  view_colors.CutLinks();
  the_colors.CutLinks();
  view_states.CutLinks();
}

void Project::Copy_(const Project& cp) {
  defaults = cp.defaults;
  specs = cp.specs;
  networks = cp.networks;
  environments = cp.environments;
  processes = cp.processes;
  logs = cp.logs;
  scripts = cp.scripts;

  obj_width = cp.obj_width;
  view_font = cp.view_font;
  view_colors = cp.view_colors;
  the_colors = cp.the_colors;
  view_states = cp.view_states;
}

void Project::UpdateAfterEdit() {
  WinBase::UpdateAfterEdit();
  view_colors.EnforceSize(COLOR_COUNT);
  the_colors.EnforceSize(COLOR_COUNT);

  root_win_pos.UpdateAfterEdit();

  if(taMisc::iv_active) {
    int i;
    for(i=0;i<COLOR_COUNT;i++) {
      view_colors.El(i)->UpdateAfterEdit();
      the_colors.El(i)->SetColor(view_colors.El(i));
    }
    InitDisplay();
  }
}

void Project::UpdateMenus() {
  mnu_updating = true;
  InitDisplay();
  WinBase::UpdateMenus();
  mnu_updating = false;
}

void Project::GetWinPos() {
  WinBase::GetWinPos();
  root_win_pos.GetWinPos();	// get this..
}

int Project::Save(ostream& strm, TAPtr par, int indent) {
  if(taMisc::iv_active) {
    taMisc::Busy();
    GetAllWinPos();
  }
  taMisc::is_saving = true;
  dumpMisc::path_tokens.Reset();
  strm << "// ta_Dump File v1.0\n";   // be sure to check version with Load
  int rval = Dump_Save_Path(strm, par, indent);
  if(rval == false) {
    if(taMisc::iv_active)  taMisc::DoneBusy();
    return rval;
  }
  strm << " {\n";

  // save defaults within project save as first item
  defaults.Dump_Save_Path(strm, par, indent+1);
  strm << " { ";
  if(defaults.Dump_Save_PathR(strm, par, indent+2))
    taMisc::indent(strm, indent+1, 1);
  strm << "  };\n";
  defaults.Dump_Save_impl(strm, par, indent+1);
  defaults.Dump_SaveR(strm, par, indent+1);

  if(Dump_Save_PathR(strm, par, indent+1))
    taMisc::indent(strm, indent, 1);
  strm << "};\n";
  Dump_Save_impl(strm, par, indent);
  Dump_SaveR(strm, par, indent);
  taMisc::is_saving = false;
  dumpMisc::path_tokens.Reset();
  SetWinName();
  if(taMisc::iv_active)  taMisc::DoneBusy();
  return true;
}

int Project::Load(istream& strm, TAPtr par) {
  int rval = WinBase::Load(strm, par); // load-em-up
  if(rval) {				  // don't do this as a dump_load_value
					  // cuz we need an updateafteredit..
    scripts.AutoRun();
  }
  if((ta_file != NULL) && !ta_file->dir.empty() && (ta_file->dir != ".")) {
    // change directories to where the project was loaded!
    chdir(ta_file->dir);
  }
  return rval;
}

void Project::InitDisplay() {
  if(!taMisc::iv_active)
    return;
  if(editor != NULL) 
    editor->InitDisplay();
}

void Project::Minimize() {
  if(!taMisc::iv_active)
    return;
  if(editor != NULL) 
    editor->Minimize();
}

void Project::Maximize() {
  if(!taMisc::iv_active)
    return;
  if(editor != NULL) 
    editor->Maximize();
}

void Project::GetBodyRep() {
  if(!taMisc::iv_active) return;
  if(body != NULL) return;
  editor = new ProjEditor(this, window);
  editor->Init();
  body = editor->GetLook();
  ivResource::ref(body);
}

ivGlyph* Project::GetPrintData(){
  return editor->print_patch;
}

int Project_MGroup::Load(istream& strm, TAPtr par) {
  int rval = PDPMGroup::Load(strm, par);
  if(rval) {
    Project* p;
    taLeafItr i;
    FOR_ITR_EL(Project, p, this->, i) {
      p->scripts.AutoRun();
    }
    if((ta_file != NULL) && !ta_file->dir.empty() && (ta_file->dir != ".")) {
      // change directories to where the project was loaded!
      chdir(ta_file->dir);
    }
  }
  return rval;
}


//////////////////////////////////////////////////
//	PDPView:  view for pdp system		//
//////////////////////////////////////////////////

void PDPView::AddNotify(TAPtr ud) {
  if((ud != NULL) && ud->InheritsFrom(TA_SchedProcess)) {
    ((SchedProcess*)ud)->displays.LinkUnique(this);
  }
}

void PDPView::RemoveNotify(TAPtr ud) {
  if((ud != NULL) && ud->InheritsFrom(TA_SchedProcess)) {
    ((SchedProcess*)ud)->displays.Remove(this);
  }
}

//////////////////////////////////////////////////
//	PDPMGroup  base mgroup for pdp system	//
//////////////////////////////////////////////////

void PDPMGroup::Initialize() {
}

void PDPMGroup::Destroy() {
}

void PDPMGroup::UpdateMenu() {
  MenuGroup_impl::UpdateMenu();
  Project* prj = GET_MY_OWNER(Project);
  if((prj != NULL) && !prj->mnu_updating)
    prj->InitDisplay();
}

//////////////////////////////////////////
//	pdpMisc: misc stuff for pdp	//
//////////////////////////////////////////

PDPRoot* 	pdpMisc::root = NULL;
char* 		pdpMisc::defaults_str = NULL;
String_Array	pdpMisc::proj_to_load;
jmp_buf  	pdpMisc::proc_begin;

void (*pdpMisc::Init_Hook)() = NULL;

ColorScaleSpec* pdpMisc::GetDefaultColor() {
  if((root == NULL) || !taMisc::iv_active)	return NULL;
  return (ColorScaleSpec*)root->colorspecs.DefaultEl();
}

static int get_unique_file_number(int st_no, const char* prefix, const char* suffix) {
  String prfx = prefix;
  String fname;
  int i;
  for(i=st_no; i<10000; i++) {	// stop at 10,000
    fname = prfx + String(i) + suffix;
    int acc = access(fname, R_OK);
    if(acc != 0)
      break;			// its ok..
  }
  fstream strm;
  strm.open(fname, ios::out);	// this should hold the place for the file
  strm.close();			// while it is being saved, etc..
  return i;
}

// for saving a recovery file if program crashes, is killed, etc.
void pdpMisc::SaveRecoverFile(int err) {
  static bool has_crashed = false;
  signal(err, SIG_DFL);		// disable catcher

  taivMisc::Cleanup(err);	// cleanup stuff in taiv

  if(err == SIGUSR1)
    cerr << "PDP++ saving network file(s) from signal: ";
  else if((err == SIGUSR2) || (err == SIGALRM))
    cerr << "PDP++ saving project file(s) from signal: ";
  else 
    cerr << "PDP++ saving recover file(s) and exiting from signal: ";
  taMisc::Decode_Signal(err);
  cerr << "\n";

  if(has_crashed) {
    cerr << "PDP++ unable to save recover file...sorry\n";
    exit(err);
  }
  has_crashed = true;		// to prevent recursive crashing..

  String net_sfx = ".net";	net_sfx += taMisc::compress_sfx;
  String wts_sfx = ".wts";	wts_sfx += taMisc::compress_sfx;
  String proj_sfx = ".proj";	proj_sfx += taMisc::compress_sfx;

  String home_dir = getenv("HOME"); // home directory if curent dir 
				       // does not work.

  taivGetFile gf;		// use a getfile for compressed writes..
  Project* p;
  int cnt = 0;
  taLeafItr i;
  FOR_ITR_EL(Project, p, pdpMisc::root->projects., i) {
    if(err == SIGUSR1) {	// usr1 is to save network at that point
      Network* net;
      taLeafItr ni;
      FOR_ITR_EL(Network, net, p->networks., ni) {
	String use_sfx = net_sfx;
	if(net->usr1_save_fmt == Network::JUST_WEIGHTS)
	  use_sfx = wts_sfx;
	cnt = get_unique_file_number(cnt, "PDP++NetSave.", use_sfx);
	gf.fname = "PDP++NetSave." + String(cnt) + use_sfx;
	ostream* strm = gf.open_write();
	if(((strm == NULL) || strm->bad()) && !home_dir.empty()) {
	  // try it with the home diretory
	  cerr << "Error saving in current directory, trying home directory";
	  gf.Close();
	  gf.fname = home_dir + "/" + gf.fname;
	  strm = gf.open_write();
	}
	if((strm == NULL) || strm->bad())
	  cerr << "SaveNetwork: could not open file: " << gf.fname << "\n";
	else {
	  if(net->usr1_save_fmt == Network::JUST_WEIGHTS)
	    net->WriteWeights(*strm);
	  else
	    net->Save(*strm);
	}
	gf.Close();
	cnt++;
      }
    }
    else {
      String prfx;
      if((err == SIGUSR2) || (err == SIGALRM))
	prfx = "PDP++Project.";
      else
	prfx = "PDP++Recover.";
      cnt = get_unique_file_number(cnt, prfx, proj_sfx);
      gf.fname = prfx + String(cnt) + proj_sfx;
      ostream* strm = gf.open_write();
      if(((strm == NULL) || strm->bad()) && !home_dir.empty()) {
	// try it with the home diretory
	cerr << "Error saving in current directory, trying home directory";
	gf.Close();
	gf.fname = home_dir + "/" + gf.fname;
	strm = gf.open_write();
      }
      if((strm == NULL) || strm->bad())
	cerr << "SaveRecoverFile: could not open file: " << gf.fname << "\n";
      else
	p->Save(*strm);
      gf.Close();
      cnt++;
    }
  }

  if((err == SIGALRM) || (err == SIGUSR1) || (err == SIGUSR2)) {
    taMisc::Register_Cleanup((SIGNAL_PROC_FUN_TYPE) SaveRecoverFile);
    has_crashed = false;
  }
  else {
    kill(getpid(), err);		// activate signal 
  }
}


int pdpMisc::WaitProc() {
  if(proj_to_load.size > 0) {
    taivGetFile* gf = NULL;
    bool delete_gf = true;
    if(taMisc::iv_active) {
      root->GetFileDlg();
      gf = root->projects.ta_file;
      delete_gf = false;
    }
    if(gf == NULL)
      gf = new taivGetFile;

    taivMisc::Busy();
    int i;
    for(i=0; i<proj_to_load.size; i++) {
      String fnm = proj_to_load[i];
      gf->fname = fnm;
      istream* strm = gf->open_read();
      if((strm == NULL) || !(gf->open_file))
	taMisc::Error("Project Load: could not open file", fnm);
      else {
	taivMisc::CreateLoadDialog();
	root->projects.Load(*strm);
	if(root->projects.size > i)
	  ((Project*)root->projects[i])->file_name = fnm;
	taivMisc::RemoveLoadDialog();
      }
      gf->Close();
      if(taMisc::iv_active) {
	winbMisc::DelayedMenuUpdate(&(root->projects));
      }
    }
    if(delete_gf)
      delete gf;
    proj_to_load.Reset();
    taivMisc::DoneBusy();
  }
  if(taMisc::iv_active) {
    winbMisc::OpenWindows();
    return winbMisc::WaitProc(); // set our waitproc
  }
  else {
    return tabMisc::Wait_DelayedRemove();
  }
}


// for pdp icon (See base.h)

#define pdp_bitmap_width 64
#define pdp_bitmap_height 64
static unsigned char pdp_bitmap_bits[] = {
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xf0, 0xff, 0x83, 0xff,
  0x1f, 0xfc, 0xff, 0x00, 0x10, 0x00, 0x82, 0x00, 0x10, 0x04, 0x80, 0x00,
  0x10, 0x00, 0x82, 0x00, 0x10, 0x04, 0x80, 0x00, 0x10, 0x1f, 0x82, 0xf8,
  0x10, 0xc4, 0x87, 0x00, 0x10, 0x32, 0x82, 0x10, 0x11, 0x84, 0x8c, 0x00,
  0x10, 0x22, 0x82, 0x10, 0x11, 0x84, 0x88, 0x00, 0x10, 0x32, 0x82, 0x10,
  0x11, 0x84, 0x8c, 0x00, 0x10, 0x1e, 0x82, 0x10, 0x11, 0x84, 0x87, 0x00,
  0x10, 0x02, 0x82, 0x10, 0x11, 0x84, 0x80, 0x00, 0x10, 0x02, 0x82, 0x10,
  0x11, 0x84, 0x80, 0x00, 0x10, 0x0f, 0x82, 0xf8, 0x10, 0xc4, 0x83, 0x00,
  0x10, 0x00, 0x82, 0x00, 0x10, 0x04, 0x80, 0x00, 0x10, 0x00, 0x82, 0x00,
  0x10, 0x04, 0x80, 0x00, 0x10, 0x00, 0x82, 0x00, 0x10, 0x04, 0x80, 0x00,
  0xf0, 0xff, 0x83, 0xff, 0x1f, 0xfc, 0xff, 0x00, 0x00, 0x78, 0x00, 0x50,
  0x00, 0xf0, 0x00, 0x00, 0x00, 0x90, 0x03, 0x88, 0x00, 0x4e, 0x00, 0x00,
  0x00, 0x20, 0x0c, 0x04, 0xc1, 0x21, 0x00, 0x00, 0x00, 0x40, 0x70, 0x02,
  0x3a, 0x10, 0x00, 0x00, 0x00, 0x40, 0x80, 0x03, 0x06, 0x08, 0x00, 0x00,
  0x00, 0x80, 0x80, 0xdc, 0x05, 0x04, 0x00, 0x00, 0x00, 0x00, 0x41, 0xf8,
  0x08, 0x02, 0x00, 0x00, 0x00, 0x00, 0x22, 0x07, 0x13, 0x01, 0x00, 0x00,
  0x00, 0x00, 0xf4, 0x00, 0xbc, 0x00, 0x00, 0x00, 0x00, 0xe0, 0xff, 0x07,
  0xff, 0x3f, 0x00, 0x00, 0x00, 0x20, 0x00, 0x04, 0x01, 0x20, 0x00, 0x00,
  0x00, 0x20, 0x00, 0x04, 0x01, 0x20, 0x00, 0x00, 0x00, 0x20, 0x00, 0x04,
  0x01, 0x20, 0x00, 0x00, 0x00, 0x20, 0x00, 0x04, 0x01, 0x20, 0x00, 0x00,
  0x00, 0x20, 0x08, 0x04, 0x41, 0x20, 0x00, 0x00, 0x00, 0x20, 0x08, 0x04,
  0x41, 0x20, 0x00, 0x00, 0x00, 0x20, 0x3e, 0x04, 0xf1, 0x21, 0x00, 0x00,
  0x00, 0x20, 0x08, 0x04, 0x41, 0x20, 0x00, 0x00, 0x00, 0x20, 0x08, 0x04,
  0x41, 0x20, 0x00, 0x00, 0x00, 0x20, 0x00, 0x04, 0x01, 0x20, 0x00, 0x00,
  0x00, 0x20, 0x00, 0x04, 0x01, 0x20, 0x00, 0x00, 0x00, 0x20, 0x00, 0x04,
  0x01, 0x20, 0x00, 0x00, 0x00, 0x20, 0x00, 0x04, 0x01, 0x20, 0x00, 0x00,
  0x00, 0xe0, 0xff, 0x07, 0xff, 0x3f, 0x00, 0x00, 0x00, 0x00, 0xf4, 0x01,
  0x7c, 0x01, 0x00, 0x00, 0x00, 0x00, 0x42, 0x9e, 0x23, 0x06, 0x00, 0x00,
  0x00, 0x00, 0x81, 0xf9, 0x11, 0x08, 0x00, 0x00, 0x00, 0x80, 0x00, 0x07,
  0x1e, 0x10, 0x00, 0x00, 0x00, 0x40, 0xf0, 0x0c, 0xe2, 0x21, 0x00, 0x00,
  0x00, 0x20, 0x0e, 0x10, 0x01, 0xde, 0x00, 0x00, 0x00, 0xf0, 0x01, 0xe0,
  0x00, 0xe0, 0x01, 0x00, 0xe0, 0xff, 0x07, 0xff, 0x3f, 0xf8, 0xff, 0x01,
  0x20, 0x00, 0x04, 0x01, 0x20, 0x08, 0x00, 0x01, 0x20, 0x00, 0x04, 0x01,
  0x20, 0x08, 0x00, 0x01, 0x20, 0x3e, 0x04, 0xf1, 0x21, 0x88, 0x0f, 0x01,
  0x20, 0x64, 0x04, 0x21, 0x22, 0x08, 0x19, 0x01, 0x20, 0x44, 0x04, 0x21,
  0x22, 0x08, 0x11, 0x01, 0x20, 0x64, 0x04, 0x21, 0x22, 0x08, 0x19, 0x01,
  0x20, 0x3c, 0x04, 0x21, 0x22, 0x08, 0x0f, 0x01, 0x20, 0x04, 0x04, 0x21,
  0x22, 0x08, 0x01, 0x01, 0x20, 0x04, 0x04, 0x21, 0x22, 0x08, 0x01, 0x01,
  0x20, 0x1e, 0x04, 0xf1, 0x21, 0x88, 0x07, 0x01, 0x20, 0x00, 0x04, 0x01,
  0x20, 0x08, 0x00, 0x01, 0x20, 0x00, 0x04, 0x01, 0x20, 0x08, 0x00, 0x01,
  0x20, 0x00, 0x04, 0x01, 0x20, 0x08, 0x00, 0x01, 0xe0, 0xff, 0x07, 0xff,
  0x3f, 0xf8, 0xff, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, };


// this is the main that should be called..

int pdpMisc::Main(int argc, char *argv[]) {
#ifdef OPT_MALLOC
  mallopt(M_MXFAST, 48); // optimize allocation for objs up to 48 bytes..
  // todo: the problem is that this comes well after first small block is
  // allocated, so it probably doesn't do any good (and may cause crashing!)
#endif

  bool	use_gui = true;
  String user_spec_def;
  int ac;
  for(ac = 1; ac < argc; ac++) {
    String tmp = argv[ac];
    if(tmp == "-nogui")
      use_gui = false;
    // project is either flagged with a -p flag or the first or second argument
    if((tmp == "-p") || ((ac <= 2) && tmp.contains(".proj"))) {
      if(tmp == "-p") {
	if(argc > ac+1) {
	  proj_to_load.Add(argv[ac+1]);
	  ac++;			// skip over it so as to not load 2x
	}
      }
      else
	proj_to_load.Add(tmp);
    }
    if((tmp == "-d") || (tmp == "-def")) {
      if(argc > ac+1)
	user_spec_def = argv[ac+1];
    }
  }
      
  if(Init_Hook != NULL) 
    (*Init_Hook)();	// call the user's init function (which will call pdp)
  else
    ta_Init_pdp();		// always has to be first

  if(use_gui)
    new ivSession("PDP++", argc, argv, PDP_options, PDP_defs);

  root = new PDPRoot();
  taBase::Ref(root);
  root->InitLinks();	// normally the owner would do this, but..
  root->name = "root";

  // tabMisc stuff
  tabMisc::root = (TAPtr)root;
  taMisc::default_scope = &TA_Project;

  // cssMisc stuff
  cssMisc::HardVars.Push(cssBI::root = new cssTA_Base(root, 1, &TA_PDPRoot,"root"));
  cssMisc::Initialize(argc, (const char**)argv);

  if(use_gui) {
    taivMisc::Initialize();
    taivM->icon_bitmap = new ivBitmap(pdp_bitmap_bits,pdp_bitmap_width,
				       pdp_bitmap_height);
    ivResource::ref(taivM->icon_bitmap);

    root->WinInit();
    root->OpenNewWindow();
    // set the update action (taken after Ok or Apply in the Edit dialog)
    taivMisc::Update_Hook = winbMisc::DelayedMenuUpdate; 
    winbMisc::group_leader = root->window;
    winbMisc::DelayedMenuUpdate(root); 	// get menus after startup..
  }
  // create colorspecs even if nogui, since they are referred to in projects
  root->colorspecs.SetDefaultColor();	// set color after starting up..

  if(cssivSession::WaitProc == NULL)
    cssivSession::WaitProc = pdpMisc::WaitProc;
  
  taMisc::Register_Cleanup((SIGNAL_PROC_FUN_TYPE) SaveRecoverFile);

#ifndef CYGWIN
  String pdp_dir = "/usr/local/pdp++"; // default pdp home directory
#else
  String pdp_dir = "C:/PDP++"; // default pdp home directory
#endif
  char* pdp_dir_env = getenv("PDPDIR");
  if(pdp_dir_env != NULL)
    pdp_dir = pdp_dir_env;

  String home_dir;
  char* home_dir_env = getenv("HOME");
  if(home_dir_env != NULL)
    home_dir = home_dir_env;

  taMisc::include_paths.AddUnique(pdp_dir);
  taMisc::include_paths.AddUnique(pdp_dir+"/css/include");
  taMisc::include_paths.AddUnique(pdp_dir+"/defaults");
  if(!home_dir.empty()) {
    taMisc::include_paths.AddUnique(home_dir+"/mypdp++");
    taMisc::include_paths.AddUnique(home_dir+"/pdp++");
    taMisc::include_paths.AddUnique(home_dir+"/mypdp++/defaults");
    taMisc::include_paths.AddUnique(home_dir+"/pdp++/defaults");
  }
  
  String prognm = argv[0];
  if(prognm.contains('/'))
    prognm = prognm.after('/',-1);
  if(prognm.contains("++"))	// the distribution version of an executable
    prognm = prognm.before("++");
  if(prognm.contains('.'))	// some kind of extention
    prognm = prognm.before('.', -1);

  // set the defaults to be the given executable's name
  if((prognm == "pdpshell") || (prognm == "pdp") || (prognm == "bp"))
    root->default_file = "bp.def";
  else
    root->default_file = prognm + ".def";

  if(!user_spec_def.empty()) {
    root->default_file = user_spec_def;
    if(!root->default_file.contains(".def"))
      root->default_file += ".def"; // justin case
  }

  String vers = "-version";
  int i;
  for(i=1; i<=argc; i++) {
    if(vers == argv[i]) {
      root->Info();
      break;
    }
  }

  cssMisc::Top->CompileRunClear(".pdpinitrc");

  if((proj_to_load.size > 0) && !use_gui)
    (*cssivSession::WaitProc)();	// load file manually, since it won't go thru waitproc
  if(taivMisc::iv_active) winbMisc::OpenWindows();
  cssMisc::Top->StartupShell(cin, cout);

  delete root;
  return 0;
}

