/* -*- C++ -*- */
/*=============================================================================
//									      //
// This file is part of the PDP++ software package.			      //
//									      //
// Copyright (C) 1995 Randall C. O'Reilly, Chadley K. Dawson, 		      //
//		      James L. McClelland, and Carnegie Mellon University     //
//     									      //
// Permission to use, copy, and modify this software and its documentation    //
// for any purpose other than distribution-for-profit is hereby granted	      //
// without fee, provided that the above copyright notice and this permission  //
// notice appear in all copies of the software and related documentation.     //
//									      //
// Permission to distribute the software or modified or extended versions     //
// thereof on a not-for-profit basis is explicitly granted, under the above   //
// conditions. 	HOWEVER, THE RIGHT TO DISTRIBUTE THE SOFTWARE OR MODIFIED OR  //
// EXTENDED VERSIONS THEREOF FOR PROFIT IS *NOT* GRANTED EXCEPT BY PRIOR      //
// ARRANGEMENT AND WRITTEN CONSENT OF THE COPYRIGHT HOLDERS.                  //
// 									      //
// Note that the taString class, which is derived from the GNU String class,  //
// is Copyright (C) 1988 Free Software Foundation, written by Doug Lea, and   //
// is covered by the GNU General Public License, see ta_string.h.             //
// The iv_graphic library and some iv_misc classes were derived from the      //
// InterViews morpher example and other InterViews code, which is             //
// Copyright (C) 1987, 1988, 1989, 1990, 1991 Stanford University             //
// Copyright (C) 1991 Silicon Graphics, Inc.				      //
//									      //
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,         //
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 	      //
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  	      //
// 									      //
// IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE FOR ANY SPECIAL,    //
// INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES  //
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT     //
// ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY,      //
// ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS        //
// SOFTWARE. 								      //
==============================================================================*/

// pdpshell.h
  
#ifndef pdpshell_h
#define pdpshell_h 1

#include <pdp/netstru.h>
#include <pdp/enviro.h>
#include <pdp/process.h>
#include <pdp/pdplog.h>
#include <ta/tdefaults.h>
#include <ta_misc/colorscale.h>
#include <ta_misc/fontspec.h>

class TypeDefault_MGroup : public PDPMGroup {
public:
  int	Dump_Load_Value(istream& strm, TAPtr par=NULL);
  // reset members before loading..
  
  void	Initialize() 		{ SetBaseType(&TA_TypeDefault); }
  void 	Destroy()		{ };
  TA_BASEFUNS(TypeDefault_MGroup);
};

class ProjViewState : public taNBase {
  // ##NO_TOKENS saves state variables associated with the project view, lookup by name
public:
  bool		in_use;		// #READ_ONLY #NO_SAVE whether this info is being used now
  bool		iconify;	// iconify (minimize) the view of this object
  bool 		draw_links;	// show the links between this object and others
  bool		selected;	// is this object selected?

  void Initialize();
  void Destroy()	{ };
  SIMPLE_COPY(ProjViewState);
  COPY_FUNS(ProjViewState, taNBase);
  TA_BASEFUNS(ProjViewState);
};

class ProjViewState_List : public taList<ProjViewState> {
  // ##NO_TOKENS #NO_UPDATE_AFTER list of project view state information
public:
  virtual void	UnSetInUse();		// unset in_use flags for all elements
  virtual void	RemoveNotInUse();	// remove all elements no longer in use

  void	Initialize() 		{ };
  void 	Destroy()		{ };
  TA_BASEFUNS(ProjViewState_List);
};

class ProjEditor;

class Project : public WinBase {
  // ##EXT_proj ##COMPRESS A Project has everything
public:
  enum ViewColors {		// indicies for view_colors
    TEXT,
    BACKGROUND,
    NETWORK,
    ENVIRONMENT,
    SCHED_PROC,
    STAT_GROUP,
    SUBPROC_GROUP,
    STAT_PROC,
    OTHER_PROC,
    PDPLOG,
    STAT_AGG,
    GEN_GROUP,
    COLOR_COUNT
  };

  WinGeometry		root_win_pos; 	// position/size of the root window for saving

  TypeDefault_MGroup    defaults;	// #NO_FIND #NO_SAVE default initial settings for objects
  BaseSpec_MGroup     	specs;		// Specifications for network parameters
  Network_MGroup	networks;	// Networks of interconnected units 
  Environment_MGroup	environments;	// Environments of patterns to present to networks
  Process_MGroup	processes;	// Processes to coordinate training/testing, etc
  PDPLog_MGroup		logs;		// Logs to display statistics in processes
  Script_MGroup		scripts;	// Scripts to control arbitrary actions

  // parameters for controlling the view
  ProjEditor*		editor;		// #IGNORE controls display of project information
  int			obj_width; 	// width of the objects in view (in units of font m's)
  FontSpec	  	view_font;	// the font to use for the labels in the display
  RGBA_List		view_colors; 	// colors to use in the project view
  TAColor_List		the_colors; 	// #IGNORE actual colors
  ProjViewState_List	view_states;	// #HIDDEN state information for view objects
  bool			mnu_updating; 	// #READ_ONLY #NO_SAVE if menu is already being updated (don't init display)

  virtual void	LoadDefaults();
  // load defaults according to root::default_file or precompiled defaults

  virtual void	InitDisplay();  // #MENU initialize the display of object information
  virtual void	Minimize();	// #MENU make display smallest possible size
  virtual void	Maximize();	// #MENU grow display to fit everything in view

  void 		GetBodyRep();
  ivGlyph*	GetPrintData();

  void	GetWinPos();
  void	UpdateMenus();

  int	Load(istream& strm, TAPtr par=NULL);
  int	Save(ostream& strm, TAPtr par=NULL, int indent=0);

  void	UpdateAfterEdit();
  void	Initialize();
  void 	Destroy()		{ CutLinks(); }
  void 	InitLinks();
  void	CutLinks();
  void	Copy_(const Project& cp);
  COPY_FUNS(Project, WinBase);
  TA_BASEFUNS(Project);
};

class Project_MGroup : public PDPMGroup {
public:
  int		Load(istream& strm, TAPtr par=NULL); // call reconnect on nets afterwards
  
  void	Initialize() 		{ SetBaseType(&TA_Project); }
  void 	Destroy()		{ };
  TA_BASEFUNS(Project_MGroup);
};


class PDPRoot : public WinBase {
  // structural root of object heirarchy
public:
  static String		version_no; 	// #READ_ONLY #SHOW current version number
  String		default_file; 	// default name of defaults file
  Project_MGroup	projects; 	// The projects
  ColorScaleSpec_MGroup colorspecs;	// Color Specs

  bool	ThisMenuFilter(MethodDef* md); // don't include saving and loading..
  void	SetWinName();
  void	GetWindow();		// make an app-window
  void  Settings();		// #MENU #MENU_ON_Object edit global settings;
  void	Info();			// #MENU get information/copyright notice
  TAPtr	Browse(const char* init_path=NULL);
  // #MENU #ARGC_0 #USE_RVAL #NO_REVERT_AFTER use object browser to find an object, starting with initial path if given
  void	Quit();			
  // #MENU #CONFIRM #MENU_SEP_BEFORE #NO_REVERT_AFTER quit from software..

  // following use project to store position info!
  void	SetWinPos(float left=0, float bottom=0, float width=0, float height=0);
  void	GetWinPos();

  void 	Initialize();
  void	InitLinks();
  void 	Destroy();
  TA_BASEFUNS(PDPRoot);
};


#endif // pdpshell_h
