/* -*- C++ -*- */
/*=============================================================================
//									      //
// This file is part of the PDP++ software package.			      //
//									      //
// Copyright (C) 1995 Randall C. O'Reilly, Chadley K. Dawson, 		      //
//		      James L. McClelland, and Carnegie Mellon University     //
//     									      //
// Permission to use, copy, and modify this software and its documentation    //
// for any purpose other than distribution-for-profit is hereby granted	      //
// without fee, provided that the above copyright notice and this permission  //
// notice appear in all copies of the software and related documentation.     //
//									      //
// Permission to distribute the software or modified or extended versions     //
// thereof on a not-for-profit basis is explicitly granted, under the above   //
// conditions. 	HOWEVER, THE RIGHT TO DISTRIBUTE THE SOFTWARE OR MODIFIED OR  //
// EXTENDED VERSIONS THEREOF FOR PROFIT IS *NOT* GRANTED EXCEPT BY PRIOR      //
// ARRANGEMENT AND WRITTEN CONSENT OF THE COPYRIGHT HOLDERS.                  //
// 									      //
// Note that the taString class, which is derived from the GNU String class,  //
// is Copyright (C) 1988 Free Software Foundation, written by Doug Lea, and   //
// is covered by the GNU General Public License, see ta_string.h.             //
// The iv_graphic library and some iv_misc classes were derived from the      //
// InterViews morpher example and other InterViews code, which is             //
// Copyright (C) 1987, 1988, 1989, 1990, 1991 Stanford University             //
// Copyright (C) 1991 Silicon Graphics, Inc.				      //
//									      //
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,         //
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 	      //
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  	      //
// 									      //
// IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE FOR ANY SPECIAL,    //
// INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES  //
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT     //
// ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY,      //
// ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS        //
// SOFTWARE. 								      //
==============================================================================*/

// stats.h

#ifndef stats_h
#define stats_h

#include <pdp/process.h>
#include <pdp/netstru.h>
#include <ta_misc/datatable.h>		// for float_RArray and Aggregate

#include <math.h>

#ifdef Aggregate
#undef Aggregate
#endif

class CritParam : public taBase {
  // ##NO_TOKENS ##NO_UPDATE_AFTER #INLINE stopping criteria params
public:
  enum Relation {
    EQUAL,		// #LABEL_=
    NOTEQUAL,		// #LABEL_!=
    LESSTHAN,		// #LABEL_<
    GREATERTHAN,	// #LABEL_>
    LESSTHANOREQUAL,	// #LABEL_<=
    GREATERTHANOREQUAL 	// #LABEL_>=
  };
  
  bool          flag;           // #LABEL_ whether to use this criterion or not
  Relation	rel;		// #LABEL_ relation of statistic to target value
  float		val;		// #LABEL_ target or comparison value
  int		cnt;		// Number of times criterion must be met before stopping
  int		n_met;		// #READ_ONLY number of times actually met

  bool 	Evaluate(float cmp);

  void	Init()	{ n_met = 0; }	// initialize the process (n_met counter)
  
  void  Initialize();
  void 	Destroy()		{ };
  void	Copy_(const CritParam& cp);
  COPY_FUNS(CritParam, taBase);
  TA_BASEFUNS(CritParam);
};

class StatVal : public DataItem {
  // ##NO_TOKENS ##NO_UPDATE_AFTER #INLINE Statistic value
public:
  float		val;		// value of statistic
  String	str_val;	// #HIDDEN_INLINE value of statistic if its a string
  CritParam	stopcrit;	// Stopping Criteria Parameters

  float		operator=(int r)	{ return val = (float) r; }
  float 	operator=(float r)	{ return val = r; }
  float 	operator+=(float r)	{ return val += r; }
  float 	operator-=(float r)	{ return val -= r; }
  float 	operator*=(float r)	{ return val *= r; }
  float 	operator/=(float r)	{ return val /= r; }
  operator 	float& () 		{ return val; }
  operator 	float* () 		{ return &val; }

  void		InitStat(float value=0.0)	{ val = value; }
  void 		Init()		{ stopcrit.Init(); }
  bool 		Crit()		{ return stopcrit.Evaluate(val); }

  void 	Initialize();
  void 	Destroy()		{ };
  void	InitLinks();
  void	Copy_(const StatVal& cp);
  COPY_FUNS(StatVal, DataItem);
  TA_BASEFUNS(StatVal);
};

class StatVal_List : public taBase_List {
  // ##NO_UPDATE_AFTER group of stat values
public:
  void operator=(int r);
  void operator=(float r);

  virtual void	InitStat(float value=0.0); // initialize statistic value
  virtual void 	Init();	  	 	  // initialize process (eg. n_met) (not value!)
  virtual bool 	Crit();		 	  // does an OR of each member

  virtual void	NameStatVals(const char* nm, const char* opts="", bool is_string=false);
  // #MENU #MENU_ON_Actions give statvals default names, options
  virtual bool	HasStopCrit();
  // #MENU #USE_RVAL check if any of the statvals have crit flags set

  virtual void	GenLogData(LogData* ld);  // generate log data

  void	Initialize() 		{ SetBaseType(&TA_StatVal); }
  void 	Destroy()		{ };
  TA_BASEFUNS(StatVal_List);
};

class StatValAgg : public Aggregate {
  // #INLINE Aggregation for StatVal-based values
public:
  void 		ComputeAgg(float& to, float fm)
  { Aggregate::ComputeAgg(to, fm); }
  void		ComputeAggNoUpdt(float& to, float fm)
  { Aggregate::ComputeAggNoUpdt(to, fm); }

  void		NewCopyAgg(StatVal* to, float fm);
  // make a new copy aggregate

  void		ComputeAgg(StatVal* to, float fm);
  // compute aggregation into `to' based on current op from `fm'
  void		ComputeAggNoUpdt(StatVal* to, float fm);
  // compute aggregation but don't update the n_updt counter (for lists)

  virtual void	ComputeAggs(StatVal_List* to, StatVal_List* fm);
  // compute aggregation for whole list of statvals (updt after list)
  virtual void	ComputeAggsNoUpdt(StatVal_List* to, StatVal_List* fm);
  // compute aggregation for whole list of statvals (no updt at all)

  virtual void	AggFromCopy(StatVal_List* fm);
  // from list is the copy_vals of another stat, goes to copy_vals in owner stat

  void 	Initialize();
  void	Destroy();
  TA_BASEFUNS(StatValAgg);
};

class AggStat : public StatValAgg {
  // #INLINE Aggregate statistics over time (processing levels)
public:
  Stat*		real_stat;	// #READ_ONLY #NO_SAVE the 'real' (non-agg) stat
  Stat* 	from;
  // which statisitic to aggregate from (a statistic of a sub process)
  bool		type_safe;	
  // #HIDDEN true if from ptr should be kept type-safe

  virtual void	FindRealStat();	// find and set the real_stat

  virtual void	SetFrom(Stat* frm); // set the from field (convenience)

  String	AppendAggName(const char* nm) const;
  String	PrependAggName(const char* nm) const;

  void 	UpdateAfterEdit();
  void 	Initialize();
  void 	Destroy()		{ CutLinks(); }
  void	CutLinks();
  void	Copy_(const AggStat& cp);
  COPY_FUNS(AggStat, StatValAgg);
  TA_BASEFUNS(AggStat);
};

class Stat : public Process {
  // Generic Statistic Process
public:
  SchedProcess* own_proc;	// #READ_ONLY #NO_SAVE The SchedProcess which owns this stat
  bool		has_stop_crit;	// #READ_ONLY true if any of the stats have a stopping crit
  int		n_copy_vals;	// #READ_ONLY the number of copy values added 
  bool		loop_init;
  // true if stat should be Init()'d before run each time when a loop stat 

  bool		log_stat;	// flag determines if stat data is logged
  AggStat	time_agg;
  // Aggregation over time (i.e., over loop of lower process), if from != NULL
  StatValAgg	net_agg;
  // Aggregation over network objects (i.e., over units)
  StatVal_List	copy_vals;	// the values of the stat if using COPY agg (net or time)
  Layer*	layer;		// restrict computation to this layer if non-null

  void		C_Code();
  LogData& 	GenLogData(LogData* ld=NULL); 	// generate log data

  // it is reccommended that you overwrite these two functions to call .Init{Stat}
  // on your StatVal{_Group} members explicitly, and agg.Init(); InitStat(); in Init()
  // these versions require a search through the members, which is slow..
  virtual float	InitStatVal();	// value to pass to the InitStat() function
  virtual void	InitStat_impl(); // #IGNORE inits generic stat vars (not member statvals)
  virtual void 	InitStat();    	// initializes the stat vars, but not the stat process
  virtual void	Init_impl();	// #IGNORE inits generic stats (not member statvals)
  virtual void 	Init();
  // initializes the process (including stopcrits), calls InitStat()
  virtual bool	Crit();    	// check criteria on stat values

  // these are looping and checking functions, SET TO {} to not compute over something
  virtual void 	Network_Run(); 			// Compute over the network
  virtual void 	Layer_Run();			// Compute over layers in the network
  virtual void 	Unit_Run(Layer* lay);		// Compute over units in a layer
  virtual void 	RecvCon_Run(Unit* unit);	// Compute over recv conns of a unit
  virtual void 	SendCon_Run(Unit* unit);	// Compute over send conns of a unit

  // these are various initialization points for initing values before computing
  virtual void 	Network_Init()		{ };	// Initialize before computing network
  virtual void 	Layer_Init(Layer*) 	{ };	// Initialize before computing layer
  virtual void 	Unit_Init(Unit*)	{ };	// Initialize before computing unit

  // these are the actual statistics functions (guaranteed to have Min_ from _Run)
  virtual void 	Network_Stat()		{ };	// Compute statistic on the network
  virtual void 	Layer_Stat(Layer*) 	{ };	// Compute statistic on a layer
  virtual void 	Unit_Stat(Unit*)	{ };	// Compute statistic on a unit
  virtual void 	Con_Stat(Unit*, Connection*, Unit*) { };
  // Compute statistic on a recv unit, connection, sending unit 

  // aggregation stuff..
  virtual void 	ComputeAggregates();
  // Compute aggregate of agg.from statistic in this stat
  virtual void 	CreateAggregates(Aggregate::Operator default_op = (Aggregate::Operator)-1);
  // #NEW_FUN #MENU #MENU_ON_Actions #ARGC_1 #UPDATE_MENUS Create aggregate stats up the processing hiearchy
  virtual void	DeleteAggregates();
  // #MENU #UPDATE_MENUS delete aggregators of this stat
  virtual void	UpdateAggregates();
  // #MENU #UPDATE_MENUS update aggregators of this stat (i.e., so they get the layer name or other change)
  virtual Stat*	FindAggregator();
  // #MENU #MENU_SEP_BEFORE #USE_RVAL Find the aggregator for this stat in the next-highest level proc

  virtual void	NameStatVals();
  // #IGNORE give default names to statvals (overload to set specific options, etc.)
  virtual bool	HasStopCrit();
  // #IGNORE check if any of the statvals have crit flags set, set has_stop_crit 
  virtual bool	CheckLayerInNet();
  // #IGONRE make sure layer pointer points to a layer in current net, set to one of same name if not

  void 	UpdateAfterEdit();

  void 	Initialize();
  void 	Destroy()		{ CutLinks(); }
  void 	InitLinks();
  void	CutLinks();
  void	Copy_(const Stat& cp);
  COPY_FUNS(Stat, Process);
  TA_BASEFUNS(Stat);
};

BaseGroup_of(Stat);

// SE_Stat and MonitorStat are so basic that they are here, and not in extra

class SE_Stat : public Stat {
  // ##COMPUTE_IN_TrialProcess Squared Error Statistic
public:
  StatVal	se;			// squared errors
  float		tolerance;		// if error is less than this, its 0

  void		RecvCon_Run(Unit*)	{ }; // don't do these!
  void		SendCon_Run(Unit*)	{ };

  void		InitStat();
  void		Init();
  bool		Crit();
  void		Network_Init();
  void		Layer_Run();		// only compute on TARG layers
  void 		Unit_Stat(Unit* unit);

  void		NameStatVals();

  void	Initialize();
  void 	Destroy()		{ };
  SIMPLE_COPY(SE_Stat);
  COPY_FUNS(SE_Stat, Stat);
  TA_BASEFUNS(SE_Stat);
};
  
class MonitorStat: public Stat {
  // ##COMPUTE_IN_TrialProcess Network Monitor Statistic
public:
  StatVal	mon;		// Value if aggregating (copy goes in copy vals)
  MemberSpace   members;	// #IGNORE memberdefs
  taBase_List	ptrs;     	// #HIDDEN #NO_SAVE actual ptrs to values
  taBase_List 	objects;	// #LINK_GROUP Group of network objects
  String        variable;	// Variable (member) to monitor
  SimpleMathSpec pre_proc_1;	// first step of pre-processing to perform
  SimpleMathSpec pre_proc_2;	// second step of pre-processing to perform
  SimpleMathSpec pre_proc_3;	// third step of pre-processing to perform

  void		InitStat();
  void		Init();
  bool		Crit();
  void		Layer_Run()	{ }; // don't run the layers
  void		Network_Init();
  void		Network_Stat();

  // these are for finding the members and building the stat
  // out of the objects and the variable
  void		ScanObjects();	// #IGNORE
  void		ScanLayer(Layer* lay); // #IGNORE
  void		ScanUnitGroup(Unit_Group* ug);	// #IGNORE
  void		ScanUnit(Unit* u,Projection* p=NULL); // #IGNORE
  void		ScanConGroup(Con_Group* cg,char* varname,Projection* p=NULL); // #IGNORE
  void		ScanProjection(Projection* p); // #IGNORE

  virtual void	SetVariable(const char* varnm); // set variable and update appropriately
  virtual void	SetObject(TAPtr obj); // clear any existing objects and set to obj and update
  virtual void	AddObject(TAPtr obj); // add obj and update

  void	NameStatVals();
  void	UpdateAfterEdit();

  void 	Initialize();
  void 	Destroy()		{ CutLinks(); }
  void	InitLinks();
  void	CutLinks();
  void 	Copy_(const MonitorStat& cp);
  COPY_FUNS(MonitorStat, Stat);
  TA_BASEFUNS(MonitorStat);
};


#endif // stats_h
