/* -*- C++ -*- */
/*=============================================================================
//									      //
// This file is part of the TypeAccess/C-Super-Script software package.	      //
//									      //
// Copyright (C) 1995 Randall C. O'Reilly, Chadley K. Dawson, 		      //
//		      James L. McClelland, and Carnegie Mellon University     //
//     									      //
// Permission to use, copy, modify, and distribute this software and its      //
// documentation for any purpose is hereby granted without fee, provided that //
// the above copyright notice and this permission notice appear in all copies //
// of the software and related documentation.                                 //
// 									      //
// Note that the PDP++ software package, which contains this package, has a   //
// more restrictive copyright, which applies only to the PDP++-specific       //
// portions of the software, which are labeled as such.			      //
//									      //
// Note that the taString class, which is derived from the GNU String class,  //
// is Copyright (C) 1988 Free Software Foundation, written by Doug Lea, and   //
// is covered by the GNU General Public License, see ta_string.h.             //
// The iv_graphic library and some iv_misc classes were derived from the      //
// InterViews morpher example and other InterViews code, which is             //
// Copyright (C) 1987, 1988, 1989, 1990, 1991 Stanford University             //
// Copyright (C) 1991 Silicon Graphics, Inc.				      //
//									      //
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,         //
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 	      //
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  	      //
// 									      //
// IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE FOR ANY SPECIAL,    //
// INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES  //
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT     //
// ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY,      //
// ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS        //
// SOFTWARE. 								      //
==============================================================================*/

// ta_group.cc

#include <ta/ta_group.h>

#ifndef NO_IV
#include <ta/ta_group_iv.h>
#endif


////////////////////////////
//      taGroup	  //
////////////////////////////

// not only sub-groups cause dirtiness, because if you add an el to a
// previously null group, it needs added, etc..

void taSubGroup::Dirty() {
  if(owner != NULL) ((TAGPtr)owner)->Dirty();
}

void taGroup_impl::Destroy() {
  if(leaf_gp != NULL) {
    taBase::unRefDone(leaf_gp);
    leaf_gp = NULL;
  }
  RemoveAll();
}

void taGroup_impl::Dirty() {
  if(leaf_gp != NULL) {
    taBase::unRefDone(leaf_gp);
    leaf_gp = NULL;
  }
  if(super_gp != NULL)
    super_gp->Dirty();
}

void taGroup_impl::InitLeafGp() const {
  if(leaf_gp != NULL)
    return;
  taGroup_impl* ncths = (taGroup_impl*)this;
  ncths->leaf_gp = new TALOG;
  taBase::Own(leaf_gp, ncths);
  InitLeafGp_impl(ncths->leaf_gp);
}

void taGroup_impl::InitLeafGp_impl(TALOG* lg) const {
  if(size > 0)
    lg->Push((taGroup_impl*)this);
  int i;
  for(i=0; i<gp.size; i++)
    FastGp_(i)->InitLeafGp_impl(lg);
}

void taGroup_impl::Copy(const taGroup_impl& cp) {
  taList_impl::Copy(cp);
  gp.Copy(cp.gp);
}

void taGroup_impl::Initialize() {
  leaves = 0;
  super_gp = NULL;
  leaf_gp = NULL;
}

void taGroup_impl::InitLinks() {
  taList_impl::InitLinks();
  gp.SetBaseType(GetTypeDef());	// more of the same type of group
  taBase::Own(gp, this);

  super_gp = GetSuperGp_();
  if(super_gp != NULL) {
    SetBaseType(super_gp->el_base);
    el_typ = super_gp->el_typ;
  }
}

bool taGroup_impl::Remove(int i) {
  if(taList_impl::Remove(i)) {
    UpdateLeafCount_(-1);
    return true;
  }
  return false;
}

void taGroup_impl::AddEl_(void* it) {
  taList_impl::AddEl_(it);
  UpdateLeafCount_(1);		// not the most efficient, but gets it at a low level
}
  
TAPtr taGroup_impl::New(int no, TypeDef* typ) {
  if(no == 0) {
#ifndef NO_IV
    if(taMisc::iv_active)
      return gpivGroupNew::New(this,typ);
#endif
    return NULL;
  }
  if(typ == NULL)
    typ = el_typ;

  if(typ->InheritsFrom(TA_taGroup_impl)) {
    if(!typ->InheritsFrom(gp.el_typ))
      typ = GetTypeDef();	// always create one of yourself..
    TAPtr rval = gp.New(no, typ);
//    UpdateAfterEdit();
    return rval;
  }
  if(!typ->InheritsFrom(el_base)) {
    taMisc::Error("*** Attempt to create type:", typ->name,
		   "in list with base type:", el_base->name);
    return NULL;
  }
  TAPtr rval = taList_impl::New(no, typ);
  return rval;
}

TAGPtr taGroup_impl::NewGp_(int no, TypeDef* typ) {
  if(no == 0) {
#ifndef NO_IV
    if(taMisc::iv_active) {
      gpivGroupNew::New(this,typ);
      return NULL;		// not sure if rval is a group or not
    }
#endif
    return NULL;
  }
  if(typ == NULL)
    typ = GetTypeDef();		// always create one of yourself..
  TAGPtr rval = (TAGPtr)gp.New(no, typ);
//  UpdateAfterEdit();
  return rval;
}

TAPtr taGroup_impl::NewEl_(int no, TypeDef* typ) {
  if(no == 0) {
#ifndef NO_IV
    if(taMisc::iv_active)
      return gpivGroupNew::New(this,typ);
#endif
    return NULL;
  }
  if(typ == NULL)
    typ = el_typ;
  TAPtr rval = taList_impl::New(no, typ);
  return rval;
}

MemberDef* taGroup_impl::FindMembeR(const char* nm, void*& ptr) const {
  String idx_str = nm;
  idx_str = idx_str.before(']');
  if(idx_str != "") {
    idx_str = idx_str.after('[');
    int idx = atoi(idx_str);
    if((size == 0) || (idx >= size)) {
      ptr = NULL;
      return NULL;
    }
    ptr = el[idx];
    return &find_md;
  }

  int i;
  if((i = FindLeaf(nm)) >= 0) {
    ptr = Leaf_(i);
    return &find_md;
  }

  MemberDef* rval;
  if((rval = GetTypeDef()->members.FindNameAddrR(nm, (void*)this, ptr)) != NULL)
    return rval;
  int max_srch = MIN(taMisc::search_depth, size);
  for(i=0; i<max_srch; i++) {
    TAPtr first_el = (TAPtr)FastEl_(i);
    if((first_el != NULL) && // only search owned objects
       ((first_el->GetOwner()==NULL) || (first_el->GetOwner() == (taBase *) this))) {
      return first_el->FindMembeR(nm, ptr);
    }
  }
  ptr = NULL;
  return NULL;
}

MemberDef* taGroup_impl::FindMembeR(TypeDef* it, void*& ptr) const {
  int i;
  if((i = FindLeaf(it)) >= 0) {
    ptr = Leaf_(i);
    return &find_md;
  }

  MemberDef* rval;
  if((rval = GetTypeDef()->members.FindTypeAddrR(it, (void*)this, ptr)) != NULL)
    return rval;
  int max_srch = MIN(taMisc::search_depth, size);
  for(i=0; i<max_srch; i++) {
    TAPtr first_el = (TAPtr)FastEl_(i);
    if((first_el != NULL) && // only search owned objects
       ((first_el->GetOwner()==NULL) || (first_el->GetOwner() == (taBase *) this))) {
      return first_el->FindMembeR(it, ptr);
    }
  }
  ptr = NULL;
  return NULL;
}

ostream& taGroup_impl::OutputR(ostream& strm, int indent) const {
  taMisc::indent(strm, indent) << name << "[" << size << "] = {\n";
  TypeDef* td = GetTypeDef();
  int i;
  for(i=0; i < td->members.size; i++) {
    MemberDef* md = td->members.FastEl(i);
    if(md->HasOption("EDIT_IN_GROUP"))
      md->Output(strm, (void*)this, indent+1);
  }

  for(i=0; i<size; i++) {
    if(el[i] == NULL)	continue;
    ((TAPtr)el[i])->OutputR(strm, indent+1);
  }

  gp.OutputR(strm, indent+1);

  taMisc::indent(strm, indent) << "}\n";
  return strm;
}

// save the path of all the elements in the group
int taGroup_impl::Dump_Save_PathR_impl(ostream& strm, TAPtr par, int indent) {
  int rval = taList_impl::Dump_Save_PathR_impl(strm, par, indent); // save first-level
  if(rval == false)
    rval = gp.Dump_Save_PathR_impl(strm, par, indent); 
  else
    gp.Dump_Save_PathR_impl(strm, par, indent);
  return rval;
}

int taGroup_impl::Dump_SaveR(ostream& strm, TAPtr par, int indent) {
  taList_impl::Dump_SaveR(strm, par, indent);
  gp.Dump_SaveR(strm, par, indent); // subgroups get saved
  return true;
}


TAGPtr taGroup_impl::GetSuperGp_() {
  if(owner == NULL)
    return NULL;
  if(owner->InheritsFrom(TA_taList)) {
    TAPtr ownr = owner->GetOwner();
    if((ownr != NULL) && (ownr->InheritsFrom(TA_taGroup_impl)))
      return (TAGPtr)ownr;
  }
  return NULL;
}

TAPtr taGroup_impl::Leaf_(int idx) const {
  if(idx >= leaves)
    return NULL;
  if(size && (idx < size))
    return (TAPtr)el[idx];

  int nw_idx = (int)idx - size;
  int i;
  TAGPtr sbg;
  for(i=0; i<gp.size; i++) {
    sbg = FastGp_(i);
    if(sbg->leaves && (sbg->leaves > nw_idx))
      return sbg->Leaf_(nw_idx);
    nw_idx -= (int)sbg->leaves;
  }
  return NULL;
}

TAGPtr taGroup_impl::LeafGp_(int idx) const {
  if(idx >= leaves)
    return NULL;
  if(size && (idx < size))
    return (TAGPtr)this;

  int nw_idx = (int)idx - size;
  int i;
  TAGPtr sbg;
  for(i=0; i<gp.size; i++) {
    sbg = FastGp_(i);
    if(sbg->leaves && (sbg->leaves > nw_idx))
      return sbg->LeafGp_(nw_idx);
    nw_idx -= (int)sbg->leaves;
  }
  return NULL;
}

void taGroup_impl::UpdateLeafCount_(int no) {
  leaves += no;
  if(super_gp != NULL)
    super_gp->UpdateLeafCount_(no);
}

bool taGroup_impl::RemoveLeaf(int idx) {
  if(idx >= leaves)
    return false;
  if(size && (idx < size))
    return Remove(idx);

  int nw_idx = (int)idx - size;
  int i;
  TAGPtr sbg;
  for(i=0; i<gp.size; i++) {
    sbg = FastGp_(i);
    if(sbg->leaves && (sbg->leaves > nw_idx))
      return sbg->RemoveLeaf(nw_idx);
    nw_idx -= (int)sbg->leaves;
  }
  return false;
}

bool taGroup_impl::RemoveLeaf(const char* it) {
  int i;
  if((i = FindLeaf(it)) < 0)
    return false;
  return RemoveLeaf(i);
}

bool taGroup_impl::RemoveLeaf(TAPtr it) {
  int i;
  if((i = FindLeaf(it)) < 0)
    return false;
  return RemoveLeaf(i);
}

void taGroup_impl::RemoveAll() {
  gp.RemoveAll();
  taList_impl::RemoveAll();
  leaves = 0;
}

void taGroup_impl::EnforceLeaves(int sz){
  if(sz > leaves)  New(sz - leaves,el_typ);
  while(leaves > sz) RemoveLeaf(leaves-1);
}

void taGroup_impl::EnforceSameStru(const taGroup_impl& cp) {
  taList_impl::EnforceSameStru(cp);
  gp.EnforceSameStru(cp.gp);
  int i;
  for(i=0; i<gp.size; i++) {
    FastGp_(i)->EnforceSameStru(*(cp.FastGp_(i)));
  }
}

int taGroup_impl::ReplaceType(TypeDef* old_type, TypeDef* new_type) {
  int nchanged = taList_impl::ReplaceType(old_type, new_type);
  nchanged += gp.ReplaceType(old_type, new_type);
  int i;
  for(i=0; i<gp.size; i++) {
    nchanged += FastGp_(i)->ReplaceType(old_type, new_type);
  }
  return nchanged;
}

int taGroup_impl::FindLeaf(const char* nm) const {
  int idx;
  if((idx = Find(nm)) >= 0)
    return idx;

  int new_idx = size;
  int i;
  TAGPtr sbg;
  for(i=0; i<gp.size; i++) {
    sbg = FastGp_(i);
    if((idx = sbg->FindLeaf(nm)) >= 0)
      return idx + new_idx;
    new_idx += (int)sbg->leaves;
  }
  return -1;
}

int taGroup_impl::FindLeaf(TAPtr it) const {
  int idx;
  if((idx = Find(it)) >= 0)
    return idx;

  int new_idx = size;
  int i;
  TAGPtr sbg;
  for(i=0; i<gp.size; i++) {
    sbg = FastGp_(i);
    if((idx = sbg->FindLeaf(it)) >= 0)
      return idx + new_idx;
    new_idx += (int)sbg->leaves;
  }
  return -1;
}

int taGroup_impl::FindLeaf(TypeDef* it) const {
  int idx;
  if((idx = Find(it)) >= 0)
    return idx;

  int new_idx = size;
  int i;
  TAGPtr sbg;
  for(i=0; i<gp.size; i++) {
    sbg = FastGp_(i);
    if((idx = sbg->FindLeaf(it)) >= 0)
      return idx + new_idx;
    new_idx += (int)sbg->leaves;
  }
  return -1;
}

TAPtr taGroup_impl::FindLeafName_(const char* it, int& idx) const {
  idx = FindLeaf(it);
  if(idx >= 0) return Leaf_(idx);
  return NULL;
}

TAPtr taGroup_impl::FindLeafType_(TypeDef* it, int& idx) const {
  idx = FindLeaf(it);
  if(idx >= 0) return Leaf_(idx);
  return NULL;
}

void taGroup_impl::Duplicate(const taGroup_impl& cp) {
  taList_impl::Duplicate(cp);
  gp.Duplicate(cp.gp);
}
void taGroup_impl::DupeUniqNameOld(const taGroup_impl& cp) {
  taList_impl::DupeUniqNameOld(cp);
  gp.DupeUniqNameOld(cp.gp);
}
void taGroup_impl::DupeUniqNameNew(const taGroup_impl& cp) {
  taList_impl::DupeUniqNameNew(cp);
  gp.DupeUniqNameNew(cp.gp);
}
void taGroup_impl::Borrow(const taGroup_impl& cp) {
  taList_impl::Borrow(cp);
  gp.Borrow(cp.gp);
}
void taGroup_impl::BorrowUnique(const taGroup_impl& cp) {
  taList_impl::BorrowUnique(cp);
  gp.BorrowUnique(cp.gp);
}
void taGroup_impl::BorrowUniqNameOld(const taGroup_impl& cp) {
  taList_impl::BorrowUniqNameOld(cp);
  gp.BorrowUniqNameOld(cp.gp);
}
void taGroup_impl::BorrowUniqNameNew(const taGroup_impl& cp) {
  taList_impl::BorrowUniqNameNew(cp);
  gp.BorrowUniqNameNew(cp.gp);
}
void taGroup_impl::Copy_Common(const taGroup_impl& cp) {
  taList_impl::Copy_Common(cp);
  gp.Copy_Common(cp.gp);
}
void taGroup_impl::Copy_Duplicate(const taGroup_impl& cp) {
  taList_impl::Copy_Duplicate(cp);
  gp.Copy_Duplicate(cp.gp);
}
void taGroup_impl::Copy_Borrow(const taGroup_impl& cp) {
  taList_impl::Copy_Borrow(cp);
  gp.Copy_Borrow(cp.gp);
}
void taGroup_impl::List(ostream& strm) const {
  taList_impl::List(strm);
  int i;
  for(i=0; i<gp.size; i++)
    FastGp_(i)->List(strm);
}
