/* -*- C++ -*- */
/*=============================================================================
//									      //
// This file is part of the TypeAccess/C-Super-Script software package.	      //
//									      //
// Copyright (C) 1995 Randall C. O'Reilly, Chadley K. Dawson, 		      //
//		      James L. McClelland, and Carnegie Mellon University     //
//     									      //
// Permission to use, copy, modify, and distribute this software and its      //
// documentation for any purpose is hereby granted without fee, provided that //
// the above copyright notice and this permission notice appear in all copies //
// of the software and related documentation.                                 //
// 									      //
// Note that the PDP++ software package, which contains this package, has a   //
// more restrictive copyright, which applies only to the PDP++-specific       //
// portions of the software, which are labeled as such.			      //
//									      //
// Note that the taString class, which is derived from the GNU String class,  //
// is Copyright (C) 1988 Free Software Foundation, written by Doug Lea, and   //
// is covered by the GNU General Public License, see ta_string.h.             //
// The iv_graphic library and some iv_misc classes were derived from the      //
// InterViews morpher example and other InterViews code, which is             //
// Copyright (C) 1987, 1988, 1989, 1990, 1991 Stanford University             //
// Copyright (C) 1991 Silicon Graphics, Inc.				      //
//									      //
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,         //
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 	      //
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  	      //
// 									      //
// IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE FOR ANY SPECIAL,    //
// INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES  //
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT     //
// ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY,      //
// ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS        //
// SOFTWARE. 								      //
==============================================================================*/

#ifndef ta_group_h
#define ta_group_h 1

#include <ta/typea.h>
#include <ta/ta_base.h>

#ifdef List			// prevent interference from iv
#undef List
#endif


class 	taGroup_impl;		// pre-declare
typedef taGroup_impl* 	TAGPtr;

typedef taList<taGroup_impl> TALOG; // list of groups (LOG)

class 	taSubGroup : public TALOG {
  // #INSTANCE ##NO_TOKENS ##NO_UPDATE_AFTER has the sub-groups for a group
public:
  void	Dirty();

  void	Initialize()	{ };
  void	Destroy()	{ };
  TA_BASEFUNS(taSubGroup);
};

class	taLeafItr {		// contains the indicies for iterating over leafs
public:
  TAGPtr	cgp;		// pointer to current group
  int		g;		// index of current group
  int		i;		// index of current leaf element
};

#define FOR_ITR_EL(T, el, grp, itr) \
for(el = (T*) grp ## FirstEl(itr); el; el = (T*) grp ## NextEl(itr))

#define FOR_ITR_GP(T, el, grp, itr) \
for(el = (T*) grp ## FirstGp(itr); el; el = (T*) grp ## NextGp(itr))

// move some of the iv interface stuff into this guy..

class taGroup_impl : public taList_impl {
  // #INSTANCE #NO_TOKENS #NO_UPDATE_AFTER implementation of a group
protected:
  virtual TAGPtr GetSuperGp_();			// Parent super-group, or NULL
  virtual void	 UpdateLeafCount_(int no); 	// updates the leaves count

public:
  int 		leaves;		// #READ_ONLY #NO_SAVE total number of leaves
  taSubGroup	gp; 		// #HIDDEN #NO_FIND #NO_SAVE sub-groups within this one
  TAGPtr	super_gp;	// #READ_ONLY #NO_SAVE super-group above this
  TALOG* 	leaf_gp; 	// #READ_ONLY #NO_SAVE 'flat' list of leaf-containing-gps 

  void		Dirty();		// when anything changes..
  bool		IsEmpty()	{ return (leaves == 0) ? true : false; }

  TAPtr 	New(int no=0, TypeDef* typ = NULL); // New based on Type

  MemberDef* 	FindMembeR(const char* nm, void*& ptr) const;    // extended to search in the group
  MemberDef* 	FindMembeR(TypeDef* it, void*& ptr) const; // extended to search in the group

  // IO routines
  ostream& 	OutputR(ostream& strm, int indent = 0) const;

  int		Dump_SaveR(ostream& strm, TAPtr par=NULL, int indent=0);
  int		Dump_Save_PathR_impl(ostream& strm, TAPtr par=NULL, int indent=0);

  ////////////////////////////////////////////////
  // 	functions that return the type		//
  ////////////////////////////////////////////////

  TAGPtr	 Gp_(int i) const	{ return gp.El(i); } 	 // #IGNORE 
  TAGPtr 	 FastGp_(int i) const	{ return gp.FastEl(i); } // #IGNORE 
  virtual TAPtr  Leaf_(int idx) const;	// #IGNORE DFS through all subroups for leaf i
  virtual TAGPtr LeafGp_(int idx) const; // #IGNORE the group having this leaf in it

  // iterator-like functions
  TAGPtr 	FirstGp_(int& g) const	// #IGNORE first sub-gp
  { g = 0; if(leaf_gp == NULL) InitLeafGp(); return leaf_gp->El(0); }
  TAGPtr 	NextGp_(int& g)	const	// #IGNORE next sub-gp
  { return leaf_gp->El(++g); }

  TAPtr	 	FirstEl_(taLeafItr& lf) const	// #IGNORE first leaf
  { TAPtr rval=NULL; lf.i = 0; lf.cgp = FirstGp_(lf.g);
    if(lf.cgp != NULL) rval=(TAPtr)lf.cgp->el[0]; return rval; }
  TAPtr	 	NextEl_(taLeafItr& lf)	const	// #IGNORE next leaf
  { TAPtr rval=NULL; if(++lf.i >= lf.cgp->size) {
    lf.i = 0; lf.cgp = leaf_gp->El(++lf.g); }
    if(lf.cgp != NULL) rval = (TAPtr)lf.cgp->el[lf.i]; return rval; }

  virtual TAGPtr NewGp_(int no, TypeDef* typ=NULL);	// #IGNORE create sub groups
  virtual TAPtr	 NewEl_(int no, TypeDef* typ=NULL);	// #IGNORE create sub groups

  virtual TAPtr FindLeafName_(const char* it, int& idx=Idx) const; 	// #IGNORE
  virtual TAPtr	FindLeafType_(TypeDef* it, int& idx=Idx) const;	// #IGNORE


  ////////////////////////////////////////////////
  // functions that don't depend on the type	//
  ////////////////////////////////////////////////

  virtual void	InitLeafGp() const;		// Initialize the leaf group
  virtual void	InitLeafGp_impl(TALOG* lg) const; // #IGNORE impl of init leaf gp
  virtual void	AddEl_(void* it); 		// #IGNORE update leaf count
  virtual bool	Remove(const char* item_nm)	{ return taList_impl::Remove(item_nm); }
  virtual bool	Remove(TAPtr item)		{ return taList_impl::Remove(item); }
  virtual bool	Remove(int idx);

  virtual bool 	RemoveLeaf(TAPtr item); 	// remove given leaf element
  virtual bool	RemoveLeaf(const char* item_nm);
  virtual bool  RemoveLeaf(int idx);	
  // Remove leaf element at leaf index
  virtual bool	RemoveLeafName(const char* item_nm)	{ return RemoveLeaf(item_nm); }
  // remove given named leaf element 
  virtual bool	RemoveLeafEl(TAPtr item)		{ return RemoveLeaf(item); }
  // Remove given leaf element
  virtual void 	RemoveAll();
  // Remove all elements of the group

  virtual bool	RemoveGp(int idx) 			{ return gp.Remove(idx); }
  // remove group at given index
  virtual bool	RemoveGp(TAGPtr group)			{ return gp.Remove(group); }
  // #MENU #FROM_GROUP_gp #MENU_ON_Edit #UPDATE_MENUS remove given group 

  virtual void	EnforceLeaves(int sz);
  // ensure that sz leaves exits by adding new ones to top group and removing old ones from end
  void	EnforceSameStru(const taGroup_impl& cp);

  int	ReplaceType(TypeDef* old_type, TypeDef* new_type);

  virtual int	FindLeaf(TAPtr item) const;  // find given leaf element (-1 = not here)
  virtual int	FindLeaf(TypeDef* item) const;
  virtual int	FindLeaf(const char* item_nm) const;
  // find named leaf element
  virtual int	FindLeafEl(TAPtr item) const	{ return FindLeaf(item); }
  // find given leaf element -1 = not here. 

  void	Duplicate(const taGroup_impl& cp);
  void	DupeUniqNameOld(const taGroup_impl& cp);
  void	DupeUniqNameNew(const taGroup_impl& cp);
  
  void	Borrow(const taGroup_impl& cp);
  void	BorrowUnique(const taGroup_impl& cp);
  void	BorrowUniqNameOld(const taGroup_impl& cp);
  void	BorrowUniqNameNew(const taGroup_impl& cp);

  void	Copy_Common(const taGroup_impl& cp);
  void	Copy_Duplicate(const taGroup_impl& cp);
  void	Copy_Borrow(const taGroup_impl& cp);

  virtual void 	List(ostream& strm=cout) const; // Display list of elements in the group

  void 	Initialize();
  void 	InitLinks();		// inherit the el_typ from parent group..
  void  Destroy();
  void 	Copy(const taGroup_impl& cp);
  TA_BASEFUNS(taGroup_impl);
};


#include <ta/ta_group_tmplt.h>

#endif // ta_group_h


