/* -*- C++ -*- */
/*=============================================================================
//									      //
// This file is part of the TypeAccess/C-Super-Script software package.	      //
//									      //
// Copyright (C) 1995 Randall C. O'Reilly, Chadley K. Dawson, 		      //
//		      James L. McClelland, and Carnegie Mellon University     //
//     									      //
// Permission to use, copy, modify, and distribute this software and its      //
// documentation for any purpose is hereby granted without fee, provided that //
// the above copyright notice and this permission notice appear in all copies //
// of the software and related documentation.                                 //
// 									      //
// Note that the PDP++ software package, which contains this package, has a   //
// more restrictive copyright, which applies only to the PDP++-specific       //
// portions of the software, which are labeled as such.			      //
//									      //
// Note that the taString class, which is derived from the GNU String class,  //
// is Copyright (C) 1988 Free Software Foundation, written by Doug Lea, and   //
// is covered by the GNU General Public License, see ta_string.h.             //
// The iv_graphic library and some iv_misc classes were derived from the      //
// InterViews morpher example and other InterViews code, which is             //
// Copyright (C) 1987, 1988, 1989, 1990, 1991 Stanford University             //
// Copyright (C) 1991 Silicon Graphics, Inc.				      //
//									      //
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,         //
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 	      //
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  	      //
// 									      //
// IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE FOR ANY SPECIAL,    //
// INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES  //
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT     //
// ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY,      //
// ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS        //
// SOFTWARE. 								      //
==============================================================================*/

// glyphs to graph by

#include <ta_misc/datagraph.h>
#include <ta_misc/fontspec.h> // for viewlabels
#include <ta_misc/colorbar.h> // for dynamic background

#include <iv_graphic/graphic.h>
#include <iv_graphic/graphic_viewer.h>
#include <iv_misc/dastepper.h>

#include <ta/enter_iv.h>
#include <InterViews/font.h>
#include <InterViews/action.h>
#include <InterViews/background.h>
#include <InterViews/box.h>
#include <InterViews/brush.h>
#include <InterViews/canvas.h>
#include <InterViews/color.h>
#include <InterViews/display.h>
#include <InterViews/label.h>
#include <InterViews/session.h>
#include <InterViews/style.h>
#include <InterViews/telltale.h>
#include <InterViews/geometry.h>
#include <InterViews/window.h> // for setting zoom_cursor
#include <IV-look/button.h>
#include <IV-look/dialogs.h>
#include <IV-look/kit.h>
#include <ta/leave_iv.h>

#include <iv_misc/requestbox.h>
#include <iv_misc/tbScrollBox.h>
#include <iv_misc/grid.h>
#include <ta_misc/axis.h>

#define GRAPH_TOP_BORDER  5.0
#define GRAPH_BOT_BORDER 10.0
#define GRAPH_LFT_BORDER 10.0
#define GRAPH_RGT_BORDER 15.0


//////////////////////////
// 	GraphButton	//
//////////////////////////
 
GraphButton::GraphButton(ivGlyph* g, ivStyle* s, ivTelltaleState* t, GraphEditor* o,
			 DA_GraphViewSpec* gvs, int f)
: ivButton(g,s,t,NULL)
{
  axis = NULL;
  axis_patch = NULL;
  field = f;
  owner = o;
  spec = gvs;

  ivLayoutKit* layout = ivLayoutKit::instance();

  state()->set(ivTelltaleState::is_choosable | 
	       ivTelltaleState::is_toggle |
	       ivTelltaleState::is_enabled | 
	       ivTelltaleState::is_chosen, true);
  
  // reparent the glyph in the surrounding goo
  body(layout->hbox(layout->margin(GetLegend(),2), 
                    taivM->big_button(g)));
  SetAxis();
}
 
GraphButton::~GraphButton() {
}

void GraphButton::SetAxis(){
  ivLayoutKit* layout = ivLayoutKit::instance();
  ivWidgetKit* wkit = ivWidgetKit::instance();
  ivGlyph* axis_rep;
  if(owner->x_axis_index == field) { // x-axis
    axis_rep = layout->hfixed
      (layout->hflexible
       (wkit->label("-X-")),18);
    state()->set(ivTelltaleState::is_chosen, false);
    state()->set(ivTelltaleState::is_active, true);
    spec->visible = false;
  }
  else 
    axis_rep = layout->margin(GetLegend(spec->axis_spec),2);

  if(body()->count() == 3)
    body()->replace(2,axis_rep);
  else
    body()->append(axis_rep); // first time only
}

// legend is 16 coords wide
ivGlyph* GraphButton::GetLegend(DA_GraphViewSpec* vspec){
  ivLayoutKit* layout = ivLayoutKit::instance();
  if(vspec==NULL) vspec = spec;
  ivGlyph* line = NULL;
  ivGlyph* points = NULL;
  if((vspec->line_type == DA_GraphViewSpec::LINE) ||
     (vspec->line_type == DA_GraphViewSpec::LINE_AND_POINTS)){
    switch(vspec->line_style){
    case DA_GraphViewSpec::SOLID:
      line = layout->vmargin
	(layout->fixed(new ivBackground(NULL,vspec->ta_color.color),16,1),2);
      break;
    case DA_GraphViewSpec::DOT:
      line = layout->hbox
	(layout->hfixed(layout->hmargin(NULL,1),1),
	 layout->margin
	 (layout->fixed(new ivBackground(NULL,vspec->ta_color.color),1,1),1,2),
	 layout->margin
	 (layout->fixed(new ivBackground(NULL,vspec->ta_color.color),1,1),1,2),
	 layout->margin
	 (layout->fixed(new ivBackground(NULL,vspec->ta_color.color),1,1),1,2),
	 layout->margin
	 (layout->fixed(new ivBackground(NULL,vspec->ta_color.color),1,1),1,2),
	 layout->margin
	 (layout->fixed(new ivBackground(NULL,vspec->ta_color.color),1,1),1,2));
      break;
    case DA_GraphViewSpec::DASH:
      line = layout->hbox
	(layout->margin
	  (layout->fixed(new ivBackground(NULL,vspec->ta_color.color),4,1),2,2),
	 layout->margin
	 (layout->fixed(new ivBackground(NULL,vspec->ta_color.color),4,1),2,2));
      break;
    case DA_GraphViewSpec::DASH_DOT:
      line = layout->hbox
	(layout->margin
	 (layout->fixed(new ivBackground(NULL,vspec->ta_color.color),4,1),2,2),
	 layout->margin
	 (layout->fixed(new ivBackground(NULL,vspec->ta_color.color),2,1),3,2));
      break;
    }
  }
  
  if((vspec->line_type == DA_GraphViewSpec::POINTS) ||
     (vspec->line_type == DA_GraphViewSpec::LINE_AND_POINTS)){
    switch(vspec->point_style){
    case DA_GraphViewSpec::SMALL_DOT:
      points = layout->hbox
	(layout->margin
	 (layout->fixed(new ivBackground(NULL,vspec->ta_color.color),2,2),1,1),
	 layout->margin
	 (layout->fixed(new ivBackground(NULL,vspec->ta_color.color),2,2),1,1),
	 layout->margin
	 (layout->fixed(new ivBackground(NULL,vspec->ta_color.color),2,2),1,1),
	 layout->margin
	 (layout->fixed(new ivBackground(NULL,vspec->ta_color.color),2,2),1,1));
      break;
    case DA_GraphViewSpec::BIG_DOT:
      points = layout->hbox
	(layout->margin
	 (layout->fixed(new ivBackground(NULL,vspec->ta_color.color),3,3),
	  2,3,1,1),
	 layout->margin
	 (layout->fixed(new ivBackground(NULL,vspec->ta_color.color),3,3),
	  2,3,1,1));
      break;
    case DA_GraphViewSpec::TICK:
      points = layout->hbox
	(layout->hmargin
	 (layout->fixed(new ivBackground(NULL,vspec->ta_color.color),2,5),3),
	 layout->hmargin
	 (layout->fixed(new ivBackground(NULL,vspec->ta_color.color),2,5),3));
      break;
    case DA_GraphViewSpec::PLUS:
      points = layout->overlay
	(layout->hbox
	 (layout->hmargin
	  (layout->fixed(new ivBackground(NULL,vspec->ta_color.color),2,5),3),
	  layout->hmargin
	  (layout->fixed(new ivBackground(NULL,vspec->ta_color.color),2,5),3)),
	 layout->hbox
	 (layout->margin
	  (layout->fixed(new ivBackground(NULL,vspec->ta_color.color),6,3),1,1),
	  layout->margin
	  (layout->fixed
	   (new ivBackground(NULL,vspec->ta_color.color),6,3),1,2)));
      break;
    case DA_GraphViewSpec::NONE: break;
    }
  }
  return layout->overlay(line,points);
}

GraphButton* GraphButton::GetButton(GraphEditor* o,
				    DA_GraphViewSpec* gvs,int f){
  ivWidgetKit* wkit = ivWidgetKit::instance();

  ivTelltaleState* ts = new ivTelltaleState;
  ivGlyph* look = wkit->palette_button_look(wkit->label(gvs->display_name), ts);
  return new GraphButton(look, wkit->style(), ts, o, gvs, f);
};


bool GraphButton::IsChosen() {
  return state()->test(ivTelltaleState::is_chosen);
}

void GraphButton::release(const ivEvent& e) {
  if(taivM->GetButton(e) == ivEvent::middle) {
    SetXAxis();
    owner->graphg->Reset();
    tabMisc::NotifyEdits(spec);
    tabMisc::NotifyEdits(owner->owner); // the graphlogview
    return;
  }
 if(taivM->GetButton(e) == ivEvent::right) {
   osboolean chosen = state()->test(ivTelltaleState::is_chosen);
   ivButton::release(e);  
   if(owner->x_axis_index == field) {
     state()->set(ivTelltaleState::is_active, true);
   }
   state()->set(ivTelltaleState::is_chosen, chosen);
   spec->Edit();
   return;
  }
  if(owner->x_axis_index != field) { // don't do x-axis
    ivButton::release(e); 
    ToggleLine();
    tabMisc::NotifyEdits(spec);
  }
}

void GraphButton::SetXAxis() {
  if(owner==NULL) return;
  owner->SetXAxis(field);
}
  

void GraphButton::ToggleLine() {
  if(owner==NULL) return;
  owner->ToggleLine(field);
}


//////////////////////////
// 	GraphLine	//
//////////////////////////
 
GraphLine::GraphLine(Grapher_G* gg,   DA_GraphViewSpec* dvs,
		     DataTable* t, int index, float_Data* xax_ar, 
		     Axis* ax, ivPatch* axpat, Axis* xax, ivColor* c)
: Graphic(gg->brush(), c, NULL, NULL, false, false, 1, NULL)
{
  owner = gg;
  spec = dvs;
  data = t;
  dt_index = index;
  labels = NULL;
  array = NULL;
  DataArray_impl* da = data->Leaf(dt_index);
  if(da->InheritsFrom(&TA_float_Data))
    array = (float_Data*)da;
  else {			// a string!
    labels = (String_Data*)da;
    da = data->Leaf(dt_index - 1); // prior float = y axis values
    if((da != NULL) && da->InheritsFrom(&TA_float_Data))
      array = (float_Data*)da;
    else
      labels = NULL;
  }
  x_axis_ar = xax_ar;
  
  axis = ax;
  x_axis = xax;
  axis_patch = axpat;

  // todo: brush width should be wider for printouts, but it gets distorted
  // according to geometry of screen -- need to print out square maybe..
  // and widths vary dramatically depending on geometry to, so const. is
  // not appropriate..

  last_size = 0;
  switch(spec->line_style) {
  case  DA_GraphViewSpec::SOLID:
    brush(new ivBrush(BRUSH_SOLID, DEFAULT_BRUSH_WIDTH)); break;
  case  DA_GraphViewSpec::DOT:
    brush(new ivBrush(BRUSH_DOT, DEFAULT_BRUSH_WIDTH)); break;
  case  DA_GraphViewSpec::DASH:
    brush(new ivBrush(BRUSH_DASH, DEFAULT_BRUSH_WIDTH)); break;
  case  DA_GraphViewSpec::DASH_DOT:
    brush(new ivBrush(BRUSH_DASH_DOT, DEFAULT_BRUSH_WIDTH)); break;
  }
}

GraphLine::~GraphLine() {
  x_targets.Reset();
  y_targets.Reset();
  remove_all();
}

void GraphLine::Reset() {
  ctrlpts(NULL,NULL,0);	// delete the coord space
  _ctrlpts = 0;
  _buf_size = Graphic::buf_size; // this should be set in Graphic::ctrlpts()
				   // since _x and _y get reallocated to
				   // this size there.
  last_size = 0;
  x_targets.Reset();
  y_targets.Reset();
  lab_cache.Reset();
}

bool GraphLine::HaveData() {
  if((array == NULL) || (array->ar == NULL) || (array->ar->size <= 0))
    return false;
  if((x_axis_ar == NULL) || (x_axis_ar->ar == NULL) || (x_axis_ar->ar->size <= 0))
    return false;
  if((axis == NULL) || (axis_patch == NULL) || (x_axis == NULL))
    return false;
  return true;
}
  
void GraphLine::Build() {
  Reset();
  AddAllPoints();
}

void GraphLine::drawclipped_gs (
    ivCanvas* c, ivCoord l, ivCoord b, ivCoord r, ivCoord t, Graphic* gs
) {
    ivCoord ll, bb, rr, tt;
    getbounds_gs(ll, bb, rr, tt, gs);

    BoxObj thisBox(ll, bb, rr, tt);
    BoxObj clipBox(l, b, r, t);
    if (clipBox.Intersects(thisBox)) {
//  let the canvas take care of the clipping
//      intersect_box._left = clipBox._left;
//      intersect_box._right = clipBox._right;
//      intersect_box._top = clipBox._top;
//      intersect_box._bottom = clipBox._bottom;
      draw_gs(c, gs);
    }
}

void GraphLine::draw_gs(ivCanvas* c, Graphic* gs){
  if(_ctrlpts <= 0)
    return;
  const ivBrush* b =  brush();
  const ivColor* stroke = gs->stroke();
  ivTransformer* tx = gs->transformer();
  if (tx != nil) {
    c->push_transform();
    c->transform(*tx);
  }
  int i =0;

  if((labels != NULL) && (array != NULL)) {		// plotting strings!
    ivWidgetKit* wkit = ivWidgetKit::instance();
    ivFontBoundingBox bbox;
    wkit->font()->font_bbox(bbox);
    float fx = bbox.width()/2.0;
    float fy = bbox.ascent() + bbox.descent();
    int mxidx = MIN(lab_cache.size, _ctrlpts);
    for (i = 0; i < mxidx; ++i) {
      if(lab_cache.FastEl(i).empty()) continue; // don't plot empty strings!
      String str = lab_cache.FastEl(i);
      float ax = _x[i];	float ay = _y[i];
      float nx,ny;
      c->transformer().transform(ax, ay,nx,ny);
      c->push_transform(); // store the line*net transform;
      ivTransformer nt = c->transformer();
      nt.invert();		// invert current transformer
      c->transform(nt);		// put inverse in there -- effectively null xform
      int len = str.length();
      nx += 3.0f;
      ny -= 3.0f;
      float cx = nx;
      int j;
      for(j=0;j<len;j++){	// draw each character
	c->character(wkit->font(),str[j],8,stroke,cx,ny);
	wkit->font()->char_bbox(str[j], bbox);
	cx += bbox.width();
      }
      c->damage(nx,ny,ny-fy,nx+(j+2)*fx);
      c->pop_transform(); // remove null transform
    }
  }
  else {			// plotting lines!
    if((spec->line_type == DA_GraphViewSpec::LINE) ||
       (spec->line_type == DA_GraphViewSpec::LINE_AND_POINTS)){
      c->new_path();
      c->move_to(_x[0], _y[0]);
      for (i = 1; i < _ctrlpts; ++i) {

	float tx,ty;
	c->transformer().transform(_x[i],_y[i],tx,ty);

	if((spec->negative_draw == false) && 
	   (i > 0) && (_x[i] < _x[i-1])){
	  c->stroke(stroke,b);
	  c->move_to(_x[i],_y[i]);
	}
	else {
	  c->line_to(_x[i], _y[i]);
	}
      }
      if (b != nil && stroke != nil) {
	c->stroke(stroke, b);
      }
    }
    ivCoord a00,a01,a10,a11,a20,a21;
    c->transformer().matrix(a00,a01,a10,a11,a20,a21);
    ivCoord pwidth = c->to_coord(1)/a00;
    ivCoord pheight= c->to_coord(1)/a11;
    if((spec->line_type == DA_GraphViewSpec::POINTS) ||
       (spec->line_type == DA_GraphViewSpec::LINE_AND_POINTS)){
      c->new_path();
      c->move_to(_x[0], _y[0]);
      for (i = 0; i < _ctrlpts; ++i) {
	if((i - spec->point_mod.off) % (spec->point_mod.m) != 0) continue;
	// draw point
	switch(spec->point_style){
	case DA_GraphViewSpec::SMALL_DOT:
	  c->fill_rect(_x[i]-pwidth,_y[i]-pheight,_x[i]+pwidth,_y[i]+pheight,stroke);
	  break;
	case DA_GraphViewSpec::BIG_DOT:
	  c->fill_rect(_x[i]-pwidth*2,_y[i]-pheight*2,
		       _x[i]+pwidth*2,_y[i]+pheight*2,stroke);
	  break;
	case DA_GraphViewSpec::PLUS:
	  c->new_path();
	  c->move_to(_x[i]-pwidth*2,_y[i]);
	  c->line_to(_x[i]+pwidth*2,_y[i]);
	  c->stroke(stroke,gs->brush());
	case DA_GraphViewSpec::TICK:
	  c->new_path();
	  c->move_to(_x[i],_y[i]-pheight*2);
	  c->line_to(_x[i],_y[i]+pheight*2);
	  c->stroke(stroke,gs->brush());
	  break;
	case DA_GraphViewSpec::NONE: break;
	}
      }
    }
    if(x_targets.size > 0) {
      ivWidgetKit* wkit = ivWidgetKit::instance();
      stroke = owner->dtvsp->bg_color.contrastcolor;
      ivFontBoundingBox bbox;
      wkit->font()->font_bbox(bbox);
      float fx=bbox.width()/2.0;
      float fy= bbox.ascent() + bbox.descent();
      float_RArray* yfar;
      float maxx = 1.0f;
      float minx = 0.0f;
      DA_GraphViewSpec* xspec = (DA_GraphViewSpec *)
	owner->dtvsp->Leaf(owner->owner->x_axis_index);
      if(xspec != NULL) {
	float_RArray* xfar = (float_RArray *) xspec->data_array->AR();
	maxx = xfar->range.max;
	minx = xfar->range.min;
      }
      yfar = (float_RArray *) spec->data_array->AR();
      float maxy = yfar->range.max;
      float miny = yfar->range.min;

      for(i=0;i<x_targets.size;i++) {
	ivCoord ax = x_targets[i];
	ivCoord ay = y_targets[i];
	// draw little box around point
	c->rect(ax-pwidth*3,ay-pheight*3,
		ax+pwidth*3,ay+pheight*3,stroke,gs->brush());
	String num(ax, "(%g , ");
	num += String(ay, "%g)");

	// determine range and position of the line
	// then adjust the placement of the label as to not go off the edge
	// of the viewed area
      
	float nx=0; // these are in untransformed canvas coords
	float ny=0;
	c->transformer().transform(ax,ay,nx,ny);
	c->push_transform(); // store the line*net transform;
	ivTransformer nt = c->transformer();
	nt.invert();			// invert current transformer
	c->transform(nt);		// put inverse in there -- effectively null xform
	int len = num.length();
	if(xspec != NULL) {
	  float xpc = - ((ax-minx) / (maxx-minx));
	  nx = nx + (xpc * (len+1)* fx);
	}
	float ypc = - ((ay-miny) / (maxy-miny));
	ny = ny + (ypc * fy);
	int j;
	for(j=0;j<len;j++){
	  // draw each character
	  c->character(wkit->font(),num[j],8,stroke,nx+((j+1)*fx),ny+5);
	}
	c->damage(nx,ny,ny-fy,nx+(j+2)*fx);
	c->pop_transform(); // remove null transform
      }
    }
  }
  if (tx != nil) {
    c->pop_transform();
  }
}

bool	GraphLine::grasp_move(const ivEvent& e, Tool&,ivCoord ix, ivCoord iy) {
  x_targets.Add(0);
  y_targets.Add(0);
  float ax = 0;
  float ay = 0;
  transformer()->inverse_transform(ix,iy,ax,ay);
  // button is stored from the grasp since the move event does not seem
  // to contain the propper button?
  last_button = taivM->GetButton(e);
  double d = sqrt((double) (((_x[0]-ax) * (_x[0]-ax)) +
			    ((_y[0]-ay) * (_y[0]-ay))));
  int target =0;
  for (int i = 1; i < _ctrlpts; ++i) {
    double dist = sqrt((double) (((_x[i]-ax) * (_x[i]-ax)) +
				 ((_y[i]-ay) * (_y[i]-ay))));
    if(dist < d) {
      d = dist;
      target = i;
    }
  }

  x_targets[x_targets.size -1] = _x[target];
  y_targets[y_targets.size -1] = _y[target];
  last_index = target;
  return true;
}

bool GraphLine::manip_move(const ivEvent& , Tool&,ivCoord , ivCoord ,
			      ivCoord , ivCoord ,
			      ivCoord cx, ivCoord cy) {
  float nx=0;
  float ny=0;
  transformer()->inverse_transform(cx,cy,nx,ny);

  // scan outward from the last point
  float d = fabs(_x[last_index]-nx);
  int target = last_index;

  for (int i = _ctrlpts/2; i >=0 ; i--) {
    if((last_index + i) < _ctrlpts){
      if(fabs(_x[last_index+i]-nx) < d){
	d = fabs(_x[last_index+i]-nx);
	target = last_index + i;
	continue;
      }
    }
    if((last_index - i) < _ctrlpts){
      if(fabs(_x[last_index-i]-nx) < d){
	d = fabs(_x[last_index-i]-nx);
	target = last_index - i;
      }
    }
  }
  if(target < 0) target = 0;
  if(target >= _ctrlpts) target = _ctrlpts-1;
  last_index = target;

  if(last_button == ivEvent::middle){ // interpolate

    // limit point to be within the x-range of the line
    DA_GraphViewSpec* xspec = (DA_GraphViewSpec *)
      owner->dtvsp->Leaf(owner->owner->x_axis_index);
    if(xspec != NULL) {
      float_RArray* xfar = (float_RArray *) xspec->data_array->AR();
      nx = MAX(xfar->range.min,nx);
      nx = MIN(xfar->range.max,nx);
    }

    float ax = nx;
    int target2;
    if(_x[target] > ax) { target--; target2 = target+1;}
    else target2 = target +1;
    if((target2 >= 0) && (target2 <= _ctrlpts)){
      float maxx = MAX(_x[target],_x[target2]);
      float minx = MIN(_x[target],_x[target2]);
      int maxindex = (_x[target] > _x[target2]) ? target : target2;
      int minindex = (_x[target] < _x[target2]) ? target : target2;
      float ay = _y[minindex] + ((ax - minx)/(maxx-minx)) *
	(_y[maxindex]-_y[minindex]);

      x_targets[x_targets.size -1] = ax;
      y_targets[y_targets.size -1] = ay;
      
      return true;
    }
  }
  x_targets[x_targets.size -1] = _x[target];
  y_targets[y_targets.size -1] = _y[target];
  return true;
}

  
bool GraphLine::effect_move(const ivEvent&, Tool& ,ivCoord , ivCoord ,
				 ivCoord  , ivCoord ) {
  return true;
}

bool GraphLine::UpdateAxis() {
  if(!HaveData()) return false;
  if(SetAxisMinMax()){ // min or max changed
    axis->UpdateAxis();
    axis->reallocate();
    return true;
  }
  return false;
}

void GraphLine::RedrawAxis() {
  if(!HaveData()) return;
//  axis_patch->undraw(); // this set's canvas to null
  axis->UpdateAxis(); // this creates news non-allocated structures
  axis_patch->reallocate(); // if canvas is null (always) this does nothing
			    // resulting in mispositioned numbers
}

int GraphLine::SetAxisMinMax() {
  float min=0.0f;
  float max=0.0f;
  int min_change = false;
  int max_change = false;
  float havedata= false;
  if(spec->axis_spec == NULL)
    taBase::SetPointer((TAPtr*)&(spec->axis_spec),spec);
  if(HaveData()) {
    min = array->ar->range.min;
    max = array->ar->range.max;
    havedata = true;
  }
  axis->fixed_min = false;
  axis->fixed_max = false;
  switch(spec->axis_spec->min_mode) {
  case DA_GraphViewSpec::FIXED:
    if(spec->axis_spec != spec) break;
    axis->fixed_min = true;
    if(axis->true_range.min != spec->range.min){
      min = spec->range.min; min_change = true;
    }
    break;
  case DA_GraphViewSpec::AUTO_SELF:
    if(spec->axis_spec != spec) break;
  case DA_GraphViewSpec::AUTO_GROUP:
    if(min < axis->true_range.min)  {  min_change = true; break; }
    if((spec->axis_spec == spec) && havedata && (array->ar->size == 1)) // first point
      min_change = true; // force a change for the first point
  }
  switch(spec->axis_spec->max_mode){
  case DA_GraphViewSpec::FIXED:
    if(spec->axis_spec != spec) break;
    axis->fixed_max = true;
    if(axis->true_range.max != spec->range.max) {
      max= spec->range.max; max_change = true;
    }
    break;
  case DA_GraphViewSpec::AUTO_SELF:
    if(spec->axis_spec != spec) break;
  case DA_GraphViewSpec::AUTO_GROUP:
    if(max > axis->true_range.max)  {  max_change = true; break; }
    if((spec->axis_spec == spec) && havedata && (array->ar->size == 1)) // first point
      max_change = true; // force a change for the first point
  }
  if((min_change == true) && (max_change == false)) // change only min
    max = axis->true_range.max;
  if((min_change == false) && (max_change == true)) // change only max
    min = axis->true_range.min;
  bool change = (min_change || max_change);
  if(min == max) change = true;
  if(change == true) change = axis->NewRange(min,max);
  return (change);
}

void GraphLine::AddAllPoints() {
  if(!HaveData()) return;
  SetAxisMinMax();
  int max_n = MIN(x_axis_ar->ar->size, array->ar->size);
  int i;
  int xa_i = x_axis_ar->ar->size - max_n;
  int ar_i = array->ar->size - max_n;
  for(i=0; i < max_n; i++, xa_i++, ar_i++) {
    ivCoord x = x_axis_ar->ar->FastEl(xa_i);
    ivCoord y = array->ar->FastEl(ar_i);
    add_point(x,y);
    if(labels != NULL)
      lab_cache.Add(labels->ar->FastEl(ar_i));
  }
  last_size = array->ar->size;
  ivCanvas* can = owner->owner->GetCanvas();
  if(can != NULL) damage_me(can);
}

void GraphLine::AddPoint(int i) { // i is number of points back from end..
  if(!HaveData()) return;

  if((i >= array->ar->size) || (i >= x_axis_ar->ar->size)) return;

  ivCoord x = x_axis_ar->ar->RevEl(i);
  ivCoord y = array->ar->RevEl(i);

  if(array->ar->size == 1) {
    if(spec->axis_spec->min_mode == DA_GraphViewSpec::FIXED) axis->fixed_min = true;
    else axis->fixed_min = false;
    if(spec->axis_spec->max_mode == DA_GraphViewSpec::FIXED) axis->fixed_max = true;
    else axis->fixed_max = false;
    axis->InitRange(y); // initialize range
  }

  add_point(x,y);
  if(labels != NULL)
    lab_cache.Add(labels->ar->RevEl(i));
  ivCanvas* can = owner->owner->GetCanvas();
  if(can != NULL) { // smart damaging of the right spot..
    if((_ctrlpts > 1) && (_t != NULL)) {
      ivCoord px = _x[_ctrlpts - 2]; // previous x
      ivCoord py = _y[_ctrlpts - 2];
      Graphic gs;
      total_gs(gs);
      corners(px,py,x,y,*(gs.transformer()));
      px -= 1.0;
      py -= 1.0;
      x += 1.0;
      y += 1.0;
      can->damage(px,py,x,y);
    }
  }
}

void GraphLine::AddLastPoint() {
  if(!HaveData()) return;
  
  if(array->ar->size != last_size + 1) {
    Build();			// size changed (probably shifted buffer)
  }
  else {
    AddPoint(0);
    last_size = array->ar->size;
  }
}


bool GraphLine::effect_border() {
  return true;
}

//////////////////////////
// 	Grapher_G	//
//////////////////////////
 
Grapher_G::Grapher_G(GraphEditor* ge, DataTable* t, DT_GraphViewSpec* dtvs) {
  owner = ge;
  data = t;
  dtvsp = dtvs;
  brush(new ivBrush((ivCoord)0.0));
  ivTransformer* tx = new ivTransformer;
  transformer(tx);
  ivResource::unref(tx);
  x_axis = NULL;
  x_axis_ar = NULL;
  x_axis_but = NULL;
  x_axis_patch = NULL;
}

Grapher_G::~Grapher_G() {
  remove_all();
  lines.Reset();
  if (x_axis_but) { ivResource::unref(x_axis_but); x_axis_but = NULL; }
}

void Grapher_G::ClearAllocation(){
  ivAllocation a;
  _a = a;
}

void Grapher_G::Reset() {
  remove_all();
  lines.Reset();
  no_text_extent = true;
  owner->viewer->root(this);
  no_text_extent = false;
  background(owner->dtvsp->bg_color.color);
  DynamicBackground* bg =   ((DynamicBackground *) owner->boxpatch->body());
  bg->SetColor(dtvsp->bg_color.color);
}

int Grapher_G::FixXAxis(){
  if((x_axis_ar == NULL) || (x_axis_but == NULL) || (x_axis == NULL))
    return false;
  DA_GraphViewSpec* xdvs = (DA_GraphViewSpec *)
    dtvsp->Leaf(owner->x_axis_index);
  float min = 0.0f;
  float max = 0.0f;

  x_axis->fixed_min = false;
  if(xdvs->min_mode == DA_GraphViewSpec::FIXED) {
    x_axis->fixed_min = true;
    min = xdvs->range.min;
  }
  else if(x_axis_ar->ar != NULL) // only one line on  X
    min = x_axis_ar->ar->range.min;

  x_axis->fixed_max = false;
  if(xdvs->max_mode == DA_GraphViewSpec::FIXED) {
    x_axis->fixed_min = true;
    max = xdvs->range.max;
  }
  else if(x_axis_ar->ar != NULL) // no Group on X
    max = x_axis_ar->ar->range.max;

  if(x_axis->NewRange(min,max)) {
    x_axis->UpdateAxis();
    x_axis->reallocate();
//    x_axis_patch->reallocate();
    x_axis_patch->redraw();
    return true;
  }
  return false;
}  

ivColor* GraphGetTextColor(void* obj){
  return ((Grapher_G*) obj)->GetTextColor();
}

void Grapher_G::Build() {
  if((owner == NULL) || (dtvsp == NULL))   return;
  remove_all();
  lines.Reset();
  
  SetXAxis(owner->x_axis_index);
  FixXAxis();
  int cnt = 0;
  taLeafItr vi;
  DA_GraphViewSpec* dvs;
  for(dvs = (DA_GraphViewSpec*)dtvsp->FirstEl(vi); dvs;
      dvs = (DA_GraphViewSpec*)dtvsp->NextEl(vi), cnt++)
  {
    if((dvs == NULL) || !dvs->visible || (cnt == owner->x_axis_index))
      continue;			// don't make a line for invisible items
    GraphButton* but = (GraphButton*)owner->buttons[cnt];
    if((but == NULL) || (but->axis == NULL) || (but->axis_patch == NULL) ||
       (x_axis_but == NULL))  continue;
    GraphLine* gl = new GraphLine(this,dvs,data,cnt,x_axis_ar, but->axis, but->axis_patch,
				  x_axis_but->axis, dvs->ta_color.color);
    lines.Add(gl);
    gl->translate(-(((_a.x_allotment().span()) -10.0f)/2.0f),0);
    gl->Build();
//    gl->RedrawAxis(); // the axis should be okay and is not part of the grapher_g
    append_(gl);
  }
  int i;
  for(i=0; i<lines.size; i++) {
    lines.FastEl(i)->Rescale();
  }
  if((owner ==NULL) || (owner->owner ==  NULL)) return;
  ViewLabel_List* vlg = NULL;
  owner->owner->FindMembeR(&TA_ViewLabel_List,((void *&) vlg));
  if(vlg ==  NULL) return;
  int v;
  for(v=0; v<vlg->size; v++) {
    ViewLabel* vl = vlg->FastEl(v);
    // reset non saved function pointers
    vl->get_color = GraphGetTextColor;
    vl->MakeText();
  }
}

void Grapher_G::AddLastPoint() {
  bool xrescale = FixXAxis();
  bool fixyaxis = false;
  taPtrList_impl rescale_list;
  int i;
  for(i=0; i<lines.size; i++) {
    lines[i]->AddLastPoint();
    if(lines[i]->UpdateAxis()) { // went out of bounds..
      fixyaxis = true;
      if(!xrescale)		// not necessary unless x is not being rescaled
	rescale_list.AddUnique_((void*)lines[i]->spec->axis_spec);
    }
  }
  for(i=0; i<lines.size; i++) {
    // if my spec was on the list, then rescale me
    if(xrescale || (rescale_list.Find_((void*)lines[i]->spec->axis_spec) >= 0))
      lines[i]->Rescale();
  }
  rescale_list.RemoveAll();
  if(fixyaxis || xrescale)  owner->linespace->redraw();
  if(fixyaxis)    owner->y_axis_patch->redraw();
}

void Grapher_G::SetXAxis(int idx) { // for all points
  DataArray_impl* xax = data->Leaf(idx);
  if((xax == NULL) || !(xax->InheritsFrom(&TA_float_Data)))
    x_axis_ar = NULL;
  else
    x_axis_ar = (float_Data*)xax;

  if(x_axis_but != NULL) {
    x_axis_but->state()->set(ivTelltaleState::is_active, false);
    x_axis_but->state()->set(ivTelltaleState::is_chosen, false);
    ivResource::unref(x_axis_but);
  }

  x_axis_but = (GraphButton*)owner->buttons[idx];
  if((x_axis_but == NULL) || (x_axis_but->axis == NULL))
    return;
  ivResource::ref(x_axis_but);
  x_axis_but->state()->set(ivTelltaleState::is_active, true);
  x_axis = x_axis_but->axis;
  x_axis_patch = x_axis_but->axis_patch;
  int i;
  for(i=0; i<lines.size; i++) {
    lines[i]->x_axis_ar = x_axis_ar;
    lines[i]->x_axis = x_axis;
  }
}


void Grapher_G::allocate(ivCanvas* c, const ivAllocation& a, ivExtension& ext) {
  GraphicMaster::allocate(c,a,ext);
  Build(); // todo maybe we should reallocate after build since the lines are new?
}


ivColor* Grapher_G::GetTextColor(){
  return dtvsp->bg_color.contrastcolor;
}

//////////////////////////
// 	GrapherCallback	//
//////////////////////////

typedef void (GlyphViewer::*ToolSetter) (unsigned int);

class GrapherCallback : public ivAction {
public:
  GrapherCallback(GlyphViewer*, ToolSetter, unsigned int);
  virtual void execute();
private:
  GlyphViewer* _gv;
  ToolSetter _tool;
  unsigned _type;
};

GrapherCallback::GrapherCallback (GlyphViewer* gv, ToolSetter tool, unsigned int t) {
  _gv = gv;
  _tool = tool;
  _type = t;
}

void GrapherCallback::execute () {
  GlyphViewer* gv = _gv;
  ToolSetter tool = _tool;
  unsigned int type = _type;
  (gv->*tool)(type);
}

//////////////////////////
// 	GraphViewer	//
//////////////////////////


GraphViewer::GraphViewer(GraphEditor* e,float w, float h, const ivColor* bg)
  : GlyphViewer(w,h,bg)
{
  editor = e;
  init_xform = NULL;
}

GraphViewer::~GraphViewer() {
  if (init_xform) {
    ivResource::unref(init_xform);
    init_xform = NULL;
  }
}

void GraphViewer::root(GraphicMaster* g) {
  GlyphViewer::root(g);
  if(init_xform != NULL) { ivResource::unref(init_xform); init_xform = NULL; }
  init_xform = new ivTransformer(_root->transformer());
  scale_axes();
}

void GraphViewer::allocate(ivCanvas* c, const ivAllocation& a, ivExtension& e){
  _root->transformer(NULL);
  if (a.equals(_a,0.000001)) {
    GlyphViewer::allocate(c,a,e);
    ((Grapher_G *) _root)->Build();
  }
  else {
    ((Grapher_G *) _root)->ClearAllocation(); // this forces root to recenter/recale
    GlyphViewer::allocate(c,a,e); // this does a root->build when it reallocates
				  // root
  }
  if(init_xform != NULL) { ivResource::unref(init_xform); init_xform = NULL; }
  init_xform = new ivTransformer(_root->transformer());
  scale_axes();
}

void GraphViewer::ClearCachedAllocation(){
  ivAllocation a;
  _a = a;
}

void GraphViewer::scale_axes(int redraw){
  if(init_xform==NULL) return;
  if(_root->transformer() == NULL) return;
  //  scaling the axes
  int i;
  for(i=0;i<editor->axis.size;i++){
    Axis* a = (Axis *) editor->axis[i];
    float vmax = 0;
    float vmin = 0;
    float nvmin =0;
    float nvmax =0;
    if(a == editor->graphg->x_axis){ // horizontal axis
      float bogus_y=0; // use vmax,vmin as temps at first;
      init_xform->inverse_transform(_a.left(),0,vmin,bogus_y);
      init_xform->inverse_transform(_a.right(),0,vmax,bogus_y);
      _root->transformer()->transform(vmin,0,nvmin,bogus_y);
      _root->transformer()->transform(vmax,0,nvmax,bogus_y);
      vmin = _a.left(); vmax = _a.right();
    }
    else{
      float bogus_x=0;
      init_xform->inverse_transform(0,_a.bottom(),bogus_x,vmin);
      init_xform->inverse_transform(0,_a.top(),bogus_x,vmax);
      _root->transformer()->transform(0,vmin,bogus_x,nvmin);
      _root->transformer()->transform(0,vmax,bogus_x,nvmax);
      vmin = _a.bottom(); vmax = _a.top();
    }
    float vrange = vmax - vmin;
    float nvrange = (nvmax - nvmin);
    a->min_percent = (nvmin - vmin)/(MAX(vrange,nvrange));
    a->max_percent = (nvmax - vmax)/(MAX(vrange,nvrange));
    a->range_scale = (vrange / nvrange);
    a->UpdateAxis();
    a->reallocate();
    if(redraw == true) a->redraw();
  }
}


void GraphViewer::rate_zoom(){
  GlyphViewer::rate_zoom();
  scale_axes();
}

void GraphViewer::grab_scroll(){
  GlyphViewer::grab_scroll();
  scale_axes(false);
}


//////////////////////////
// 	GraphEditor	//
//////////////////////////

GraphEditor::GraphEditor(TAPtr own,DataTable* d, DT_GraphViewSpec* dtvs, ivWindow* w) {
  owner = own;
  data = d;
  dtvsp = dtvs;
  win = w;

  x_axis_index = 0;
  graphg = NULL;
  viewer = NULL;

  linespace = NULL;
  tool_gp = NULL;
  
  fieldpatch = NULL;
  field_rep = NULL;

  last_size = 0;

  y_axis_patch = x_axis_patch = NULL;

  boxpatch = NULL;
  resize_vbox = NULL;
  resize_vbox2 = NULL;
  resize_hbox = NULL;

  SetOwnerXAxis = NULL;
}
 
GraphEditor::~GraphEditor() {
  buttons.Reset();
  axis.Reset();
  axis_patch.Reset();
  // resource unref already does the nil check for you
  ivResource::unref(graphg);	graphg = NULL;
  ivResource::unref(viewer);	viewer = NULL;
  ivResource::unref(linespace);	linespace = NULL;
  ivResource::unref(resize_hbox);	resize_hbox = NULL;
  ivResource::unref(resize_vbox);	resize_vbox = NULL;
  ivResource::unref(resize_vbox2); resize_vbox2 = NULL;
  ivResource::unref(boxpatch);	boxpatch = NULL;
  ivResource::unref(field_rep);	field_rep = NULL;
  ivResource::unref(fieldpatch);	fieldpatch = NULL;
  ivResource::unref(x_axis_patch); x_axis_patch = NULL;
  ivResource::unref(y_axis_patch); y_axis_patch = NULL;
}

void GraphEditor::Init() {

  graphg = new Grapher_G(this, data, dtvsp);
  ivResource::ref(graphg);

  viewer = new GraphViewer(this,400, 150,dtvsp->bg_color.color);
  ivResource::ref(viewer);
  viewer->root(graphg);

  field_rep = new tbScrollBox();
  ((tbScrollBox*) field_rep)->naturalnum = 5;
  ivResource::ref(field_rep);
}


ivButton* GraphEditor::MEButton(ivTelltaleGroup* gp, char* txt, ivAction* a) {
  ivWidgetKit* wkit = ivWidgetKit::instance();

//  ivButton* rval = wkit->palette_button(txt, a);
//  rval->state()->join(gp);
//  return rval;

  ivTelltaleState* t = new Teller((TelltaleFlags)0, (ivAction*)NULL);
  ivButton* rval = new NoUnSelectButton
    (wkit->palette_button_look(wkit->label(txt), t), wkit->style(),t,a);
  rval->state()->set(ivTelltaleState::is_choosable | ivTelltaleState::is_toggle |
	ivTelltaleState::is_enabled, true);
  rval->state()->join(gp);
  return rval;
}

void GraphEditor::AddLastPoint() {
  if(graphg->transformer() == NULL) { // no transformer
//    viewer->initgraphic(); // recenter. hmmm this is a protected function
    viewer->root(graphg); // more cumbersome but it is the only way we have acess to initgraphic
  }
  graphg->AddLastPoint();
}

void GraphEditor::Reset() {
  fieldpatch->redraw();
  linespace->redraw();
  GetFields(field_rep);
  ivMonoGlyph* mg = (ivMonoGlyph *) fieldpatch->body(); // box
  mg->change(0);
  mg = (ivMonoGlyph *) mg->component(4); // box
  mg->change(0);
  fieldpatch->reallocate();
  if (x_axis_index > data->leaves){
      x_axis_index = 0;
      if(SetOwnerXAxis != NULL) SetOwnerXAxis(owner,x_axis_index);
  }
  GetAxis();
  graphg->Reset();
}

void GraphEditor::UpdateDisplay() {
  if(last_size != data->leaves) {
    Reset();
  }
  graphg->Build();
  viewer->Update_All();
  linespace->redraw();
  fieldpatch->redraw();
//  win->repair();		// friendlier?
}

void GraphEditor::InitDisplay() {
  Reset();
  UpdateDisplay();
}

void GraphEditor::SetXAxis(int idx) {
  // first have the old axis button get it's new axis
  GraphButton* button = (GraphButton *) buttons[x_axis_index];
  x_axis_index = idx;
  if(button != NULL) {
    button->SetAxis();
    button->state()->set(ivTelltaleState::is_chosen, button->spec->visible);
  }
  button = (GraphButton *) buttons[idx];
  if(button != NULL) button->SetAxis();
  GetAxis(); // this recreates all the axes
  UpdateDisplay(); // sets x axis in grapher_g and rebuilds all lines 
  if(SetOwnerXAxis != NULL) SetOwnerXAxis(owner,x_axis_index);
}

void GraphEditor::ToggleLine(int idx) {
  if(idx >= data->leaves)
    return;
  DA_ViewSpec* dvs = dtvsp->Leaf(idx);
  dvs->visible = !(dvs->visible);
  if(x_axis_index >= data->size) Reset();
  else {
    GetAxis();       // ^^ Reset also gets the buttons, but we don't want 
    graphg->Reset(); // to do that unless the x axis is out of range
                     // so we just call these two functions directly
  }
//  if(viewer->init_xform != NULL) viewer->scale_axes();
  UpdateDisplay();
}


#include <ta/enter_iv.h>
declareActionCallback(GraphEditor)
implementActionCallback(GraphEditor)
#include <ta/leave_iv.h>

ivGlyph* GraphEditor::GetTools() {
  ivLayoutKit* layout = ivLayoutKit::instance();
  ivWidgetKit* wkit = ivWidgetKit::instance();

  tool_gp = new ivTelltaleGroup;

  ivButton* msbutton = 
    MEButton(tool_gp,"Mv/Scan",
	     new GrapherCallback(viewer,&GlyphViewer::cur_tool,Tool::move));
  
  ((NoUnSelectButton *) msbutton)->chooseme();

  return layout->vbox
    (layout->hbox
     (layout->hnatural
       (MEButton(tool_gp,"Select",
		 new GrapherCallback(viewer,
				     &GlyphViewer::cur_tool,Tool::select)), 60),
      layout->hglue(),
      layout->hnatural(msbutton,60),
      layout->hglue(),
      layout->hnatural
       (MEButton(tool_gp,"ReScale",
		 new GrapherCallback(viewer,
				     &GlyphViewer::cur_tool,Tool::scale)), 60)),
     taivM->vfsep,
     wkit->menu_item_separator_look(),
     layout->hbox(layout->hglue(),wkit->fancy_label("Graph Labels"),layout->hglue()),
     taivM->vfsep,
     layout->hbox
     (layout->hflexible
      (layout->hnatural
       (wkit->push_button("New", new ActionCallback(GraphEditor)
			  (this, &GraphEditor::Create_Default_Label)),50)),
      layout->hglue(),
      layout->hflexible
      (layout->hnatural
       (wkit->push_button("Delete", new ActionCallback(GraphEditor)
			  (this, &GraphEditor::Delete_Labels)),50)),
      layout->hglue(),
      layout->hflexible
      (layout->hnatural
       (wkit->push_button("Edit", new ActionCallback(GraphEditor)
			  (this, &GraphEditor::Edit_Labels)),50))));
}


ViewLabel* GraphEditor::Create_Label(char* label_name){
  if(owner ==NULL) return NULL;
  ViewLabel_List* vlg = NULL;
  owner->FindMembeR(&TA_ViewLabel_List,((void *&) vlg));
  if (vlg ==  NULL) return NULL;
  ViewLabel *vl = (ViewLabel *) vlg->New(1);
  vl->get_color = GraphGetTextColor;
  if(label_name != NULL) {
    vl->name = label_name;
  }
  else{
    vl->name = String("GraphView Label") + String(vlg->size);
  }
  vl->MakeText();
  InitDisplay();
  taivMisc::RecordScript(vl->GetPath() +".New(1);\n");
  taivMisc::RecordScript(vl->GetPath() + String(".El[") + String(vlg->size-1)
			  + String("].name = \"") + vl->name +
			  String("\";\n"));
  return vl;
}

void GraphEditor::Delete_Labels(){
  taBase_List* labelgroup = &(graphg->selectgroup);
  taBase_List* viewlabelgroup;
  owner->FindMembeR(&TA_ViewLabel_List,((void *&) viewlabelgroup));
  if(labelgroup->size == 0){
    if(taMisc::Choice("Delete All Labels?","Ok","Cancel") == 0)
      labelgroup = viewlabelgroup;
    else labelgroup = NULL;
  }
  if(labelgroup == NULL) return;
  TAPtr o;
  int i;
  for(i=labelgroup->size-1;i>=0;i--){
    o = labelgroup->FastEl(i);
    if(viewlabelgroup != NULL) {
      String recordstring = viewlabelgroup->GetPath() + ".Remove("
	+ o->GetPath() + ");";
      viewlabelgroup->Remove(o);
      taivMisc::RecordScript(recordstring);
    }
    if(labelgroup != viewlabelgroup)     labelgroup->Remove(i); // select group
  }
}

void GraphEditor::Edit_Labels(){
  if(graphg->selectgroup.size > 1){
    taivEdit* gc = (taivEdit*)graphg->selectgroup.GetTypeDef()->ive;
    if(gc != NULL) 
      gc->Edit( (void*) &(graphg->selectgroup),win);
  }
  else if (graphg->selectgroup.size == 1){
    taivEdit* gc = (taivEdit*)graphg->selectgroup[0]->GetTypeDef()->ive;
    if(gc != NULL) 
      gc->Edit( (void*)(graphg->selectgroup[0]),win);
  }
}

void GraphEditor::InitColors(){
  dtvsp->UpdateLineFeatures();
}

ivGlyph* GraphEditor::GetFields(ivScrollBox* sb) {
  ivGlyphIndex c;
  for(c=sb->count()-1; c>= 0; c--) // clear out any existing
    sb->remove(c);
  buttons.Reset();
  ivResource::flush();
  if(data->leaves <= 0)
    return (ivGlyph*)sb;

  int cnt = 0;
  taLeafItr i, vi;
  DataArray_impl* da;
  DA_GraphViewSpec* dvs;
  for(da=(DataArray_impl*)data->FirstEl(i), dvs=(DA_GraphViewSpec*)dtvsp->FirstEl(vi);
      ((da != NULL) && (dvs != NULL));
      da=(DataArray_impl*)data->NextEl(i), dvs=(DA_GraphViewSpec*)dtvsp->NextEl(vi), cnt++)
  {
    GraphButton* fb = GraphButton::GetButton(this, dvs,cnt);
    buttons.Add(fb);
    if(da->InheritsFrom(TA_String_Data) && !da->HasDispOption(" DISP_STRING, ")) {
	continue;
    }
    sb->append(fb);
    if(!dvs->visible) {
      ((GraphButton*)buttons[cnt])->state()->set(ivTelltaleState::is_chosen, false);
    }
  }
  return (ivGlyph *) sb;
}

ivCanvas* GraphEditor::GetCanvas(){
  if(viewer != NULL) {
    return(viewer->canvas());
  }
  else return (NULL);
}


ivGlyph* GraphEditor::GetAxis() {
  ivLayoutKit* layout = ivLayoutKit::instance();
  
  ivGlyph* result = layout->hbox();
  last_size = data->leaves;
  if(buttons.size < data->leaves)
    return result;

  y_axis_patch->body(result);
  ivResource::flush();		// get rid of old axis's

  axis.Reset();
  axis_patch.Reset();
  x_axis_patch->body(new ivPatch(NULL));
  ivResource::flush();

  int cnt;
  taLeafItr di, vi;
  DataArray_impl* da;
  DA_GraphViewSpec* dvs;
  for(cnt = 0, da=(DataArray_impl*)data->FirstEl(di), dvs=(DA_GraphViewSpec*)dtvsp->FirstEl(vi);
      ((da != NULL) && (dvs != NULL));
      da=(DataArray_impl*)data->NextEl(di), dvs=(DA_GraphViewSpec*)dtvsp->NextEl(vi), cnt++)
  {
    if(dvs->axis_spec == NULL)
      taBase::SetPointer((TAPtr*)&(dvs->axis_spec),dvs);
    float amin = 0.0f; float amax = 0.0f;
    // set initial min and max if available
    if(!da->InheritsFrom(TA_String_Data) && (((float_Data *)da)->ar != NULL)) {
      float_Data* fd = (float_Data *)da;
      amin = fd->ar->range.min; amax = fd->ar->range.max;
    }
    if(cnt == x_axis_index) { // the Xaxis - use contrast color
      XAxis* xax = new XAxis(0.0f,0.0f,dtvsp->bg_color.contrastcolor);
      xax->n_ticks = dvs->n_ticks;
      if(dvs->min_mode == DA_GraphViewSpec::FIXED)
	xax->fixed_min = true;
      if(dvs->max_mode == DA_GraphViewSpec::FIXED)
	xax->fixed_max = true;
      axis.Add(xax);
      ivPatch* ap = new ivPatch(xax);
      axis_patch.Add(ap);
      x_axis_patch->body(ap);
      ((GraphButton*)buttons[cnt])->axis = xax;
      ((GraphButton*)buttons[cnt])->axis_patch = ap;
    } // self-axes
    else {
      if(!dvs->visible ||
	 (da->InheritsFrom(TA_String_Data) && !da->HasDispOption(" DISP_STRING, ")))
	continue;
      if((dvs->axis_spec != dvs) && (dvs->axis_spec->visible))
	continue;		// save users of other axes for later
      YAxis* yax = new YAxis(amin,amax,dvs->ta_color.color);
      yax->n_ticks = dvs->n_ticks;
      if(dvs->min_mode == DA_GraphViewSpec::FIXED)
	yax->fixed_min = true;
      if(dvs->max_mode == DA_GraphViewSpec::FIXED)
	yax->fixed_max = true;
      axis.Add(yax);
      // the 35  vvv is to create space for the text numbers
      ivPatch* ap = new ivPatch(layout->margin(yax,35,0,0,0));
      axis_patch.Add(ap);
      result->append(ap);
      ((GraphButton*)buttons[cnt])->axis = yax;
      ((GraphButton*)buttons[cnt])->axis_patch = ap;
    }
  }
  // now a second pass for all the users of other axes
  for(cnt = 0, da=(DataArray_impl*)data->FirstEl(di), dvs=(DA_GraphViewSpec*)dtvsp->FirstEl(vi);
      ((da != NULL) && (dvs != NULL));
      da=(DataArray_impl*)data->NextEl(di), dvs=(DA_GraphViewSpec*)dtvsp->NextEl(vi), cnt++)
  {
    if((cnt == x_axis_index) || !dvs->visible ||
       (da->InheritsFrom(TA_String_Data) && !da->HasDispOption(" DISP_STRING, ")))
      continue;
    if((dvs->axis_spec == dvs) || !dvs->axis_spec->visible)
      continue;		// got its own axis

    int ai = dtvsp->FindLeaf(dvs->axis_spec);
    YAxis* yax = (YAxis *) ((GraphButton*)buttons[ai])->axis;
    ivPatch* ap = ((GraphButton*)buttons[ai])->axis_patch;
    ((GraphButton*)buttons[cnt])->axis = yax;
    ((GraphButton*)buttons[cnt])->axis_patch = ap;
  }

  if(resize_vbox) resize_vbox->modified(0); // notify box's
  if(resize_vbox2) resize_vbox2->modified(0);
  if(resize_hbox) resize_hbox->modified(0);

  if(boxpatch){
    ((CachePatch *) boxpatch)->impl->fakeflag = CachePatch_Impl::ONCE;
    boxpatch->reallocate();
  }
  return result;
}


ivGlyph* GraphEditor::GetLook() {
  ivLayoutKit* layout = ivLayoutKit::instance();
  ivWidgetKit* wkit = ivWidgetKit::instance();
  
  GetFields(field_rep);
  ivPolyGlyph* fieldbox = layout->vbox
    (GetTools(),
     taivM->vfsep,
     wkit->menu_item_separator_look(),
     taivM->vfsep,
     layout->hbox
     (wkit->vscroll_bar(field_rep),
      taivM->hfsep,
      layout->hfixed(field_rep, taivM->big_button_width + ((16 + 2 + 2)* 2 ))));
  
  linespace = new ivPatch(viewer);
  ivResource::ref(linespace);
  fieldpatch = new ivPatch(fieldbox);
  ivResource::ref(fieldpatch);
  
  x_axis_patch = new ivPatch(NULL);
  ivResource::ref(x_axis_patch);
  
  y_axis_patch = new ivPatch(NULL);
  ivResource::ref(y_axis_patch);
  
  GetAxis();	// the y axis's 
  
  // margin = lrbt
  resize_vbox =
    layout->vbox
    (layout->margin(y_axis_patch,  5.0,  0.0,
		    GRAPH_BOT_BORDER, GRAPH_TOP_BORDER),
     layout->shape_of_xy
     (layout->margin(y_axis_patch, 5.0,  0.0,
		    GRAPH_BOT_BORDER, GRAPH_TOP_BORDER),
      layout->margin(x_axis_patch, GRAPH_LFT_BORDER, GRAPH_RGT_BORDER,
		     20.0, 0.0)));
  ivResource::ref(resize_vbox);

//   resize_vbox2 = 
//     layout->vbox
//     (layout->hcenter
//      (layout->margin
//       (layout->flexible
//        (new ivBackground(linespace,dtvsp->bg_color.color)),5,15,5,5),0),
//      layout->hcenter
//      (layout->margin(x_axis_patch,5,15,20,5),0));

  resize_vbox2 = 
    layout->vbox
    (layout->hcenter(layout->margin(layout->flexible(linespace),
				    GRAPH_LFT_BORDER,GRAPH_RGT_BORDER,
				    GRAPH_BOT_BORDER,GRAPH_TOP_BORDER),0),
     layout->hcenter(layout->margin(x_axis_patch,
				    GRAPH_LFT_BORDER,GRAPH_RGT_BORDER,
				    20.0,0.0),0));
  ivResource::ref(resize_vbox2);

  resize_hbox = 
    layout->hbox(layout->vcenter(resize_vbox,0),
		 layout->vcenter(resize_vbox2,0));
  ivResource::ref(resize_hbox);

//  boxpatch = new CachePatch(wkit->inset_frame(resize_hbox));
  boxpatch = new CachePatch(new DynamicBackground
			      (resize_hbox,dtvsp->bg_color.color));
//  boxpatch = new ivPatch(new DynamicBackground
//			 (resize_hbox,dtvsp->bg_color.color));
  ivResource::ref(boxpatch);

  return 
    (wkit->inset_frame
     (layout->hbox
      (layout->vcenter(wkit->outset_frame
		       (new ivBackground(fieldpatch, wkit->background())),0),
       layout->vcenter(taivM->hfsep,0),
       layout->vcenter(boxpatch,0))));
}


void GraphLine::Rescale() {
  if(!HaveData())	return;
  
  // Rescale does potentially 4 things (maybe more later)
  // 1) it can scale the line vertically to be the height of the window
  // 2) it can translate the line to be centererd vertically
  // 3) it can scale the line horizontally to be the width of the window
  // 4) it can translate the line horizontally to scroll the line or
  //       start at the left of the window.
  
  // Currently the line is given an initial horizontal translation (118)
  // so that it starts drawing at the left side of the window. We need to
  // preserve this translation, but maybe also add to it to scroll.
  
  // Vertically we need to do the translation before the scaling.
  // This is because the translation is in the line coordinates
  //   ex: if the max and min of the line are (1.2,1.6) then we need
  //       to translate the line by -1.4 so that it is centered around
  //       zero before scaling it to fit the window.
  
  // Horizontaly the translation comes after the scaling, since it is
  // in screen coordinates that we want to scroll. ex: move 1 pixel to left.
  
  // the parent object's matrix may also translate this so 0,0 is the center
  // of the screen. This is good for vertical , but not for horizontal, so
  // that is why there is an initial horizontal translation(118) to the line.

  // x_min...y_max are the corners of the rectangle enclosing
  // the line. The corners() function multiplies those values
  // by the current matrix to get the projection of those corners.

  ivCoord l, b, r, t;
  l= x_axis->range.min;		// be sure to use the axis version of the scale..
  r= x_axis->range.max;
  b= axis->range.min;
  t= axis->range.max;

  if(_t) corners(l,b,r,t, *_t);

  // If the line is one dimensional then do not scale, or
  // else we will get a division by zero error.

//   if(t==b || r==l) {
//     return;
//   }

  
  // Get the owner's (the graph_g's) size, so we know how
  // to adjust ourselves.

  // this should be the viewer's size and not the grph_g's size
  // incase the viewer is zoomed.

  ivDisplay* disp = ivSession::instance()->default_display();
  ivCoord one_pixel = disp->to_coord(1);

  ivAllocation* a = owner->GetAllocation();
  ivAllotment ax = a->allotment(Dimension_X);
  ivAllotment ay = a->allotment(Dimension_Y);
  ivCoord width = (ax.end() - ax.begin()) - one_pixel;
  ivCoord height = (ay.end() - ay.begin()) - one_pixel;

  // give a border

//  height -= 10;
//  width  -= 10;

  // hscale and vscale are the values we need to scale the curent
  // matrix to get the correct scaling.
  // **** The else case is probably flawed, and may not be necessary? ****

  ivCoord hscale, vscale;
  if (t>b) {
    vscale = (height/(t-b));
  }
  else { vscale = 1;}

  if (r>l) {
    hscale = (width/(r-l));
  }
  else { hscale = 1;}

  // rvscale is the real vertical scale we would multiply by the
  // Identity matrix to scale correctly.
  ivCoord rvscale = height/(axis->range.Range());
  ivCoord rhscale = width/(x_axis->range.Range());
    
  // these are some coords we will use later

  float b00,b01,b10,b11,b20,b21;

  // _t is our transform matrix;
  // get how much we are currently translated verticaly,
  // and translate us back to zero.

  if(_t == NULL){
    transformer(new ivTransformer);
    // skip the translation since the new transformer is not translated
  }
  else { 
    _t->matrix(b00,b01,b10,b11,b20,b21);
    _t->translate(-b20,-b21);
  }
    
  // now adjust the scale
  _t->scale(hscale,vscale);
    
  // and translate the correct amount
  _t->translate(-(x_axis->range.MidPoint()* rhscale),
		-(axis->range.MidPoint() * rvscale));
}
