/* -*- C++ -*- */
/*=============================================================================
//									      //
// This file is part of the TypeAccess/C-Super-Script software package.	      //
//									      //
// Copyright (C) 1995 Randall C. O'Reilly, Chadley K. Dawson, 		      //
//		      James L. McClelland, and Carnegie Mellon University     //
//     									      //
// Permission to use, copy, modify, and distribute this software and its      //
// documentation for any purpose is hereby granted without fee, provided that //
// the above copyright notice and this permission notice appear in all copies //
// of the software and related documentation.                                 //
// 									      //
// Note that the PDP++ software package, which contains this package, has a   //
// more restrictive copyright, which applies only to the PDP++-specific       //
// portions of the software, which are labeled as such.			      //
//									      //
// Note that the taString class, which is derived from the GNU String class,  //
// is Copyright (C) 1988 Free Software Foundation, written by Doug Lea, and   //
// is covered by the GNU General Public License, see ta_string.h.             //
// The iv_graphic library and some iv_misc classes were derived from the      //
// InterViews morpher example and other InterViews code, which is             //
// Copyright (C) 1987, 1988, 1989, 1990, 1991 Stanford University             //
// Copyright (C) 1991 Silicon Graphics, Inc.				      //
//									      //
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,         //
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 	      //
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  	      //
// 									      //
// IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE FOR ANY SPECIAL,    //
// INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES  //
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT     //
// ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY,      //
// ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS        //
// SOFTWARE. 								      //
==============================================================================*/

#ifndef datagraph_h
#define datagraph_h

#include <ta/taiv_type.h>
#include <ta_misc/datatable.h>
#include <ta_misc/colorscale.h>
#include <ta_misc/colorbar.h>
#include <ta_misc/fontspec.h>

#include <ta/enter_iv.h>
#include <InterViews/monoglyph.h>
#include <InterViews/layout.h>
#include <InterViews/patch.h>
#include <IV-look/kit.h>
#include <OS/list.h>
#include <ta/leave_iv.h>

#include <iv_graphic/graphic.h>
#include <iv_graphic/graphic_viewer.h>

class ivBrush;
class ivColor;
class ivStyle;
class ivTransformer;
class ivWindow;
class ivTelltaleGroup;
class ivButton;
class ivAction;
class ivDialogKit;

class Axis;
class Grapher_G;		// the graphicmaster
class GraphEditor;		// the overall editor

class GraphButton : public ivButton {
public:
  GraphEditor* 	owner;	
  DA_GraphViewSpec* spec;
  int		field;		// field number
  Axis*		axis;		// axis for this field
  ivPatch*	axis_patch;	// axis_patch for this field

  void		release(const ivEvent& e); // process mouse release..

  ivGlyph*	GetLegend(DA_GraphViewSpec* vspec=NULL);
  virtual void	ToggleLine();	// toggle line on/off
  virtual void	SetXAxis();	// set this one to be the x axis (middle button)

  virtual void	SetAxis(); // set the righthand axis rep appropriately

  bool		IsChosen();	// is this button selected

  static GraphButton* GetButton(GraphEditor* o, DA_GraphViewSpec* gvs,int f);
  
  GraphButton(ivGlyph* g, ivStyle* s, ivTelltaleState* t,
	      GraphEditor* o, DA_GraphViewSpec* gvs, int f);
  ~GraphButton();
};


class GraphLine : public Graphic {
public:
  Grapher_G*		owner;
  DA_GraphViewSpec* 	spec;

  Axis*		axis;		// our axis
  Axis*		x_axis;		// the x_axis
  ivPatch*	axis_patch;

  DataTable*	data;		// overall data table for whole graph
  float_Data*   array;		// my data to plot from
  String_Data*	labels;		// if there are labels (then array = y axis points)
  int		dt_index;	// data table field index
  int		last_size;	// last size of the array (for addlastpoint)
  float_Data*	x_axis_ar;	// x axis to plot by

  BoxObj	intersect_box;  // region in which to draw
  float_Array	x_targets;
  float_Array	y_targets;
  String_Array	lab_cache;	// cached labels

  int		last_button;	// last button pressed
  int		last_index;	// last index

  void		Reset();	// clear existing points
  void		Build();	// build line from datatable
  int		SetAxisMinMax(); // set the min and max values
  void		AddAllPoints(); // add all points from graphtable

  void 		AddPoint(int i); // add point from table index i
  void		AddLastPoint();	// add new datapt from end of table

  bool		UpdateAxis();	  // send min/max to axis and redraw it (true if updt)
  void		RedrawAxis();	  // forcibly redraw the axis..
  void		Rescale();	// rescale based on owner's size

  bool		HaveData();	// checks if my array is ok

  bool 		effect_border();
  void		draw_gs(ivCanvas* c, Graphic* gs);
  void 		drawclipped_gs(ivCanvas* c, ivCoord l, ivCoord b,
			       ivCoord r, ivCoord t, Graphic* gs);

  bool		grasp_move(const ivEvent&, Tool&,ivCoord ix, ivCoord iy);
  bool		manip_move(const ivEvent&, Tool&,ivCoord ix, ivCoord iy,
			   ivCoord lx, ivCoord ly,
			   ivCoord cx, ivCoord cy);
  bool		effect_move(const ivEvent&, Tool&,ivCoord ix,
			    ivCoord iy, ivCoord fx, ivCoord fy);

  bool		grasp_scale(const ivEvent&, Tool&,ivCoord,ivCoord,ivCoord,ivCoord,
			    ivCoord,ivCoord) {return true;}
  bool		manip_scale(const ivEvent&,Tool&,ivCoord,ivCoord,ivCoord,ivCoord,
			    ivCoord,ivCoord,ivCoord,ivCoord) {return true;}

  GraphLine(Grapher_G* gg,   DA_GraphViewSpec* gvs,
	    DataTable* t, int index, float_Data* xax_ar, 
	    Axis* ax, ivPatch* axpat, Axis* xax, ivColor* c);
  ~GraphLine();
};


// probably should define a Graphic_List that does the ref, unref stuff
// this would make things easier here..

class Grapher_G : public GraphicMaster { // an overloaded polygraphic
public:
  taPtrList<GraphLine>	lines;	// no delete on resetting the list..
  GraphEditor*		owner;
  DataTable*		data;
  DT_GraphViewSpec*	dtvsp;
  Axis*			x_axis;
  float_Data*		x_axis_ar;
  GraphButton*		x_axis_but; // button for the x axis
  ivPatch*		x_axis_patch;

  virtual void		ClearAllocation();
  virtual void		Reset(); // clear all the lines in the graph
  virtual void		Build(); // create/draw all the lines in graph
  
  virtual void		AddLastPoint();	// add last point in datatable 

  virtual void		SetXAxis(int idx); // set x axis for all lines
  virtual int		FixXAxis(); // updates x_axis;

  virtual void		allocate(ivCanvas*, const ivAllocation&, ivExtension&);
  ivAllocation*		GetAllocation(){return &_a;}
  ivColor*		GetTextColor(); // return the background contrast color

  Grapher_G(GraphEditor* onr, DataTable* t, DT_GraphViewSpec* dtvs);
  ~Grapher_G();
};


class GraphViewer : public GlyphViewer {
protected:
  void rate_zoom();
  void grab_scroll();
public:
  GraphEditor* editor;
  ivTransformer* init_xform;

  void scale_axes(int redraw=true);
  GraphicMaster* root() { return GlyphViewer::root(); }
  void root(GraphicMaster*);
  void allocate(ivCanvas*, const ivAllocation&, ivExtension&);
  void ClearCachedAllocation();

  GraphViewer(GraphEditor* e,float w, float h, const ivColor* bg);
  ~GraphViewer();
};

class GraphEditor {
  // ##NO_INSTANCE ##NO_MEMBERS ##NO_TOKENS ##NO_CSS
  // graph editor: contains all the junk for making a graph editor, not part of glyph hierarchy
public:
  TAPtr			owner;	// the GraphLogView in general
  int			x_axis_index;	// index of index in array

  ivGlyph_List		buttons; 	// buttons representing fields
  ivGlyph_List		axis_patch; 	// patches for each axis
  ivGlyph_List		axis;	      	// the axis's 

  DataTable*		data;		// the data to be graphed
  DT_GraphViewSpec*	dtvsp;		// the display spec for the data
  Grapher_G*		graphg;		// graphic master graphg
  int			last_size; 	// data->size last time updated..

  ivWindow*		win;
  GraphViewer* 		viewer;	
  ivPatch*		linespace;
  
  ivTelltaleGroup*	tool_gp; 

  ivPatch*		fieldpatch; 	// field button representation
  ivScrollBox*	        field_rep; 

  ivPatch*		y_axis_patch;
  ivPatch*		x_axis_patch;

  ivPatch*		boxpatch; 	// this covers the the axi & lines
  ivPolyGlyph*		resize_vbox; 
  ivPolyGlyph*		resize_vbox2;
  ivPolyGlyph*		resize_hbox; 

  void			(*SetOwnerXAxis)(TAPtr, int);

  virtual void		Init();		// initialization
  virtual ivGlyph* 	GetLook();	// gets the whole thing (only call this once)
  virtual void		InitColors();

  virtual ivGlyph* 	GetTools();	// get the tool buttons
  virtual ivGlyph* 	GetFields(ivScrollBox* sb);// get the variable buttons
  virtual ivGlyph*	GetAxis(); 	// Delete and recreate the axis

  virtual ivCanvas*	GetCanvas(); // return the viewer's canvas

  // get a "pallete button" that is mutually exclusive
  virtual ivButton*	MEButton(ivTelltaleGroup* gp, char* txt, ivAction* a);

  // this is the main user interface
  virtual void		Reset();	// clear everything
  virtual ViewLabel*	Create_Label(char* label_name = NULL);	// create a label
  virtual void		Create_Default_Label(){Create_Label();}
  virtual void		Delete_Labels();	// delete selected labels
  virtual void		Edit_Labels();	// edit selected labels

  virtual void		AddLastPoint(); // add last point in the table (update)
  virtual void		InitDisplay();	  // reinit stuff
  virtual void		UpdateDisplay(); // redo the whole display

  virtual void		SetXAxis(int idx); 	// changes x axis
  virtual void		ToggleLine(int idx); 	// toggles display of given line

  GraphEditor(TAPtr own, DataTable* d, DT_GraphViewSpec* dtvs, ivWindow* w);
  virtual ~GraphEditor();
};



#endif // datagraph_h
