/* -*- C++ -*- */
/*=============================================================================
//									      //
// This file is part of the TypeAccess/C-Super-Script software package.	      //
//									      //
// Copyright (C) 1995 Randall C. O'Reilly, Chadley K. Dawson, 		      //
//		      James L. McClelland, and Carnegie Mellon University     //
//     									      //
// Permission to use, copy, modify, and distribute this software and its      //
// documentation for any purpose is hereby granted without fee, provided that //
// the above copyright notice and this permission notice appear in all copies //
// of the software and related documentation.                                 //
// 									      //
// Note that the PDP++ software package, which contains this package, has a   //
// more restrictive copyright, which applies only to the PDP++-specific       //
// portions of the software, which are labeled as such.			      //
//									      //
// Note that the taString class, which is derived from the GNU String class,  //
// is Copyright (C) 1988 Free Software Foundation, written by Doug Lea, and   //
// is covered by the GNU General Public License, see ta_string.h.             //
// The iv_graphic library and some iv_misc classes were derived from the      //
// InterViews morpher example and other InterViews code, which is             //
// Copyright (C) 1987, 1988, 1989, 1990, 1991 Stanford University             //
// Copyright (C) 1991 Silicon Graphics, Inc.				      //
//									      //
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,         //
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 	      //
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  	      //
// 									      //
// IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE FOR ANY SPECIAL,    //
// INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES  //
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT     //
// ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY,      //
// ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS        //
// SOFTWARE. 								      //
==============================================================================*/

#ifndef datatable_h
#define datatable_h

#include <ta/ta_group.h>
#include <ta_misc/ta_misc_TA_type.h>
#include <ta_misc/tdgeometry.h>
#include <ta_misc/minmax.h>
#include <ta_misc/aggregate.h>
#include <ta_misc/colorscale.h> // for DT && DA Viewspecs


class float_RArray : public float_Array {
  // #NO_UPDATE_AFTER float array with range
public:
  enum DistMetric {		// generalized distance metrics
    SUM_SQUARES,		// sum of squares:  sum[(x-y)^2]
    EUCLIDIAN,			// Euclidian distance (sqrt of sum of squares)
    HAMMING, 			// Hamming distance: sum[abs(x-y)]
    COVAR,			// covariance: sum[(x-<x>)(y-<y>)]
    CORREL,			// correlation: sum[(x-<x>)(y-<y>)] / sqrt(sum[x^2 y^2])
    INNER_PROD,			// inner product: sum[x y]
    CROSS_ENTROPY		// cross entropy: sum[x ln(x/y) + (1-x)ln((1-x)/(1-y))]
  };

  MinMax	range;		// min-max range of the data 

  virtual void	UpdateRange(float item)	{ range.UpdateRange(item); }
  virtual void	UpdateAllRange();	
  // #MENU #MENU_ON_Actions update range for all items in array

  void		Set(int i, const float& item)
  { float_Array::Set(i, item); UpdateRange(item); }
  void		Add(const float& item);
  void		Insert(const float& item, int idx, int n_els=1);

  float		Pop();
  bool		Remove(const float& item);
  bool		Remove(uint indx, int n_els=1);
  void 		CopyVals(const taArray_impl& from, int start=0, int end=-1, int at=0);
  
  virtual float	MaxVal(int& idx, int start=0, int end = -1) const;
  // #MENU #MENU_ON_Actions #USE_RVAL value and index of the (first) element that has the maximum value
  virtual float	AbsMaxVal(int& idx, int start=0, int end = -1) const;
  // #MENU #MENU_ON_Actions #USE_RVAL value and index of the (first) element that has the maximum absolute value
  virtual float	MinVal(int& idx, int start=0, int end = -1) const;
  // #MENU #USE_RVAL value and index of the (first) element that has the minimum value
  virtual float	Sum(int start=0, int end = -1) const;
  // #MENU #MENU_ON_Actions #USE_RVAL compute the sum of the values in the array
  virtual float	Mean(int start=0, int end = -1) const;
  // #MENU #USE_RVAL compute the mean of the values in the array
  virtual float	Var(float mean=0, bool use_mean=false, int start=0, int end = -1) const;
  // #MENU #USE_RVAL compute the variance of the values, opt with given mean
  virtual float	StdDev(float mean=0, bool use_mean=false, int start=0, int end = -1) const;
  // #MENU #USE_RVAL compute the standard deviation of the values, opt with given mean
  virtual float	SEM(float mean=0, bool use_mean=false, int start=0, int end = -1) const;
  // #MENU #USE_RVAL compute the standard error of the mean of the values, opt with given mean
  virtual float	SSLength(int start=0, int end = -1) const;
  // #MENU #USE_RVAL sum-of-squares length of the array

  // generalized distance measures follow
  virtual float	SumSquaresDist(const float_RArray& oth, bool norm = false, 
			       float tolerance=0.0f, int start=0, int end = -1) const;
  // compute sum-squares dist between this and the oth, tolerance is by element
  virtual float	EuclidDist(const float_RArray& oth, bool norm = false, 
			   float tolerance=0.0f, int start=0, int end = -1) const;
  // compute Euclidian dist between this and the oth, tolerance is by element
  virtual float	HammingDist(const float_RArray& oth, bool norm = false, 
			    float tolerance=0.0f, int start=0, int end = -1) const;
  // compute Hamming dist between this and the oth, tolerance is by element
  virtual float	Covar(const float_RArray& oth, int start=0, int end = -1) const;
  // compute the covariance of this array the oth array
  virtual float	Correl(const float_RArray& oth, int start=0, int end = -1) const;
  // compute the correlation of this array with the oth array
  virtual float	InnerProd(const float_RArray& oth, bool norm = false, 
			  int start=0, int end = -1) const;
  // compute the inner product of this array and the oth array
  virtual float	CrossEntropy(const float_RArray& oth, int start=0, int end = -1) const;
  // compute cross entropy between this and other array, this is 'p' other is 'q'

  virtual float	Dist(const float_RArray& oth, DistMetric metric, bool norm = false, 
			float tolerance=0.0f, int start=0, int end = -1) const;
  // compute generalized distance metric with other array (calls appropriate fun above)
  static bool	LargerFurther(DistMetric metric);
  // returns true if a larger value of given distance metric means further apart

  virtual void	Histogram(const float_RArray& oth, float bin_size);
  // this gets a histogram (counts) of number of values in other array

  virtual void	AggToArray(const float_RArray& from, Aggregate& agg,
			   int start=0, int end = -1);
  // aggregate values from other array to this one using aggregation params of agg
  virtual float	AggToVal(Aggregate& agg, int start=0, int end = -1) const;
  // compute aggregate of values in this array using aggregation params of agg

  virtual float	NormLen(float len=1.0f, int start=0, int end = -1);
  // #MENU normalize array to total given length (1.0), returns scale
  virtual float	NormSum(float sum=1.0f, float min_val=0.0f, int start=0, int end = -1);
  // #MENU normalize array to total given sum (1.0) and min_val (0) (uses range), returns scale
  virtual float	NormMax(float max=1.0f, int start=0, int end = -1);
  // #MENU normalize array to given maximum value, returns scale
  virtual float	NormAbsMax(float max=1.0f, int start=0, int end = -1);
  // #MENU normalize array to given absolute maximum value, returns scale
  virtual void	SimpleMath(const SimpleMathSpec& math_spec, int start=0, int end = -1);
  // #MENU apply standard kinds of simple math operators to values in the array
  virtual void	SimpleMathArg(const float_RArray& arg_ary, const SimpleMathSpec& math_spec, int start=0, int end = -1);
  // apply simple math operators to values, other array provides 'arg' value for math_spec
  virtual int	Threshold(float thresh=.5f, float low=0.0f, float high=1.0f, int start=0, int end = -1);
  // #MENU threshold values in the array, low vals go to low, etc.

  virtual void 	WritePoint(const TwoDCoord& geom, int x, int y, float color=1.0, bool wrap=true);
  // write a single point, assuming geometry geom
  virtual void 	RenderLine(const TwoDCoord& geom, int xs, int ys, int xe, int ye,
			   float color=1.0, bool wrap=true);
  // #MENU render a line from given x,y starting, ending coords in 2d space of geometry geom

  virtual void 	WriteXPoints(const TwoDCoord& geom, int x, int y, const float_RArray& color,
			     int wdth=1, bool wrap=true);
  // write a series of points of given width in x dimension using colors in order
  virtual void 	WriteYPoints(const TwoDCoord& geom, int x, int y, const float_RArray& color,
			     int wdth=1, bool wrap=true);
  // write a series of points of given width in y dimension using colors in order
  virtual void 	RenderWideLine(const TwoDCoord& geom, int xs, int ys, int xe, int ye,
			       const float_RArray& color, int wdth=1, bool wrap=true);
  // #MENU render a wide line from given x,y starting, ending coords in 2d space of geometry geom


  // updtate range after
  void		ShiftLeft(int nshft);
  int		Dump_Load_Value(istream& strm, TAPtr par=NULL);

  void		Reset(){float_Array::Reset();range.Init(0.0f);}

  void	Initialize()		{ };
  void	Destroy()		{ };
  void	InitLinks();
  void 	Copy_(const float_RArray& cp);
  COPY_FUNS(float_RArray, float_Array);
  TA_BASEFUNS(float_RArray);
};


// specific ones are in the _tmplt file (String_Data, float_Data)

class DataArray_impl : public taNBase {
  // ##NO_TOKENS #NO_UPDATE_AFTER holds array data
protected:
  static String	opt_after;	// #IGNORE temp string for option after function
public:
  String	disp_opts;	// viewer default display options

  // these are overloaded for the specific types
  virtual taArray_base* 	AR()	{ return NULL; } // the array pointer
  virtual void	NewAR() { };
  // #MENU #MENU_ON_Object create an array for yourself
  virtual void	SetAR(taArray_base*) { };	// set AR to existing array

  bool		HasDispOption(const char* opt) const
  { return disp_opts.contains(opt); } // check if a given display option is set
  String& 	DispOptionAfter(const char* opt)
  { opt_after = disp_opts.after(opt); opt_after = opt_after.before(',');
    return opt_after; }

  void	Initialize();
  void  InitLinks();
  void	Destroy()	{ };
  void 	Copy_(const DataArray_impl& cp);
  COPY_FUNS(DataArray_impl, taNBase);
  TA_BASEFUNS(DataArray_impl);
};

class DataTable : public taGroup<DataArray_impl> {
  // #NO_UPDATE_AFTER table of data
public:
  virtual void	ResetData();
  // #MENU #MENU_ON_Actions reset all of the data arrays
  virtual void	RemoveRow(int row_num);
  // #MENU Remove an entire row of data
  virtual void	AddBlankRow();
  // #MENU add a new row to the data table, returns new row number

  virtual void	AddRowToArray(float_RArray& ar, int row_num) const;
  // add a row of the datatable to given array
  virtual void	AggRowToArray(float_RArray& ar, int row_num, Aggregate& agg) const;
  // aggregate a row of the datatable to given array using parameters in agg
  virtual float	AggRowToVal(int row_num, Aggregate& agg) const;
  // aggregate a row of the datatable to a value using parameters in agg
  virtual void	AddArrayToRow(float_RArray& ar);
  // add contents of array to datatable
  virtual void	AggArrayToRow(float_RArray& ar, int row_num, Aggregate& agg);
  // aggregate contents of array to datatable at given row
  virtual void	PutArrayToRow(float_RArray& ar, int row_num);
  // just put array values into given row of data

  int  		MinLength();		// #IGNORE
  int  		MaxLength();		// #IGNORE

  void	UpdateAfterEdit();
  void	Initialize();
  void	Destroy();
  void	InitLinks();
  TA_BASEFUNS(DataTable);
};

#include <ta_misc/datatable_tmplt.h>

class ClustNode;

class ClustLink : public taBase {
  // #INLINE ##NO_TOKENS ##NO_UPDATE_AFTER a link in the cluster tree with distance
public:
  float		dist;		// distance to this node from parent
  ClustNode*	node;		// cluster node

  void	Initialize();
  void	Destroy()	{ CutLinks(); }
  void	CutLinks();
  void	Copy_(const ClustLink& cp);
  COPY_FUNS(ClustLink, taBase);
  TA_BASEFUNS(ClustLink);
};
  

class ClustNode : public taNBase {
  /* ##INLINE ##NO_TOKENS #NO_UPDATE_AFTER node in clustering algorithm
     use one with leaves as children as a root node for cluster */
public:
  float_RArray*	pat;		// pattern I point to (if leaf)
  float		par_dist;	// distance from parent cluster
  float		nn_dist;	// nearest neighbor (within cluster) distance
  float		tmp_dist;	// temporary distance value (for computations)
  float		y;		// y axis value
  taBase_List	children;	// my sub-nodes
  taBase_List	nns;		// nearest neighbor(s)

  ClustNode*	GetChild(int i)	const { return ((ClustLink*)children[i])->node; }
  ClustNode*	GetNN(int i)	const { return ((ClustLink*)nns[i])->node; }
  float& 	GetNNDist(int i) const { return ((ClustLink*)nns[i])->dist; }

  virtual void	AddChild(ClustNode* nd, float dst = 0.0f);
  // add new child (via ClustLink)
  virtual void	LinkNN(ClustNode* nd, float dst = 0.0f);
  // add new neighbor (via ClustLink)
  virtual bool	RemoveChild(ClustNode* nd);// remove link with this node
  virtual int	FindChild(ClustNode* nd); // find child with this node link

  virtual void 	Cluster(float_RArray::DistMetric metric=float_RArray::EUCLIDIAN,
		    bool norm=false, float tol=0.0f);
  // generate the cluster: call on a root node with a flat list of leaf children

  virtual void	Graph(ostream& strm);
  // generate commands to drive graph (or xgraph) for plotting cluster
  virtual void	Graph_impl(ostream& strm);
  // #IGNORE implementation

  virtual void	XGraph(const char* fnm, const char* title);
  // generate graph in given file name with given title, and call xgraph on result

  virtual void	GraphData(DataTable* dt);
  // generate graph as X, Y, label points in a datatable, suitable for graphing
  virtual void	GraphData_impl(DataTable* dt);
  // #IGNORE implementation

  virtual void	NNeighbors(float_RArray::DistMetric metric,
		    bool norm, float tol);
  // #IGNORE find nearest neighbors for everyone

  virtual bool	ClustOnClosest(float_RArray::DistMetric metric);
  // #IGNORE generate a new cluster by combining closest nearest neighbors together

  float	Dist(const ClustNode& oth, float_RArray::DistMetric metric,
	     bool norm, float tol) const;
  // #IGNORE compute distance between this node and other, averaging over individual patterns

  virtual void	SetYs(float y_init = -1.0);
  // #IGNORE traverse the tree computing y values (starting from initial given value)

  virtual float	SetParDists(float par_d, float_RArray::DistMetric metric);
  // #IGNORE traverse the tree computing par_dist values (starting from given dist)

  void	Initialize();
  void	Destroy()	{ };
  SIMPLE_COPY(ClustNode);
  COPY_FUNS(ClustNode, taNBase);
  TA_BASEFUNS(ClustNode);
};

class DA_ViewSpec : public taNBase { 
  // ##SCOPE_DT_ViewSpec base specification for the display of log data_array (DA)
public:
  DataArray_impl*	data_array;	// #READ_ONLY #NO_SAVE the data array
  String		display_name;	// name used in display
  bool	        	visible;	// visibility flag

  virtual void		UpdateView();
  // #BUTTON Update view to reflect current changes

  virtual int		BuildFromDataArray(DataArray_impl* tda=NULL);

  static String	CleanName(String& name);

  void	UpdateAfterEdit();
  void	Initialize();
  void	Destroy()	{ CutLinks(); }
  void	CutLinks();
  void	Copy_(const DA_ViewSpec& cp);
  COPY_FUNS(DA_ViewSpec,taNBase);
  TA_BASEFUNS(DA_ViewSpec);
};

class DT_ViewSpec :  public taGroup<DA_ViewSpec> {
  // base specification for the display of log data_table (DT)
public:
  DataTable*		data_table;	// #READ_ONLY the data table;
  String		display_name;	// name used in display
  bool	        	visible; 	// visibility flag

  virtual int		BuildFromDataTable(DataTable* tdt=NULL);

  void	UpdateAfterEdit();
  void	Initialize();
  void	Destroy();
  void	InitLinks();
  void	CutLinks();
  void	Copy_(const DT_ViewSpec& cp);
  COPY_FUNS(DT_ViewSpec, taGroup<DA_ViewSpec>);
  TA_BASEFUNS(DT_ViewSpec);
};

class DA_TextViewSpec: public DA_ViewSpec {
  // data-array view spec for text-based display
public:
  int		width;			// number of columns to use for display

  int		BuildFromDataArray(DataArray_impl* tda=NULL);

  void 	Initialize();
  void	Destroy();
  void Copy_(const DA_TextViewSpec& cp) { width = cp.width;}
  COPY_FUNS(DA_TextViewSpec,DA_ViewSpec);
  TA_BASEFUNS(DA_TextViewSpec);
};

class DA_NetViewSpec: public DA_ViewSpec {
  // data-array view spec for netview based display
public:
  int	label_index;	// index into the netview's label list
  void  Initialize();
  void  Destroy();
  void Copy_(const DA_NetViewSpec& cp) { label_index = cp.label_index;}
  COPY_FUNS(DA_NetViewSpec,DA_ViewSpec);
  TA_BASEFUNS(DA_NetViewSpec);
};
  
class DA_GraphViewSpec: public DA_ViewSpec {
  // data-array view spec for graph-based display
public:
  enum GV_Range {
    AUTO_SELF,			// autoscales to min/max of this array
    AUTO_GROUP,			// autoscale to min/max of all arrays which use this scale
    FIXED 			// does not auto scale
  };
  enum LineType {
    LINE = 0,			// just a line, no pts
    POINTS = 1,			// just pts, no line
    LINE_AND_POINTS = 2 	// both
  };
  enum LineStyle {
    SOLID 	= 0,		// -----
    DOT		= 1,		// .....
    DASH    	= 2,		// - - -
    DASH_DOT    = 3 		// _._._
  };
  enum PointStyle {
    NONE        = 0,            //
    SMALL_DOT	= 1,		// .
    BIG_DOT	= 2,		// o
    TICK	= 3,		// |
    PLUS	= 4 		// +
  };

  RGBA		line_color;	 // Color of the line
  TAColor	ta_color;	 // #IGNORE actual color pointer

  LineType	line_type;	// The way the line is drawn
  LineStyle	line_style;	// The style in which the line is drawn
  PointStyle	point_style;	// The style in which the points are drawn
  Modulo	point_mod;      // When to skip the drawing of point

  bool		negative_draw;	// Draw lines in the negative direction?

  DA_GraphViewSpec* axis_spec;  // #NO_NULL spec which indicates the axis to use
  GV_Range 	min_mode;	// how the minmium value on the axis is computed
  GV_Range 	max_mode;	// how the maximum value on the axis is computed
  MinMax 	range;		// range of the axis for fixed modes
  int		n_ticks;	// number of ticks to use in display

  int		BuildFromDataArray(DataArray_impl* tda=NULL);

  void 	Initialize();
  void	Destroy();
  void 	InitLinks();
  void  CutLinks();
  void  UpdateAfterEdit();
  void Copy_(const DA_GraphViewSpec& cp);
  COPY_FUNS(DA_GraphViewSpec,DA_ViewSpec);
  TA_BASEFUNS(DA_GraphViewSpec);
};

class DT_GraphViewSpec : public DT_ViewSpec {
  // controls display of datatable in a graph format
public:
  enum ColorType {		// defines standard color sequences
    C_RAINBOW,			// color, rainbow
    C_GREYSCALE,		// color, greyscale
    M_MONO,			// monochrome
    P_RAINBOW,			// printer, rainbow
    P_GREYSCALE,		// printer, greyscale
    P_MONO,			// printer, monochrome
    CUSTOM 			// custom, let user set colors
  };

  enum SequenceType {		// defines sequences of line features
    COLORS,			// use sequential colors
    LINES,			// use sequential line styles
    POINTS,			// use sequential point styles
    NONE 			// no features
  };

  RGBA			background;	// background color
  TAColor		bg_color;	// #IGNORE actual color object
  ColorType		color_type;	// palette of colors to use
  SequenceType		sequence_1;	// first (innermost) sequence of features
  SequenceType		sequence_2;	// second sequence of features
  SequenceType		sequence_3;	// third sequence of features

  virtual char*	ColorName(int color_no);
  // #MENU #MENU_ON_Actions #USE_RVAL gets color name for given line number for color_type
  virtual int 	ColorCount();		 // number of colors in palatte
  virtual void	SetBgColor();
  // sets background color based on color type
  virtual void	UpdateLineFeatures(bool visible_only=true);
  // #MENU #MENU_ON_Actions apply specified sequences to update line features
  virtual void	ApplyOneFeature(DA_GraphViewSpec* dagv, SequenceType seq, int val,
				bool& set_ln, bool& set_pt);
  // #IGNORE

  int		BuildFromDataTable(DataTable* tdt=NULL);

  void	UpdateAfterEdit();
  void	Initialize();
  void	Destroy();
  void	InitLinks();
  void	Copy_(const DT_GraphViewSpec& cp);
  COPY_FUNS(DT_GraphViewSpec, DT_ViewSpec);
  TA_BASEFUNS(DT_GraphViewSpec);
};

class DA_GridViewSpec : public DA_ViewSpec {
  // information for display of a data array in a grid display
public:
  enum DisplayStyle {
    TEXT,			// Draw using text only
    BLOCK,			// Draw using color block only
    TEXT_AND_BLOCK 		// Draw using both color block with text
  };

  PosTDCoord	pos;		// position of the data in absolute coordinates
  DisplayStyle  display_style;	// can display as text, block, or both
  bool		scale_on;	// adjust overall scale including this data

  int		BuildFromDataArray(DataArray_impl* tda=NULL);

  void 	Initialize();
  void	Destroy();
  void 	InitLinks();
  void	Copy_(const DA_GridViewSpec& cp);
  COPY_FUNS(DA_GridViewSpec, DA_ViewSpec);
  TA_BASEFUNS(DA_GridViewSpec);
};

class DT_GridViewSpec : public DT_ViewSpec {
  // information for display of a datatable in a grid display
public:
  enum BlockFill {		// ways that grid blocks can be filled
    COLOR,			// color indicates value
    AREA,			// area indicates value
    LINEAR 			// linear size of square side indicates value
  };

  enum MatrixLayout { // order of display of the grid elements
    DEFAULT,			// use current default layout
    LFT_RGT_BOT_TOP, // [3412] Incr col first, then decr row, start at bot left
    LFT_RGT_TOP_BOT, // [1234] Incr col first, then incr row, start at top left
    BOT_TOP_LFT_RGT, // [2413] Decr row first, then incr col, start at bot left
    TOP_BOT_LFT_RGT  // [1324] Incr row first, then incr col, start at top left
  };

  PosTDCoord	pos;		// position of the datatable in absolute coordinates
  PosTDCoord	geom;		// relative geometry (maximum extent) of the datatable, just for El's, not subgroups
  PosTDCoord	full_geom;	// #HIDDEN #NO_SAVE full absolute geometry (maximum extent) of everything under this one
  MatrixLayout	layout;		// current layout of the data table

  bool          use_gp_name;    // use the group name instead of the El names
  DA_GridViewSpec::DisplayStyle  display_style;	// can display as text, block, or both
  bool		scale_on;	// adjust overall scale including this data (or not)

  virtual int 	UpdateLayout(MatrixLayout ml=DEFAULT);
  // #MENU #MENU_ON_Actions enforce the geometry to fit with no spaces or overlap, returns maxx
  virtual void 	UpdateGeom();
  // #MENU Get the Geometry from the positions of visibles

  virtual int	BuildFromDataTable(DataTable* tdt=NULL);
  virtual void	GetMinMaxScale(MinMax& mm, bool first=true); // get min and max data range for scaling

  void 	UpdateAfterEdit();
  void 	Initialize();
  void	Destroy();
  void 	InitLinks();
  void	Copy_(const DT_GridViewSpec& cp);
  COPY_FUNS(DT_GridViewSpec, DT_ViewSpec);
  TA_BASEFUNS(DT_GridViewSpec);
};

#endif // datatable_h
