/* -*- C++ -*- */
/*=============================================================================
//									      //
// This file is part of the TypeAccess/C-Super-Script software package.	      //
//									      //
// Copyright (C) 1995 Randall C. O'Reilly, Chadley K. Dawson, 		      //
//		      James L. McClelland, and Carnegie Mellon University     //
//     									      //
// Permission to use, copy, modify, and distribute this software and its      //
// documentation for any purpose is hereby granted without fee, provided that //
// the above copyright notice and this permission notice appear in all copies //
// of the software and related documentation.                                 //
// 									      //
// Note that the PDP++ software package, which contains this package, has a   //
// more restrictive copyright, which applies only to the PDP++-specific       //
// portions of the software, which are labeled as such.			      //
//									      //
// Note that the taString class, which is derived from the GNU String class,  //
// is Copyright (C) 1988 Free Software Foundation, written by Doug Lea, and   //
// is covered by the GNU General Public License, see ta_string.h.             //
// The iv_graphic library and some iv_misc classes were derived from the      //
// InterViews morpher example and other InterViews code, which is             //
// Copyright (C) 1987, 1988, 1989, 1990, 1991 Stanford University             //
// Copyright (C) 1991 Silicon Graphics, Inc.				      //
//									      //
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,         //
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 	      //
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  	      //
// 									      //
// IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE FOR ANY SPECIAL,    //
// INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES  //
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT     //
// ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY,      //
// ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS        //
// SOFTWARE. 								      //
==============================================================================*/

#ifndef fun_lookup_h
#define fun_lookup_h

#include <ta_misc/datatable.h>
#include <ta_misc/ta_misc_TA_type.h>

class FunLookup : public float_RArray {
  // #NO_UPDATE_AFTER function lookup for non-computable functions
public:
  MinMaxRange	x_range;	// range of the x axis
  float		res;		// resolution of the function

  float&	Yval(float x)	// get y value at given x value (no interpolation)
  { return El( (int) ((x - x_range.min) / res)); }
  
  float		Xval(int idx)	// get x value for given index position within list
  { return x_range.min + ((float)idx * res); }

  float		Eval(float x);	// this does linear interpolation between points..
  
  virtual void	AllocForRange(); // allocate values for given range and resolution
  
  virtual void	Plot(ostream& strm); // generate file for plotting function

  virtual void	Convolve(const FunLookup& src, const FunLookup& con);
  // convolve source array with convolve array and put result here

  void	Initialize();
  void	Destroy()		{ };
  void	InitLinks();
  void 	Copy_(const FunLookup& cp);
  COPY_FUNS(FunLookup, float_RArray);
  TA_BASEFUNS(FunLookup);
};

#endif // fun_lookup_h
