/* -*- C++ -*- */
/*=============================================================================
//									      //
// This file is part of the TypeAccess/C-Super-Script software package.	      //
//									      //
// Copyright (C) 1995 Randall C. O'Reilly, Chadley K. Dawson, 		      //
//		      James L. McClelland, and Carnegie Mellon University     //
//     									      //
// Permission to use, copy, modify, and distribute this software and its      //
// documentation for any purpose is hereby granted without fee, provided that //
// the above copyright notice and this permission notice appear in all copies //
// of the software and related documentation.                                 //
// 									      //
// Note that the PDP++ software package, which contains this package, has a   //
// more restrictive copyright, which applies only to the PDP++-specific       //
// portions of the software, which are labeled as such.			      //
//									      //
// Note that the taString class, which is derived from the GNU String class,  //
// is Copyright (C) 1988 Free Software Foundation, written by Doug Lea, and   //
// is covered by the GNU General Public License, see ta_string.h.             //
// The iv_graphic library and some iv_misc classes were derived from the      //
// InterViews morpher example and other InterViews code, which is             //
// Copyright (C) 1987, 1988, 1989, 1990, 1991 Stanford University             //
// Copyright (C) 1991 Silicon Graphics, Inc.				      //
//									      //
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,         //
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 	      //
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  	      //
// 									      //
// IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE FOR ANY SPECIAL,    //
// INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES  //
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT     //
// ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY,      //
// ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS        //
// SOFTWARE. 								      //
==============================================================================*/

#ifndef tdgeometry_h
#define tdgeometry_h

#include <ta/ta_base.h>
#include <ta_misc/ta_misc_TA_type.h>

class TwoDCoord;
class PosTwoDCoord;
class TDCoord;
class PosTDCoord;
class FloatTwoDCoord;
class FloatTDCoord;

class TwoDCoord : public taBase {
  // ##NO_TOKENS #NO_UPDATE_AFTER #INLINE a value in 2D coordinate space
public:
  int 		x;  		// horizontal
  int 		y;  		// vertical

  void 		invert()			{ x = -x; y = -y; }
  void 		operator += (const TwoDCoord& td)	{ x += td.x; y += td.y; }
  void 		operator -= (const TwoDCoord& td)	{ x -= td.x; y -= td.y; }
  void 		operator *= (const TwoDCoord& td)	{ x *= td.x; y *= td.y; }
  void 		operator /= (const TwoDCoord& td)	{ x /= td.x; y /= td.y; }
  TwoDCoord 	operator + (const TwoDCoord& td) const;
  TwoDCoord 	operator - (const TwoDCoord& td) const;
  TwoDCoord 	operator * (const TwoDCoord& td) const;
  TwoDCoord 	operator / (const TwoDCoord& td) const;
  void		operator = (const FloatTwoDCoord& cp);

  bool		FitN(int n);		// adjust x and y to fit x total elements

  void		SetGtEq(int n)	// set each to be greater than or equal to n
  { x = MAX(n, x);  y = MAX(n, y); }
  void		SetLtEq(int n)	// set each to be less than or equal to n
  { x = MIN(n, x);  y = MIN(n, y); }

  int		SqDist(const TwoDCoord& td) const { // squared distance between two coords
    int xd = x - td.x; int yd = y - td.y;
    return ((xd * xd) + (yd * yd));
  }
  float		Dist(const TwoDCoord& td)  const { return sqrtf((float)SqDist(td)); }
  // distance between two coords

  static bool	WrapClipOne(bool wrap, int& c, int max); 
  // wrap-around or clip one dimension, true if clipped (coord set to -1)
  bool		WrapClip(bool wrap, const TwoDCoord& max) {
    return WrapClipOne(wrap, x, max.x) || WrapClipOne(wrap, y, max.y);
  } // wrap-around or clip coordinates within 0,0 - max range, true if clipped

  void 	Initialize() 		{ x = y = 0; }
  void 	Destroy()		{ };
  void 	Copy_(const TwoDCoord& cp);
  COPY_FUNS(TwoDCoord, taBase);
  TA_BASEFUNS(TwoDCoord);
  TwoDCoord(const FloatTwoDCoord& cp);	// conversion constructor
};

class PosTwoDCoord : public TwoDCoord {
  // #NO_UPDATE_AFTER #INLINE positive-only value in 2D coordinate space
public:
  void	UpdateAfterEdit();
  void	Initialize()		{ x = y = 0; }
  void	Destroy()		{ };
  TA_BASEFUNS(PosTwoDCoord);
};

class TDCoord : public TwoDCoord {
  // ##NO_TOKENS #NO_UPDATE_AFTER #INLINE a value in 3D coordinate space
public:
  int 		z;  		// depth

  void 		invert()			{ x = -x; y = -y; z = -z; }
  void 		operator += (const TDCoord& td)	{ x += td.x; y += td.y; z += td.z; }
  void 		operator -= (const TDCoord& td)	{ x -= td.x; y -= td.y; z -= td.z; }
  void 		operator *= (const TDCoord& td)	{ x *= td.x; y *= td.y; z *= td.z; }
  void 		operator /= (const TDCoord& td)	{ x /= td.x; y /= td.y; z /= td.z; }
  TDCoord 	operator + (const TDCoord& td) const;
  TDCoord 	operator - (const TDCoord& td) const;
  TDCoord 	operator * (const TDCoord& td) const;
  TDCoord 	operator / (const TDCoord& td) const;
  void		operator = (const FloatTDCoord& cp);

  bool		FitNinXY(int n);	// adjust x and y to fit x total elements

  void		SetGtEq(int n)	// set each to be greater than or equal to n
  { x = MAX(n, x);  y = MAX(n, y); z = MAX(n, z); }
  void		SetLtEq(int n)	// set each to be less than or equal to n
  { x = MIN(n, x);  y = MIN(n, y); z = MIN(n, z); }

  int		SqDist(const TDCoord& td) const { // squared distance between two coords
    int xd = x - td.x; int yd = y - td.y; int zd = z - td.z;
    return ((xd * xd) + (yd * yd) + (zd * zd));
  }
  float		Dist(const TDCoord& td)  const { return sqrtf((float)SqDist(td)); }
  // distance between two coords

  bool		WrapClip(bool wrap, const TDCoord& max) {
    return TwoDCoord::WrapClip(wrap, max) || WrapClipOne(wrap, z, max.z);
  } // wrap-around or clip coordinates within 0,0 - max range, -1 if clipped

  void 	Initialize() 		{ x = y = z = 0; }
  void 	Destroy()		{ };
  void 	Copy_(const TDCoord& cp);
  COPY_FUNS(TDCoord, TwoDCoord);
  TA_BASEFUNS(TDCoord);
  TDCoord(const FloatTDCoord& cp);	// conversion constructor
};

class PosTDCoord : public TDCoord {
  // #NO_UPDATE_AFTER #INLINE positive-only value in 3D coordinate space
public:
  void	UpdateAfterEdit();
  void	Initialize()		{ x = y = z = 0; }
  void	Destroy()		{ };
  TA_BASEFUNS(PosTDCoord);
};

class FloatTwoDCoord : public taBase {
  // ##NO_TOKENS #NO_UPDATE_AFTER #INLINE a value in 2D coordinate space
public:
  float		x;  		// horizontal
  float		y;  		// vertical

  void 		invert()				{ x = -x; y = -y; }
  void 		operator += (const FloatTwoDCoord& td)	{ x += td.x; y += td.y; }
  void 		operator -= (const FloatTwoDCoord& td)	{ x -= td.x; y -= td.y; }
  void 		operator *= (const FloatTwoDCoord& td)	{ x *= td.x; y *= td.y; }
  void 		operator /= (const FloatTwoDCoord& td)	{ x /= td.x; y /= td.y; }
  FloatTwoDCoord operator + (const FloatTwoDCoord& td) const;
  FloatTwoDCoord operator - (const FloatTwoDCoord& td) const;
  FloatTwoDCoord operator * (const FloatTwoDCoord& td) const;
  FloatTwoDCoord operator / (const FloatTwoDCoord& td) const;
  void		operator = (const TwoDCoord& cp);

  float		SqDist(const FloatTwoDCoord& td) const { // squared distance between two coords
    float xd = x - td.x; float yd = y - td.y;
    return ((xd * xd) + (yd * yd));
  }
  float		Dist(const FloatTwoDCoord& td) const { return sqrtf(SqDist(td)); }
  // distance between two coords

  void 	Initialize() 		{ x = y = 0; }
  void 	Destroy()		{ };
  void 	Copy_(const FloatTwoDCoord& cp);
  COPY_FUNS(FloatTwoDCoord, taBase);
  TA_BASEFUNS(FloatTwoDCoord);
  FloatTwoDCoord(const TwoDCoord& cp);	// conversion constructor
};

class FloatTDCoord : public FloatTwoDCoord {
  // ##NO_TOKENS #NO_UPDATE_AFTER #INLINE a real value in 3D coordinate space
public:
  float 	z;  		// depth

  void 		invert()			{ x = -x; y = -y; z = -z; }
  void 		operator += (const FloatTDCoord& td)	{ x += td.x; y += td.y; z += td.z; }
  void 		operator -= (const FloatTDCoord& td)	{ x -= td.x; y -= td.y; z -= td.z; }
  void 		operator *= (const FloatTDCoord& td)	{ x *= td.x; y *= td.y; z *= td.z; }
  void 		operator /= (const FloatTDCoord& td)	{ x /= td.x; y /= td.y; z /= td.z; }
  FloatTDCoord 	operator + (const FloatTDCoord& td) const;
  FloatTDCoord 	operator - (const FloatTDCoord& td) const;
  FloatTDCoord 	operator * (const FloatTDCoord& td) const;
  FloatTDCoord 	operator / (const FloatTDCoord& td) const;
  void		operator = (const TDCoord& cp);

  float		SqDist(const FloatTDCoord& td) const { // squared distance between two coords
    float xd = x - td.x; float yd = y - td.y; float zd = z - td.z;
    return ((xd * xd) + (yd * yd) + (zd * zd));
  }
  float		Dist(const FloatTDCoord& td) const { return sqrtf(SqDist(td)); }
  // distance between two coords

  void 	Initialize() 		{ x = y = z = 0.0f; }
  void 	Destroy()		{ };
  void 	Copy_(const FloatTDCoord& cp);
  COPY_FUNS(FloatTDCoord, FloatTwoDCoord);
  TA_BASEFUNS(FloatTDCoord);
  FloatTDCoord(const TDCoord& cp);	// conversion constructor
};

#endif // tdgeometry_h
